
#include "NIBO_NDS3.h"
#include <Arduino.h>
#include <../Wire/Wire.h>

#define NDS3_I2C_ADDR 42

#define MAKE_WORD(h, l)  ((((uint16_t)((uint8_t)h)) << 8) | ((uint8_t)l))

enum {
  CMD_RESET = 0x00,
  CMD_MOVE = 0x01,
  CMD_MEASURE = 0x02,
  CMD_READ = 0x03,
  CMD_GETPOS = 0x04,
  CMD_GETDIST = 0x05,
  CMD_GETBUSY = 0x06,
  CMD_SETDDR = 0x10,
  CMD_SETPORT = 0x11,
  CMD_GETPIN = 0x12,
  CMD_GETANALOG = 0x13,
  CMD_MOVESERVO = 0x20,
  CMD_GETSERVOCALIB = 0x21,
  CMD_SETSERVOCALIB = 0x22,
  CMD_GETDISTCALIB = 0x23,
  CMD_SETDISTCALIB = 0x24
};


NDS3_class::NDS3_class() {
}


void NDS3_class::begin() {
  address = NDS3_I2C_ADDR;
  Wire.begin();
}


void NDS3_class::begin(uint8_t _address) {
  address = _address;
}


uint8_t NDS3_class::read_8() {
  return Wire.read();
}


uint16_t NDS3_class::read_16() {
  uint8_t lo = Wire.read();
  uint8_t hi = Wire.read();
  return MAKE_WORD(hi, lo);
}


void NDS3_class::write_8(uint8_t val) {
  Wire.write(val);
}


void NDS3_class::write_16(uint16_t val) {
  Wire.write(highByte(val));
  Wire.write(lowByte(val));
}


void NDS3_class::command(uint8_t cmd) {
  Wire.beginTransmission(address);
  Wire.write(cmd);  
}


uint8_t NDS3_class::request(uint8_t size) {
  Wire.endTransmission(false);
  Wire.requestFrom(address, size);
  return Wire.available();
}


void NDS3_class::reset() {
  command(CMD_RESET);
  request(1);
}


void NDS3_class::move(uint8_t dir) {
  command(CMD_MOVE);
  write_8(dir);
  request(1);
}


void NDS3_class::measure(uint8_t dir, int8_t step) {
  command(CMD_MEASURE);
  write_8(dir);
  write_8(step);
  request(1);
}


void NDS3_class::read(uint8_t dir, uint8_t step) {
  command(CMD_READ);
  write_8(dir);
  write_8(step);
  uint8_t s = request(16);
  
  while ((dir<=180) && (s--)) {
    distances[dir] = read_8();
    dir += step;
  }
}


uint8_t NDS3_class::get_pos() {
  command(CMD_GETPOS);
  if (request(1)==1) {
    return read_8();
  }
  return 0;
}


uint8_t NDS3_class::get_dist() {
  command(CMD_GETDIST);
  if (request(1)==1) {
    return read_8();
  }
  return 0;
}


uint8_t NDS3_class::get_busy() {
  command(CMD_GETBUSY);
  if (request(1)==1) {
    return read_8();
  }
  return 0;
}

void NDS3_class::io_update() {
  /* should be replaced in protocol by single command! */
  port_commit();
  ddr_commit();
  pin_update();
}


void NDS3_class::ddr_commit() {
  command(CMD_SETDDR);
  write_8(io_ddr);
  request(1);
}


void NDS3_class::port_commit() {
  command(CMD_SETPORT);
  write_8(io_ddr);
  request(1);
}


void NDS3_class::pin_update() {
  command(CMD_GETPIN);
  if (request(1)==1) {
    io_pin = read_8();
  }
}


uint16_t NDS3_class::get_analog(uint8_t idx) {
  command(CMD_GETANALOG);
  if (request(2)==2) {
    return read_16();
  }
  return 0;
}


void NDS3_class::move_servo(uint16_t usecs) {
  command(CMD_MOVESERVO);
  write_16(usecs);
  request(1);
}


void NDS3_class::servocalib_update() {
  command(CMD_GETSERVOCALIB);
  if (request(6)==6) {
    servocalib_lo = read_16();
    servocalib_mid = read_16();
    servocalib_hi = read_16();
  }
}


void NDS3_class::servocalib_commit() {
  command(CMD_SETSERVOCALIB);
  write_16(servocalib_lo);
  write_16(servocalib_mid);
  write_16(servocalib_hi);
  request(1);
}


void NDS3_class::distcalib_update() {
  command(CMD_GETDISTCALIB);
  if (request(6)==6) {
    distcalib_v0 = read_16();
    distcalib_d0 = read_16();
    distcalib_alpha = read_16();
  }
}


void NDS3_class::distcalib_commit() {
  command(CMD_SETDISTCALIB);
  write_16(distcalib_v0);
  write_16(distcalib_d0);
  write_16(distcalib_alpha);
  request(1);
}




NDS3_class NDS3;


