/*
  NIBO2 library for ARDUINO
  License: BSD-License
  (c) 2013 by Nils Springob, nicai-systems
*/

/*! @file    NIBO_NDS3.h
 *  @brief   NIBO2 library for ARDUINO
 *  @author  Nils Springob (nils@nicai-systems.de)
 *  @date    2013-11-05
 */


#ifndef __NIBO_NDS3_H__
#define __NIBO_NDS3_H__

#ifdef ARDUINO
#include <Stream.h>
#else
#error "missing ARDUINO libs!"
#endif



class NDS3_class {
  
  private:
    uint8_t address;
    uint8_t read_8();
    uint16_t read_16();
    void write_8(uint8_t);
    void write_16(uint16_t);
    void command(uint8_t);
    uint8_t request(uint8_t);
    
    /* should be replaced in protocol by single command! */
    void ddr_commit();
    void port_commit();
    void pin_update();    
    
  public:
    NDS3_class();

    /*!
     * Initialisierung des NDS3 mit Standard Adresse
     */
    void begin();

    /*!
     * Initialisierung des NDS3 mit spezieller I2C Adresse
     */
    void begin(uint8_t address);

    /*!
     * Setzt das aktuelle Kommando zurück, nds3_get_busy() wird anschliessend 0 zurück geben.
     */
    void reset();

    /*!
     * Startet eine Drehung in die angegeben Richtung.
     * Die Drehung ist abgeschlossen, sobald nds3_get_busy() 0 zurück gibt.
     * @param dir Richtung in Grad [0 ... 180]
     */
    void move(uint8_t dir);

    /*!
     * Startet einen Scan. Der Sensor wird von der aktuellen Richtung mit der
     * Schritweite Step zur angegebenen Richtung gedreht.
     * Die Messung ist abgeschlossen, sobald nds3_get_busy() 0 zurück gibt.
     * @param dir Richtung in Grad [0 ... 180]
     * @param step Schrittweite in Grad
     */
    void measure(uint8_t dir, int8_t step);

    uint8_t distances[181];

    /*!
     * Liefert die Ergebnisse des letzten Scans zurück. Dir ist die Startrichtung,
     * step die Schrittweite. Beispiel:
     * \code
       // Scan von 0 bis 180 Grad in 5 Grad Schritten
       nds3_move(0);
       while(nds3_get_busy());
       nds3_measure(180, 5);
       while(nds3_get_busy());
       nds3_read(0, 5);
       \endcode
     * @param dir Richtung in Grad [-90 ... 90]
     * @param step Schrittweite in Grad
     */
    void read(uint8_t dir, uint8_t step);

    /*!
     * Liefert die aktuelle Position
     * @return Richtung in Grad [0 ... 180]
     */
    uint8_t get_pos();

    /*!
     * Liefert die aktuell gemessene Distanz
     * @return Distanz in cm
     */
    uint8_t get_dist();

    /*!
     * Prüfen, ob das NDS3 Modul ein (Bewegungs-)Kommando bearbeitet.
     * @returns 0: untätig
     */
    uint8_t get_busy();


    /*!
     * Data Direction Register (DDR)
     * Wenn das zugehörige Bit den Wert 0 hat, kann der Pin als Eingang verwendet werden, hat er den Wert 1 ist es ein Ausgang
     * Bits (7-0): --B2--B0--A3A2A1A0
     */
    uint8_t io_ddr;

    /*!
     * Ausgabe Register (PORT)
     * Wert für die Ausgabe  0 =^ 0V / 1 =^ 5V
     * Bits (7-0): --B2--B0--A3A2A1A0
     */
    uint8_t io_port;

    /*!
     * Port Input Register (PIN)
     * Eingangs-Logikpegel am Portpin: 0 =^ 0V / 1 =^ 5V
     * Bits (7-0): --B2--B0--A3A2A1A0
     */
    uint8_t io_pin;

    /*!
     * Aktualisiere PIN, PORT und DDR
     * Die Variablen io_ddr und io_port werden zum NDS3 übertragen,
     * die Variable io_pin wird vom NDS3 aktualisiert.
     */
    void io_update();

    /*!
     * Liefert den Wert des Analog-Eingangs Idx zurück
     * idx: 0-3 Spannung am Portpin A0-A3 (DDR muss auf Eingang geschaltet sein!)
     */
    uint16_t get_analog(uint8_t idx);

    /*!
     * Rohpulsweite des Servos in Microsekunden setzen
     */
    void move_servo(uint16_t usecs);
    
    /*!
     * Maximale Pulsweite des Servos in Microsekunden
     */
    uint16_t servocalib_hi;
    
    /*!
     * Mittlere Pulsweite des Servos in Microsekunden
     */
    uint16_t servocalib_mid;
    
    /*!
     * Minimale Pulsweite des Servos in Microsekunden
     */
    uint16_t servocalib_lo;
    
    /*!
     * Übertrage Servo-Kalibrierwerte vom NDS3 in die Variablen
     */
    void servocalib_update();
    
    /*!
     * Übertrage Servo-Kalibrierwerte aus den Variablen zum NDS3
     */
    void servocalib_commit();
    
    /*!
     * Voltage - Offset für Distanzmessung (* 5.0/1024 V)
     */
    uint16_t distcalib_v0;
    
    /*!
     * Voltage - Offset für Distanzmessung (* 1 cm)
     */
    uint16_t distcalib_d0;
    
    /*!
     * Voltage - Faktor
     */
    uint16_t distcalib_alpha;
    
    /*!
     * Übertrage Distanz-Kalibrierwerte vom NDS3 in die Variablen.
     * Die Werte werden im EEPROM gespeichert.
     * Die Formel lautet: d = alpha/(v-v0) - d
     * d: Distanz (*cm)
     * v: Spannung (*5.0/1024 V)
     */
    void distcalib_update();
    
    /*!
     * Übertrage Distanz-Kalibrierwerte aus den Variablen zum NDS3
     */
    void distcalib_commit();    

};


extern NDS3_class NDS3;



#endif


