/*
  NIBObee library for ARDUINO
  License: BSD-License
  (c) 2013 by Nils Springob, nicai-systems
*/

#include "NIBObee_BGX1.h"
#include "Arduino.h"
#include "../Wire/Wire.h"
#include "utility/bgx1_protocol.h"

#define BGX1_I2C_ADDR 11

#define MAKE_WORD(h, l)  ((((uint16_t)((uint8_t)h)) << 8) | ((uint8_t)l))


/**********************************************/


BGX1_class::BGX1_class() : term(this) {
}


void BGX1_class::begin() {
  address = BGX1_I2C_ADDR;
  Wire.begin();
}


void BGX1_class::begin(uint8_t _address) {
  address = _address;
}


uint8_t BGX1_class::read_8() {
  return Wire.read();
}


uint16_t BGX1_class::read_16() {
  uint8_t lo = Wire.read();
  uint8_t hi = Wire.read();
  return MAKE_WORD(hi, lo);
}


void BGX1_class::write_8(uint8_t val) {
  Wire.write(val);
}


void BGX1_class::write_16(uint16_t val) {
  Wire.write(highByte(val));
  Wire.write(lowByte(val));
}


void BGX1_class::command(uint8_t cmd) {
  Wire.beginTransmission(address);
  Wire.write(cmd);  
}


uint8_t BGX1_class::request(uint8_t size) {
  Wire.endTransmission(false);
  Wire.requestFrom(address, size);
  return Wire.available();
}


void BGX1_class::requestPosRes() {
  if (request(2)==2) {
    pos_x = read_8();
    pos_y = read_8();
  }
}


void BGX1_class::reset() {
  command(CMD_Reset);
  request(0);
}


uint16_t BGX1_class::getVersion() {
  command(CMD_GetVersion);
  if (request(2)==2) {
    return read_16();
  }
  return 0;
}


uint8_t BGX1_class::getStatus() {
  command(CMD_GetStatus);
  if (request(1)==1) {
    return read_8();
  }
  return 0;
}


void BGX1_class::setStatus(uint8_t status) {
  command(CMD_SetStatus);
  write_8(status);
  request(0);
}


void BGX1_class::move(uint8_t x, uint8_t y) {
  command(CMD_Move);
  write_8(x);
  write_8(y);
  request(0);
}


void BGX1_class::draw_mode(uint8_t mode) {
  command(CMD_Mode);
  write_8(mode);
  request(0);
}


void BGX1_class::fill(uint8_t pattern) {
  command(CMD_FillAll);
  write_8(pattern);
  request(0);
}


size_t BGX1_class::write(uint8_t c) {
  command(CMD_Print);
  write_8(c);
  requestPosRes();
  return 1;
}


size_t BGX1_class::write(const uint8_t *buffer, size_t size) {
  uint8_t res = size;
  command(CMD_Print);
  while (size--) {
    write_8(*buffer++);
  }
  requestPosRes();
  return res;
}


void BGX1_class::print_text(const char * text) {
  command(CMD_Print);
  while (*text) {
    write_8(*text++);
  }
  requestPosRes();
}


uint8_t BGX1_class::text_width(const char * text) {
  command(CMD_TextWidth);
  while (*text) {
    write_8(*text++);
  }
  if (request(1)==1) {
    return read_8();
  }
  return 0;
}


void BGX1_class::set_proportional(uint8_t id) {
  command(CMD_SelectFont);
  write_8(id);
  request(0);
}


void BGX1_class::hline(uint8_t length) {
  command(CMD_HLine);
  write_8(length);
  requestPosRes();
}


void BGX1_class::vline(uint8_t length) {
  command(CMD_VLine);
  write_8(length);
  requestPosRes();
}


void BGX1_class::box(uint8_t width, uint8_t height) {
  command(CMD_Box);
  write_8(width);
  write_8(height);
  requestPosRes();
}


void BGX1_class::lineTo(uint8_t x, uint8_t y) {
  command(CMD_LineTo);
  write_8(x);
  write_8(y);
  requestPosRes();
}


void BGX1_class::syncIO() {
  command(CMD_SyncPort);
  write_8(io_ddr);
  write_8(io_port);
  if (request(1)==1) {
    io_pin = read_8();
  }
}


uint16_t BGX1_class::getAnalog(uint8_t index) {
  command(CMD_GetAnalog);
  write_8(index);
  if (request(2)==2) {
    return read_16();
  }
  return 0;
}


void BGX1_class::syncInterface() {
  command(CMD_SyncInterface);
  write_8(leds);
  if (request(1)==1) {
    keys = read_8();
  }
}


void BGX1_class::setIllumination(uint16_t value) {
  command(CMD_SetIllumination);
  write_16(value);
  request(0);
}


/**********************************************/


BGX1_term::BGX1_term(BGX1_class * _bgx1) : bgx1(_bgx1) {
  
}


void BGX1_term::clear() {
  bgx1->command(CMD_TermClear);
  bgx1->request(0);
}


void BGX1_term::cursor(uint8_t col, uint8_t row) {
  bgx1->command(CMD_TermGoto);
  bgx1->write_8(col);
  bgx1->write_8(row);
  bgx1->request(0);
}


void BGX1_term::scroll(uint8_t lines) {
  bgx1->command(CMD_TermScroll);
  bgx1->write_8(lines);
  bgx1->request(0);
}


size_t BGX1_term::write(uint8_t c) {
  bgx1->command(CMD_TermPrint);
  bgx1->write_8(c);
  bgx1->request(1);
  return 1;
}


size_t BGX1_term::write(const uint8_t *buffer, size_t size) {
  uint8_t res = size;
  bgx1->command(CMD_TermPrint);
  while (size--) {
    bgx1->write_8(*buffer++);
  }
  bgx1->request(1);
  return res;
}


/**********************************************/


BGX1_class BGX1;
