/*  BSD-License:

Copyright (c) 2007 by Nils Springob, nicai-systems, Germany

All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

  * Redistributions of source code must retain the above copyright notice,
    this list of conditions and the following disclaimer.
  * Redistributions in binary form must reproduce the above copyright notice,
    this list of conditions and the following disclaimer in the documentation
    and/or other materials provided with the distribution.
  * Neither the name nicai-systems nor the names of its contributors may be
    used to endorse or promote products derived from this software without
    specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

*/

/*! @file    utility/motpid.cpp
 *  @brief   Zuordnung der physikalischen Pins zu symbolischen Namen
 *  @author  Nils Springob (nils@nicai-systems.de)
 *  @date    2009-08-19
 */
#define __STDC_LIMIT_MACROS
#include <avr/interrupt.h>
#include <avr/eeprom.h>
#include <util/atomic.h>
#include <stdint.h>
#include "utility/iodefs.h"
#include "utility/analog.h"
#include "utility/line.h"
#include "utility/base.h"
#include "utility/odometry.h"
#include "utility/clock.h"
#include "utility/motpwm.h"
#include "utility/motpid.h"

#include "NIBObee.h"



#define PWM_TIMER_MAX 511
#define PWM_TIMER_CNT TCNT1
#define PWM_TIMER_CCRA TCCR1A
#define PWM_TIMER_CCRA_INIT _BV(COM1A1) | _BV(COM1A0) | _BV(COM1B1) | _BV(COM1B0) | _BV(WGM11)
#define PWM_TIMER_CCRB TCCR1B
#define PWM_TIMER_CCRB_INIT _BV(CS10)
#define PWM_TIMER_OCRA OCR1A
#define PWM_TIMER_OCRA_INIT 0
#define PWM_TIMER_OCRB OCR1B
#define PWM_TIMER_OCRB_INIT 0
#ifndef TIMSK
# define PWM_TIMER_IMSK TIMSK1
#else
# define PWM_TIMER_IMSK TIMSK
#endif
#define PWM_TIMER_IMSK_INIT 0
#define PWM_TIMER_IMSK_CMPA _BV(OCIE1A)
#define PWM_TIMER_IMSK_CMPB _BV(OCIE1B) 

#define PWM_TIMER_IMSK_OIE   _BV(TOIE1)

// Clock = 15 Mhz
// PS = 1
// freq = 15Mhz/1022 = 14.677 kHz
// call PID with 100 Hz:
#define MOTPID_CALL_CNT 147

/*! PID will be called with 100 Hz */
#define MOTPID_FREQ 100

#define MOTPID_KP 5
#define MOTPID_KI 6
#define MOTPID_KD 7




#ifdef __cplusplus
extern "C" {
#endif



uint8_t motpid_cnt;

extern volatile int8_t odometry_pid_left;
extern volatile int8_t odometry_pid_right;
extern volatile int16_t odometry_left;
extern volatile int16_t odometry_right;

#define MOTOR_LEFT 0
#define MOTOR_RIGHT 1


motpid_t motpid[2];


void motpid_reset(uint8_t side) {
  motpid[side].a2 = 0;
  motpid[side].a1 = 0;
  motpid[side].o1 = 0;
  motpid[side].position_speed = 20;
}


void motpid_change_mode(uint8_t new_mode, uint8_t side) {
  if (motpid[side].mode==new_mode)
    return;
  motpid_reset(side);
  motpid[side].mode=new_mode;
}


void motpid_setSpeed(int16_t left, int16_t right) {
  motpid_change_mode(MOTPID_MODE_SPEED, 0);
  motpid_change_mode(MOTPID_MODE_SPEED, 1);
  motpid[0].value = left;
  motpid[1].value = right;
}


static int16_t motpid_calculate_single(int16_t a2, int16_t a1, int16_t o1, int16_t setpoint, int16_t reading) {
  int16_t temp = o1;
  if (temp >= +16384) temp = +16384;
  if (temp <= -16384) temp = -16384;
  temp += MOTPID_KI * (setpoint - reading);
  temp += MOTPID_KD * (2*a1 - a2 - reading);
  temp += MOTPID_KP * (a1 - reading);
  return temp;
}

// 100 Hz
void motpid_left_timer_int(void) {
  motpid[0].speed = (3*motpid[0].speed + (MOTPID_FREQ/2)*odometry_pid_left)/4;
  motpid[0].speed2 = (3*motpid[0].speed2 + motpid[0].speed)/4;
  if (motpid[0].mode) {
    int16_t value = motpid[0].speed;
    int16_t setpoint = motpid[0].value;
    if (motpid[0].mode & MOTPID_MODE_POS) {
      setpoint -= odometry_left;
      if (setpoint>motpid[0].position_speed) setpoint=motpid[0].position_speed;
      if (setpoint<-motpid[0].position_speed) setpoint=-motpid[0].position_speed;
      if (motpid[0].mode & MOTPID_FLAG_DEC) {
        // decrementing
        if (setpoint>=0) {
          setpoint = 0;
          motpid[0].a2 = motpid[0].a1 = motpid[0].o1 = 0;
        } else if (setpoint>-3) {
          setpoint = -3;
        }
      } else {
        // incrementing
        if (setpoint<=0) {
          setpoint = 0;
          motpid[0].a2 = motpid[0].a1 = motpid[0].o1 = 0;
        } else if (setpoint<3) {
          setpoint = 3;
        }
      }
    }

    motpid[0].o1 = motpid_calculate_single(motpid[0].a2,  motpid[0].a1, motpid[0].o1, setpoint,  value);
    motpid[0].a2 = motpid[0].a1;
    motpid[0].a1 = value;

    motpwm_setLeft(motpid[0].o1 / 16);
  }
  odometry_pid_left = 0;
}


// 100 Hz
void motpid_right_timer_int(void) {
  motpid[1].speed = (3*motpid[1].speed + (MOTPID_FREQ/2)*odometry_pid_right)/4;
  motpid[1].speed2 = (3*motpid[1].speed2 + motpid[1].speed)/4;
  if (motpid[1].mode) {
    int16_t value = motpid[1].speed;
    int16_t setpoint = motpid[1].value;
    if (motpid[1].mode & MOTPID_MODE_POS) {
      setpoint -= odometry_right;
      if (setpoint>motpid[1].position_speed) setpoint=motpid[1].position_speed;
      if (setpoint<-motpid[1].position_speed) setpoint=-motpid[1].position_speed;
      if (motpid[1].mode & MOTPID_FLAG_DEC) {
        // decrementing
        if (setpoint>=0) {
          setpoint = 0;
          motpid[1].a2 = motpid[1].a1 = motpid[1].o1 = 0;
        } else if (setpoint>-3) {
          setpoint = -3;
        }
      } else {
        // incrementing
        if (setpoint<=0) {
          setpoint = 0;
          motpid[1].a2 = motpid[1].a1 = motpid[1].o1 = 0;
        } else if (setpoint<3) {
          setpoint = 3;
        }
      }

    }

    motpid[1].o1 = motpid_calculate_single(motpid[1].a2,  motpid[1].a1, motpid[1].o1, setpoint,  value);
    motpid[1].a2 = motpid[1].a1;
    motpid[1].a1 = value;

    motpwm_setRight(motpid[1].o1 / 16);
  }
  odometry_pid_right = 0;
}


ISR(TIMER1_OVF_vect) {
  static uint8_t counter=0;
  static uint8_t running=0;
  counter++;
  if (running) {
    return;
  }
  running=1;
  if (counter>=15) {
    counter-=15;
    clock_inc_irq(1, 22);
  }

  motpid_cnt++;
  if (motpid_cnt>=MOTPID_CALL_CNT) {
    motpid_left_timer_int();
    //schedule_irq();
    motpid_right_timer_int();
    //schedule_irq();
    //comdata_signal();
    motpid_cnt=0;
  }
  running = 0;
}


void motpid_init() {
  if (!(nibobee_initialization&NIBOBEE_MOTPWM_INITIALIZED)) {
    motpwm_init();
  }
  PWM_TIMER_IMSK |= PWM_TIMER_IMSK_OIE;
  motpid_reset(0);
  motpid_reset(1);
  motpid[0].mode = MOTPID_MODE_OFF;
  motpid[1].mode = MOTPID_MODE_OFF;
}


void motpid_release(uint8_t side) {
  motpid_change_mode(MOTPID_MODE_OFF, side);
}


#ifdef __cplusplus
} // extern "C"
#endif




/*! @file    motpwm.c
 *  @brief   Zuordnung der physikalischen Pins zu symbolischen Namen
 *  @author  Nils Springob (nils@nicai-systems.de)
 *  @date    2009-08-19
 */


#ifdef __cplusplus
extern "C" {
#endif



//OC1A is IO_MOTOR_BIT_PWM_LEFT
//OC1B is IO_MOTOR_BIT_PWM_RIGHT

int16_t motpwm_motor_l;
int16_t motpwm_motor_r;
uint8_t motpwm_flags;

void motpwm_init() {
  nibobee_initialization |= NIBOBEE_MOTPWM_INITIALIZED;
  set_output_group(IO_MOTOR);
  activate_output_group(IO_MOTOR);
  PWM_TIMER_CCRA = PWM_TIMER_CCRA_INIT;
  PWM_TIMER_CCRB = PWM_TIMER_CCRB_INIT;
  PWM_TIMER_IMSK |= PWM_TIMER_IMSK_INIT;
}


void motpwm_stop() {
  ATOMIC_BLOCK(ATOMIC_RESTORESTATE) {
    motpwm_motor_l = 0;
    motpwm_motor_r = 0;
    PWM_TIMER_OCRA = 0;
    PWM_TIMER_OCRB = 0;
  }
}


void motpwm_stopLeft() {
  ATOMIC_BLOCK(ATOMIC_RESTORESTATE) {
    motpwm_motor_l = 0;
    PWM_TIMER_OCRA = 0;
  }
}


void motpwm_stopRight() {
  ATOMIC_BLOCK(ATOMIC_RESTORESTATE) {
    motpwm_motor_r = 0;
    PWM_TIMER_OCRB = 0;
  }
}


void motpwm_setLeft(int16_t value) {
  value/=2;
  if (value>PWM_TIMER_MAX) value=PWM_TIMER_MAX;
  if (value<-PWM_TIMER_MAX) value=-PWM_TIMER_MAX;
  
  if (value<0) {
    motpwm_motor_l = - value;
    if (get_output_groupbit(IO_MOTOR, DIR_LEFT)) {
      // direction has changed, stop first
      motpwm_flags&=0x02;
      PWM_TIMER_OCRA = 0;
      PWM_TIMER_IMSK |= PWM_TIMER_IMSK_CMPA;
    } else {
      PWM_TIMER_OCRA = motpwm_motor_l;
    }
  } else {
    motpwm_motor_l = value;
    if (get_output_groupbit(IO_MOTOR, DIR_LEFT)) {
      PWM_TIMER_OCRA = motpwm_motor_l;
    } else {
      // direction has changed, stop first
      motpwm_flags|=0x01;
      PWM_TIMER_OCRA = 0;
      PWM_TIMER_IMSK |= PWM_TIMER_IMSK_CMPA;
    }
  }
}


void motpwm_setRight(int16_t value) {
  value/=2;
  if (value>PWM_TIMER_MAX) value=PWM_TIMER_MAX;
  if (value<-PWM_TIMER_MAX) value=-PWM_TIMER_MAX;
  
  if (value<0) {
    motpwm_motor_r =  - value;
    if (get_output_groupbit(IO_MOTOR, DIR_RIGHT)) {
      PWM_TIMER_OCRB = motpwm_motor_r;
    } else {
      // direction has changed, stop first
      motpwm_flags|=0x02;
      PWM_TIMER_OCRB = 0;
      PWM_TIMER_IMSK |= PWM_TIMER_IMSK_CMPB;
    }
  } else {
    motpwm_motor_r =  value;
    if (get_output_groupbit(IO_MOTOR, DIR_RIGHT)) {
      // direction has changed, stop first
      motpwm_flags&=0x01;
      PWM_TIMER_OCRB = 0;
      PWM_TIMER_IMSK |= PWM_TIMER_IMSK_CMPB;
    } else {
      PWM_TIMER_OCRB = motpwm_motor_r;
    }
  }
}


ISR(TIMER1_COMPA_vect) {
  if (get_input_groupbit(IO_MOTOR, PWM_LEFT)) {
    set_output_groupbitval(IO_MOTOR, DIR_LEFT, motpwm_flags&0x01);
    PWM_TIMER_OCRA = motpwm_motor_l;
    PWM_TIMER_IMSK &= (uint8_t) (~PWM_TIMER_IMSK_CMPA);
  }
}


ISR(TIMER1_COMPB_vect) {
  if (get_input_groupbit(IO_MOTOR, PWM_RIGHT)) {
    set_output_groupbitval(IO_MOTOR, DIR_RIGHT, motpwm_flags&0x02);
    PWM_TIMER_OCRB = motpwm_motor_r;
    PWM_TIMER_IMSK &= (uint8_t) (~PWM_TIMER_IMSK_CMPB);
  }
}


#ifdef __cplusplus
} // extern "C"
#endif








