/*  BSD-License:

Copyright (c) 2007 by Nils Springob, nicai-systems, Germany

All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

  * Redistributions of source code must retain the above copyright notice,
    this list of conditions and the following disclaimer.
  * Redistributions in binary form must reproduce the above copyright notice,
    this list of conditions and the following disclaimer in the documentation
    and/or other materials provided with the distribution.
  * Neither the name nicai-systems nor the names of its contributors may be
    used to endorse or promote products derived from this software without
    specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

*/

/*! @file    utility/odometry.cpp
 *  @brief   Zuordnung der physikalischen Pins zu symbolischen Namen
 *  @author  Nils Springob (nils@nicai-systems.de)
 *  @date    2013-09-02
 */
#define __STDC_LIMIT_MACROS
#include <avr/interrupt.h>
#include <avr/eeprom.h>
#include <util/atomic.h>
#include <stdint.h>
#include "utility/iodefs.h"
#include "utility/analog.h"
#include "utility/line.h"
#include "utility/base.h"
#include "utility/clock.h"
#include "utility/motpwm.h"
#include "utility/motpid.h"

#include "NIBObee.h"


/*---------------------------------------------------------------------------*/


/*! @file    odometry.c
 *  @brief   Zuordnung der physikalischen Pins zu symbolischen Namen
 *  @author  Nils Springob (nils@nicai-systems.de)
 *  @date    2009-08-19
 */


// WGM=2, Phase Correct PWM, 9-bit (->WGM11)
// Set OC1A on compare match when upcounting. Clear OC1A on compare match when downcounting. (->COM1A1 COM1A0)
// Set OC1B on compare match when upcounting. Clear OC1B on compare match when downcounting. (->COM1B1 COM1B0)
// clkI/O/1 (No prescaling) (->CS10)
//
// Clock = 15 Mhz
// freq = 15Mhz/(2*(512-1)) = 14.677 kHz
// 


#define PWM_TIMER_MAX 511

#define PWM_TIMER_CNT        TCNT1
#define PWM_TIMER_CCRA       TCCR1A
#define PWM_TIMER_CCRA_INIT  _BV(COM1A1) | _BV(COM1A0) | _BV(COM1B1) | _BV(COM1B0) | _BV(WGM11)
#define PWM_TIMER_CCRB       TCCR1B
#define PWM_TIMER_CCRB_INIT  _BV(CS10)
#define PWM_TIMER_OCRA       OCR1A
#define PWM_TIMER_OCRA_INIT  0
#define PWM_TIMER_OCRB       OCR1B
#define PWM_TIMER_OCRB_INIT  0
#ifndef TIMSK
# define PWM_TIMER_IMSK      TIMSK1
#else
# define PWM_TIMER_IMSK      TIMSK
#endif
#define PWM_TIMER_IMSK_INIT  0
#define PWM_TIMER_IMSK_CMPA  _BV(OCIE1A)
#define PWM_TIMER_IMSK_CMPB  _BV(OCIE1B)
#define PWM_TIMER_IMSK_OIE   _BV(TOIE1)


// Clock = 15 Mhz
// PS = 1
// freq = 15Mhz/1022 = 14.677 kHz
// call PID with 100 Hz:
#define MOTPID_CALL_CNT 147

/*! PID will be called with 100 Hz */
#define MOTPID_FREQ 100

#define MOTPID_KP 5
#define MOTPID_KI 6
#define MOTPID_KD 7


#ifdef __cplusplus
extern "C" {
#endif


volatile int8_t odometry_pid_left;
volatile int8_t odometry_pid_right;
volatile int16_t odometry_left;
volatile int16_t odometry_right;

void odometry_init() {
  // enable int0 and int1 on rising edge, one tick per hole...
#ifndef GICR
  // ATmega644
  EICRA |= _BV(ISC11) | _BV(ISC10) | _BV(ISC01) | _BV(ISC00);
  EIMSK |= _BV(INT1) |  _BV(INT0);
#else
  // ATmega16
  MCUCR |= _BV(ISC11) | _BV(ISC10) | _BV(ISC01) | _BV(ISC00); 
  GICR |= _BV(INT1) |  _BV(INT0);
#endif

  set_output_groupbitval(IO_MOTOR, DIR_LEFT, 1);
  set_output_groupbitval(IO_MOTOR, DIR_RIGHT, 0);
  odometry_left = 0;
  odometry_right = 0;
}


int16_t odometry_getLeft(uint8_t reset) {
  ATOMIC_BLOCK(ATOMIC_RESTORESTATE) {
    int16_t res = odometry_left;
    if (reset) {
      odometry_left = 0;
    }
    return res;
  }
  return 0;
}


int16_t odometry_getRight(uint8_t reset) {
  ATOMIC_BLOCK(ATOMIC_RESTORESTATE) {
    int16_t res = odometry_right;
    if (reset) {
      odometry_right = 0;
    }
    return res;
  }
  return 0;
}

int16_t odometry_get(int id) {
  if (id) {
    return odometry_getRight(0);
  } else {
    return odometry_getLeft(0);
  }
}


ISR(INT0_vect) {
  if (get_output_groupbit(IO_MOTOR, DIR_LEFT)) odometry_pid_left++, odometry_left++; else odometry_pid_left--, odometry_left--;
}


ISR(INT1_vect) {
  if (get_output_groupbit(IO_MOTOR, DIR_RIGHT)) odometry_pid_right--, odometry_right--; else odometry_pid_right++, odometry_right++;
}


#ifdef __cplusplus
} // extern "C"
#endif
