/*  BSD-License:

Copyright (c) 2007 by Nils Springob, nicai-systems, Germany

All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

  * Redistributions of source code must retain the above copyright notice,
    this list of conditions and the following disclaimer.
  * Redistributions in binary form must reproduce the above copyright notice,
    this list of conditions and the following disclaimer in the documentation
    and/or other materials provided with the distribution.
  * Neither the name nicai-systems nor the names of its contributors may be
    used to endorse or promote products derived from this software without
    specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

*/

/*! @file    analog.h
 *  @brief   Zugriff auf die analogen Eingänge.
 *           Die analogen Eingänge werden automatisch gemessen, dazu müssen die globalen Interrupts aktiviert sein!
 *  @author  Nils Springob (nils.springob@nicai-systems.com)
 *  @date    2015-02-25
 */

#ifndef _NIBOBURGER_ANALOG_H_
#define _NIBOBURGER_ANALOG_H_


#ifdef __cplusplus
extern "C" {
#endif

enum {
  ANALOG_BL,   //!< Bodensensor Links / Sensor Vorne Links-Aussen
  ANALOG_BC,   //!< Bodensensor Mitte / Sensor Hinten Mitte
  ANALOG_BR,   //!< Bodensensor Rechts / Sensor Vorne Rechts-Aussen
  ANALOG_FL,   //!< Sensor Vorne Links
  ANALOG_FR,   //!< Sensor Vorne Rechts
  ANALOG_FLL,  //!< Sensor Vorne Links aussen
  ANALOG_FRR,  //!< Sensor Vorne Rechts aussen
  /* speziell */
  ANALOG_BCL,  //!< Bodensensor Links, Licht von Mitte Sensor
  ANALOG_BCR,  //!< Bodensensor Rechts, Licht von Mitte Sensor
  ANALOG_KEY,  //!< Funktionstasten
  ANALOG_VOLT, //!< Versorgungsspannung
  ANALOG_CNT
};


enum {
  ANALOG_PASSIVE       = 0, //!< passive Messung (LED aus)
  ANALOG_ACTIVE        = 1, //!< aktive Messung (LED an)
  ANALOG_DIFFERENTIAL  = 2  //!< passive+aktive Messung (Differenz)
};


enum {
  ANALOG_UPDATE_ASAP,    //!< immer Messen
  ANALOG_UPDATE_10MS,    //!< Messdurchlauf alle 10 ms starten
  ANALOG_UPDATE_MANUAL,  //!< Messdurchlauf manuell starten
};


/*!
 * Initialisierung des Liniensensors.
 * Der AD-Wandler misst im Interruptmodus automatisch alle Werte, und schaltet
 * auch die IR-LEDs der Bodensensoren bei Bedarf ein und aus.
 * Die Interrupts müssen aktiviert sein, damit die Sensorwerte gemessen werden!
 */
void analog_init();


/*!
 * Liefert den Wert eines analogen Eingangs mit dem index @param idx zurück.
 * Die Interrupts müssen aktiviert sein, damit die Werte gemessen werden!
 */
uint16_t analog_getValue(uint8_t idx);


/*!
 * Liefert den Wert eines analogen Eingangs mit dem index @param idx zurück.
 * Der Parameter @param mode darf die Werte ANALOG_PASSIVE, ANALOG_ACTIVE
 * oder ANALOG_DIFFERENTIAL annehmen.
 * Die Interrupts müssen aktiviert sein, damit die Werte gemessen werden!
 */
uint16_t analog_getValueExt(uint8_t idx, uint8_t mode);


/*!
 * Liefert einen Zufallswert zurück. Der Wert errechnet sich aus der Summe
 * aller gemessenen Analog-Spannungen. Der Wert selbst sollte nur als Seed
 * (Entropiequelle) verwendet werden, da er monoton wachsend ist. Der Wert
 * wird um so zufälliger, je mehr analoge Messungen vorher durchgeführt
 * wurden.
 * @return Random Seed
 */
uint16_t analog_getRandomSeed();


/*!
 * Aktiviert den automatischen Toggle-Mode (active/passive)
 */
void analog_setExtToggleMode(uint8_t idx, uint8_t activate);


/*!
 * Diese Funktion wird nach einem kompletten Durchgang aller Sampling-Kanäle aufgerufen
 */
void analog_irq_hook();


/*!
 * Diese Funktion wartet bis neue analoge Werte gemessen wurden
 */
void analog_wait_update();



/*!
 * Diese Funktion liefert TRUE wenn es seit dem letzten Aufruf neue analoge Messwerte gibt
 */
uint8_t analog_has_update();



#ifdef __cplusplus
} // extern "C"
#endif


#endif // _NIBOBURGER_ANALOG_H_

