/*  BSD-License:

Copyright (c) 2007 by Nils Springob, nicai-systems, Germany

All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

  * Redistributions of source code must retain the above copyright notice,
    this list of conditions and the following disclaimer.
  * Redistributions in binary form must reproduce the above copyright notice,
    this list of conditions and the following disclaimer in the documentation
    and/or other materials provided with the distribution.
  * Neither the name nicai-systems nor the names of its contributors may be
    used to endorse or promote products derived from this software without
    specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

*/

/*! @file    NIBObee.cpp
 *  @brief   Zuordnung der physikalischen Pins zu symbolischen Namen
 *  @author  Nils Springob (nils@nicai-systems.de)
 *  @date    2013-09-02
 */
#define __STDC_LIMIT_MACROS
#include <avr/interrupt.h>
#include <avr/eeprom.h>
#include <util/atomic.h>
#include <util/delay.h>
#include <stdint.h>
#include <stdlib.h>
#include "nibobee/iodefs.h"
#include "nibobee/analog.h"
#include "nibobee/line.h"
#include "nibobee/base.h"
#include "nibobee/odometry.h"
#include "nibobee/clock.h"
#include "nibobee/motpwm.h"
#include "nibobee/motpid.h"

#include "NIBObee.h"



EnginePartClass::EnginePartClass(unsigned char _id) : id(_id) {
}

int EnginePartClass::getTicks() {
  if (id) {
    return odometry_getRight(0);
  } else {
    return odometry_getLeft(0);
  }
}

void EnginePartClass::resetTicks() {
  if (id) {
    odometry_getRight(1);
  } else {
    odometry_getLeft(1);
  }
}


/*--------------------------------------------------------*/

EngineClass::EngineClass() : left(0), right(1) {
}


void EngineClass::begin() {
  motpwm_init();
  motpid_init();
  odometry_init();
}


int EngineClass::getTargetDist() {
  return motpid_getTargetDist();
}


void EngineClass::setPWM(int left, int right) {
  motpwm_setRight(right);
  motpwm_setLeft(left);
}


void EngineClass::setSpeed(int left, int right) {
  motpid_setSpeed(left, right);  
}


void EngineClass::setTargetAbs(int left, int right, unsigned int speed) {
  motpid_setTarget(left, right, speed);
}


void EngineClass::setTargetRel(int left, int right, unsigned int speed) {
  motpid_setTargetRel(left, right, speed);
}


void EngineClass::stopImmediate() {
  motpid_stop(1);
}


void EngineClass::stop() {
  motpid_stop(0);
}


/*---------------------------------------------------------------------------*/


LineSensorClass::LineSensorClass() {
}


void LineSensorClass::begin() {
  line_init();
  line_readPersistent();
}


unsigned int LineSensorClass::get(unsigned char id, unsigned char absolute) {
  if (id<3) {
    if (absolute) {
      return line_getAbs(id);
    } else {
      return line_get(id);
    }
  }
  return 0;
}


unsigned int LineSensorClass::getL(unsigned char absolute) {
  return get(0, absolute);
}


unsigned int LineSensorClass::getC(unsigned char absolute) {
  return get(1, absolute);
}


unsigned int LineSensorClass::getR(unsigned char absolute) {
  return get(2, absolute);
}


void LineSensorClass::calibrateWhite() {
  line_calibrateWhite();
}

void LineSensorClass::calibrateBlack() {
  line_calibrateBlack();
}


void LineSensorClass::storeCalibration() {
  line_writePersistent();
}


/*---------------------------------------------------------------------------*/


FeelerSensorClass::FeelerSensorClass(unsigned char _id) {
  id = _id; 
}


void FeelerSensorClass::begin() {  
  IO_SENS_PORT |= id ? (_BV(IO_SENS_BIT_RIGHT1) | _BV(IO_SENS_BIT_RIGHT2)) 
                     : (_BV(IO_SENS_BIT_LEFT1) | _BV(IO_SENS_BIT_LEFT2));
}


int FeelerSensorClass::get() {
  uint8_t val = IO_SENS_PIN;
  if (id==0) { 
    val &= _BV(IO_SENS_BIT_LEFT1) | _BV(IO_SENS_BIT_LEFT2);
    if (val==_BV(IO_SENS_BIT_LEFT1)) return -1;
    if (val==_BV(IO_SENS_BIT_LEFT2)) return 1;
  } else {
    val &= _BV(IO_SENS_BIT_RIGHT1) | _BV(IO_SENS_BIT_RIGHT2);
    if (val==_BV(IO_SENS_BIT_RIGHT1)) return -1;
    if (val==_BV(IO_SENS_BIT_RIGHT2)) return 1;
  }  
  return 0;
}




/*---------------------------------------------------------------------------*/

ExtSensorClass::ExtSensorClass(unsigned char _id) {
  id = _id;
}

void ExtSensorClass::activate(unsigned char mode) {
  switch(mode) {
    case SENSOR_DISABLED: // digital bit disconnected
    case SENSOR_PULSE:    // toggle digital bit
    case SENSOR_LOW:      // digital bit output low (0 V)
    case SENSOR_HIGH:     // digital bit output high (5 V)
      break;
  }
}

void ExtSensorClass::deactivate() {
  activate(SENSOR_DISABLED);
}

unsigned int ExtSensorClass::get(unsigned char absolute) {
  int16_t val16;
  if (absolute) {
    val16 = (int16_t)analog_getValue(ANALOG_EXT2+id);
  } else {
    val16 = (int16_t)analog_getValue(ANALOG_EXT2H+id)-(int16_t)analog_getValue(ANALOG_EXT2+id);
  }
  return val16;
}


/*---------------------------------------------------------------------------*/


NIBObeeClass::NIBObeeClass() {
}


void NIBObeeClass::begin() {
  IO_LEDS_DDR |= IO_LEDS_MASK;
  Engine.begin();
  LineSensor.begin();
  FeelerL.begin();
  FeelerR.begin();
  _delay_ms(100);
}


unsigned int NIBObeeClass::getVoltage() {
  unsigned int voltage = analog_getValue(ANALOG_VOLT);
  voltage = (4*12480U)/voltage;
  voltage *= 25;
  return voltage;
}


void NIBObeeClass::checkVoltage() {
  unsigned int voltage = getVoltage();
  if (voltage < 4600) {
    cli();
    while(1) {
      IO_LEDS_PORT = 0x0a;
      _delay_ms(50);
      IO_LEDS_PORT = 0x00;
      _delay_ms(50);
      IO_LEDS_PORT = 0x05;
      _delay_ms(50);
      IO_LEDS_PORT = 0x00;
      _delay_ms(50);
    }
  }
}


void NIBObeeClass::setLed(int led, int value) {
  if (value) {
    IO_LEDS_PORT |= (1<<led);
  } else {
    IO_LEDS_PORT &= ~(1<<led);
  }  
}


int NIBObeeClass::getLed(int led) {
  return (IO_LEDS_PORT & (1<<led))?1:0;  
}


unsigned int NIBObeeClass::getAnalog(unsigned char channel, unsigned char active) {
  if (active) {
    switch (channel) {
      case 0: channel = ANALOG_EXT0; break;
      case 1: channel = ANALOG_EXT1; break;
      case 2: channel = ANALOG_EXT2H; break;
      case 3: channel = ANALOG_EXT3H; break;
      case 4: channel = ANALOG_L1; break;
      case 5: channel = ANALOG_C1; break;
      case 6: channel = ANALOG_R1; break;
      default: return 0;
    }
  } else {
    switch (channel) {
      case 0: channel = ANALOG_EXT0; break;
      case 1: channel = ANALOG_EXT1; break;
      case 2: channel = ANALOG_EXT2; break;
      case 3: channel = ANALOG_EXT3; break;
      case 4: channel = ANALOG_L0; break;
      case 5: channel = ANALOG_C0; break;
      case 6: channel = ANALOG_R0; break;
      default: return 0;
    }
  }
  return analog_getValue(channel);
}


unsigned int NIBObeeClass::getRandomSeed() {
  return analog_getRandomSeed();
}


/*---------------------------------------------------------------------------*/

NIBObeeClass NIBObee;
EngineClass Engine;
LineSensorClass LineSensor;
FeelerSensorClass FeelerL(0);
FeelerSensorClass FeelerR(1);
ExtSensorClass ExtSensorL(0);
ExtSensorClass ExtSensorR(1);
