/*  BSD-License:

Copyright (c) 2007 by Nils Springob, nicai-systems, Germany

All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

  * Redistributions of source code must retain the above copyright notice,
    this list of conditions and the following disclaimer.
  * Redistributions in binary form must reproduce the above copyright notice,
    this list of conditions and the following disclaimer in the documentation
    and/or other materials provided with the distribution.
  * Neither the name nicai-systems nor the names of its contributors may be
    used to endorse or promote products derived from this software without
    specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

*/

#include "nibo/niboconfig.h"
#include "nibo/iodefs.h"
#include "nibo/delay.h"
#include "nibo/display.h"
#include "nibo/gfx.h"
#include "nibo/i2cmaster.h"
#include "nibo/motco.h"
#include "nibo/irco.h"
#include "nibo/adc.h"
#include "nibo/floor.h"
#include "nibo/leds.h"
#include "nibo/pwm.h"
#include "nibo/bot.h"

#include <avr/interrupt.h>
#include <avr/pgmspace.h>

#define LOBYTE(x)        (uint8_t)((uint16_t)x)
#define HIBYTE(x)        (uint8_t)(((uint16_t)x)>>8)
#define MAKE_WORD(hi,lo) ((hi*0x100)+lo)

typedef enum {
  MODE_STOP,
  MODE_STRAIGHTON,
  MODE_TURN_LEFT,
  MODE_TURN_RIGHT,
  MODE_ROTATE_LEFT,
  MODE_ROTATE_RIGHT,
  MODE_BACK
} mode_t;

int16_t tspeed_l = 0;
int16_t tspeed_r = 0;

uint8_t headlight_target;
uint8_t headlight_act;

void print_hex (uint8_t val)
{
  char c1=val/16;
  char c2=val%16;
  if (c1<10) c1+='0'; else c1+='a'-10;
  if (c2<10) c2+='0'; else c2+='a'-10;
  gfx_print_char(c1);
  gfx_print_char(c2);
}


int main(void) {
  sei(); // enable interrupts
  leds_init();
  pwm_init();
  bot_init();
  i2c_init();
  floor_init();

  if (display_init()!=0)
  {
    leds_set_displaylight(50);
    if (display_type==2)
    {
      gfx_init();
    }
  }
  
  gfx_fill(0x00);
  gfx_move(15, 0);
  gfx_set_proportional(1);
  gfx_print_text("nibo obstacle test");
  gfx_set_proportional(0);

  int8_t mode=2;

  irco_startMeasure();
  delay(10);
  //motco_setSpeedParameters(5, 4, 6); // ki, kp, kd
  motco_setSpeedParameters(15, 20, 10); // ki, kp, kd
  
  while (1)
  {
    delay(50);
    // Analog values:
    floor_disable_ir();
    delay(1);
    floor_update();
    int16_t value_fl = -floor_l;
    int16_t value_fr = -floor_r;
    int16_t value_ll = -line_l;
    int16_t value_lr = -line_r;
    floor_enable_ir();
    delay(1);
    floor_update();
    value_fl += floor_l;
    value_fr += floor_r;
    value_ll += line_l;
    value_lr += line_r;
    floor_disable_ir();

    if (value_fl<0) value_fl=0;
    if (value_fr<0) value_fr=0;
    if (value_ll<0) value_ll=0;
    if (value_lr<0) value_lr=0;
    if (value_fl>255) value_fl=255;
    if (value_fr>255) value_fr=255;
    if (value_ll>255) value_ll=255;
    if (value_lr>255) value_lr=255;

    gfx_move(30, 20);
    print_hex(value_fr);
    gfx_print_char(' ');
    print_hex(value_lr);
    gfx_print_char(' ');
    print_hex(value_ll);
    gfx_print_char(' ');
    print_hex(value_fl);

    // Distance

    int16_t speed_l = 0;
    int16_t speed_r = 0;

    // Request distance data
    if (!irco_update())
    {
      gfx_move(10, 10);
      gfx_set_proportional(1);
      gfx_print_text("IRCO Error   ");
      gfx_set_proportional(0);
      continue;
    }

    uint8_t dist_l  = irco_distance[4];
    uint8_t dist_vl = irco_distance[3];
    uint8_t dist_v  = irco_distance[2];
    uint8_t dist_vr = irco_distance[1];
    uint8_t dist_r  = irco_distance[0];

    gfx_move(25, 30);
    print_hex(dist_r);
    gfx_print_char(' ');
    print_hex(dist_vr);
    gfx_print_char(' ');
    print_hex(dist_v);
    gfx_print_char(' ');
    print_hex(dist_vl);
    gfx_print_char(' ');
    print_hex(dist_l);

    uint8_t ledr = 0x00;
    uint8_t ledg = 0x00;

    if (dist_r<0x30) ledg|=0x80;
    if (dist_vr<0x30) ledg|=0x40;
    if (dist_v<0x30) ledg|=0x30;
    if (dist_vl<0x30) ledg|=0x08;
    if (dist_l<0x30) ledg|=0x04;

    if (dist_r>0x90) ledr|=0x80;
    if (dist_vr>0x90) ledr|=0x40;
    if (dist_v>0x90) ledr|=0x30;
    if (dist_vl>0x90) ledr|=0x08;
    if (dist_l>0x90) ledr|=0x04;

    uint16_t sum = dist_r + dist_vr
                 + dist_v + dist_vl
                 + dist_l;
    sum /= 5;
    
    uint8_t dmax = (sum<0xe0)?(sum+0x20):0xff;
    uint8_t dmin = (sum>0x20)?(sum-0x20):0x00;

    if (dist_r<dmin) ledg|=0x80;
    if (dist_vr<dmin) ledg|=0x40;
    if (dist_v<dmin) ledg|=0x30;
    if (dist_vl<dmin) ledg|=0x08;
    if (dist_l<dmin) ledg|=0x04;

    if (dist_r>dmax) ledr|=0x80;
    if (dist_vr>dmax) ledr|=0x40;
    if (dist_v>dmax) ledr|=0x30;
    if (dist_vl>dmax) ledr|=0x08;
    if (dist_l>dmax) ledr|=0x04;

    ledr=0;
      
    if ((dist_l<254) && (dist_l<=dist_vl) && (dist_l<=dist_v) && (dist_l<=dist_vr) && (dist_l<=dist_r))
      speed_l=-10, speed_r=10, ledr= 0x04, mode=-1; // links frei!
    if ((dist_r<254) && (dist_r<=dist_l) && (dist_r<=dist_vl) && (dist_r<=dist_v) && (dist_r<=dist_vr))
      speed_l=10, speed_r=-10, ledr= 0x80, mode=1; // rechts frei!
    if ((dist_vl<254) && (dist_vl<=dist_l) && (dist_vl<=dist_v) && (dist_vl<=dist_vr) && (dist_vl<=dist_r))
      speed_l=40, speed_r=50, ledr= 0x08, mode=0; // vorne links frei!
    if ((dist_vr<254) && (dist_vr<=dist_l) && (dist_vr<=dist_vl) && (dist_vr<=dist_v) && (dist_vr<=dist_r))
      speed_l=50, speed_r=40, ledr= 0x40, mode=0; // vorne rechts frei!
    if (dist_v<254)
    {
      if (dist_v>5) dist_v-=5;
//      if ((dist_v<=dist_l) && (dist_v<=dist_vl) && (dist_v<=dist_vr) && (dist_v<=dist_r))
      if ((dist_v<=dist_vl) && (dist_v<=dist_vr))
        speed_l=60, speed_r=60, ledr= 0x30, mode=0; // vorne frei!
    }
      
    IO_LEDS_RED_PORT=ledr;
    IO_LEDS_GREEN_PORT=ledg;

    if (ledr&0x30)
      headlight_target=10;
    else
      headlight_target=0;

    if (headlight_act<headlight_target)
      headlight_act++;
    if (headlight_act>headlight_target)
      headlight_act--;

    if (headlight_act)
      leds_set_headlights(1<<(headlight_act-1));
    else
      leds_set_headlights(0);
    
    uint16_t lval=(dist_vl+dist_l)/2;
    uint16_t rval=(dist_vr+dist_r)/2;

    if (lval>rval) speed_r-=5; // rechts mehr Platz
    if (rval>lval) speed_l-=5; // links mehr Platz

    if (((tspeed_l<0) && (speed_l>0)) ||
        ((tspeed_l>0) && (speed_l<0)) ||
        ((tspeed_r<0) && (speed_r>0)) ||
        ((tspeed_r>0) && (speed_r<0)))
    {
      tspeed_l = 0;
      tspeed_r = 0;
      
      if (!motco_stop())
      {
        gfx_move(10, 10);
        gfx_set_proportional(1);
        gfx_print_text("MOTCO Error");
        gfx_set_proportional(0);
        continue;
      }
    }
    else
    {
      tspeed_l = (3*tspeed_l+speed_l)/4;
      tspeed_r = (3*tspeed_r+speed_r)/4;
    
      if (!motco_setSpeed(tspeed_l, tspeed_r))
      {
        gfx_move(10, 10);
        gfx_set_proportional(1);
        gfx_print_text("MOTCO_Error");
        gfx_set_proportional(0);
        continue;
      }
    }

    gfx_move(25, 40);
    gfx_set_proportional(1);
    gfx_print_text("r:");
    gfx_set_proportional(0);
    print_hex(HIBYTE(tspeed_r));
    print_hex(LOBYTE(tspeed_r));
    gfx_set_proportional(1);
    gfx_print_text(" l:");
    gfx_set_proportional(0);
    print_hex(HIBYTE(tspeed_l));
    print_hex(LOBYTE(tspeed_l));

    delay(10);
    if (!motco_update())
    {
      gfx_move(10, 10);
      gfx_set_proportional(1);
      gfx_print_text("MOTCO-Error");
      gfx_set_proportional(0);
      continue;
    }

    gfx_move(25, 50);
    gfx_set_proportional(1);
    gfx_print_text("r:");
    gfx_set_proportional(0);
    print_hex(HIBYTE(motco_speed_r));
    print_hex(LOBYTE(motco_speed_r));
    gfx_set_proportional(1);
    gfx_print_text(" l:");
    gfx_set_proportional(0);
    print_hex(HIBYTE(motco_speed_l));
    print_hex(LOBYTE(motco_speed_l));


    gfx_move(10, 10);
    gfx_print_text("         ");
  }
}

