/*  BSD-License:

Copyright (c) 2007 by Nils Springob, nicai-systems, Germany

All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

  * Redistributions of source code must retain the above copyright notice,
    this list of conditions and the following disclaimer.
  * Redistributions in binary form must reproduce the above copyright notice,
    this list of conditions and the following disclaimer in the documentation
    and/or other materials provided with the distribution.
  * Neither the name nicai-systems nor the names of its contributors may be
    used to endorse or promote products derived from this software without
    specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

*/

#include "nibo/iodefs.h"
#include <util/delay.h>

#define LOBYTE(val)((uint8_t)(0xff&(((uint16_t)(val))>>0)))
#define HIBYTE(val)((uint8_t)(0xff&(((uint16_t)(val))>>8)))


/* Taken from the Atmel example:
/* | Enter programming mode            | 'P' |       |      | 13d |   1  |
;* | Report autoincrement address      | 'a' |       |      | 'Y' |      |
;* | Set address                       | 'A' | ah al |      | 13d |   2  |
;* | Write program memory, low byte    | 'c' |    dd |      | 13d |   3  |
;* | Write program memory, high byte   | 'C' |    dd |      | 13d |   3  |
;* | Issue Page Write                  | 'm' |       |      | 13d |      |
;* | Read program memory               | 'R' |       |dd(dd)|     |   4  |
;* | Write data memory                 | 'D' |    dd |      | 13d |      |
;* | Read data memory                  | 'd' |       |   dd |     |      |
;* | Chip erase                        | 'e' |       |      | 13d |      |
;* | Write lock bits                   | 'l' |    dd |      | 13d |      |
;* | Write fuse bits                   | 'f' |    dd |      | 13d |  11  |
;* | Read fuse and lock bits           | 'F' |       |   dd |     |  11  |
;* | Leave programming mode            | 'L' |       |      | 13d |   5  |
;* | Select device type                | 'T' |    dd |      | 13d |   6  |
;* | Read signature bytes              | 's' |       | 3*dd |     |      |
;* | Return supported device codes     | 't' |       | n*dd | 00d |   7  |
;* | Return software identifier        | 'S' |       | s[7] |     |   8  |
;* | Return sofware version            | 'V' |       |dd dd |     |   9  |
;* | Return hardware version           | 'v' |       |dd dd |     |   9  |
;* | Return programmer type            | 'p' |       |   dd |     |  10  |
;* | Set LED                           | 'x' |    dd |      | 13d |  12  |
;* | Clear LED                         | 'y' |    dd |      | 13d |  12  |
;* | Universial command                | ':' |  3*dd |   dd | 13d |      |
;* | New universal command             | '.' |  4*dd |   dd | 13d |      |
;* | Special test command              | 'Z' |  2*dd |   dd |     |      |
*/

uint8_t deviceType;
uint16_t addr;
uint8_t pgm_mode;


uint8_t exchange(uint8_t val) {
  uint8_t res=0x00;
  for (uint8_t i=0; i<8; ++i) {
    set_output_bitval(IO_ISP_MOSI, (val&0x80));
    val <<= 1;
    res <<= 1;
    res |= get_input_bit(IO_ISP_MISO)?0x01:0x00;

    set_output_bit(IO_ISP_SCK);
    _delay_us(8);
    clear_output_bit(IO_ISP_SCK);
    _delay_us(8);
  }
  return res;
}

void activateReset(uint8_t dev) {
  if (dev&_BV(IO_RESET_A_BIT))
    clear_output_bit(IO_RESET_A);
  if (dev&_BV(IO_RESET_B_BIT))
    clear_output_bit(IO_RESET_B);
  if (dev&_BV(IO_RESET_5_BIT))
    clear_output_bit(IO_RESET_5);
  if (dev&_BV(IO_RESET_3_BIT))
    clear_output_bit(IO_RESET_3);
}

void deactivateResets() {
  set_output_bit(IO_RESET_A);
  set_output_bit(IO_RESET_B);
  set_output_bit(IO_RESET_5);
  set_output_bit(IO_RESET_3);
}

uint8_t enterProgrammingMode(uint8_t dev)
{
  activate_output_bit(IO_RESET_5);
  activate_output_bit(IO_RESET_3);
  activate_output_bit(IO_RESET_A);
  activate_output_bit(IO_RESET_B);
  activate_output_bit(IO_ISP_SCK);
  activate_output_bit(IO_ISP_MOSI);
  deactivate_output_bit(IO_ISP_MISO);
  
  clear_output_bit(IO_ISP_SCK);
  //SPCR = _BV(SPE) | _BV(MSTR) | _BV(SPR1) | _BV(SPR0);

  deactivateResets();
  _delay_ms(10);
  activateReset(dev);

  _delay_ms(10);
  _delay_ms(10);
  
  exchange(0xac);          // SPI (byte 1)
  exchange(0x53);          // SPI (byte 2)

  uint8_t result=0;

  deviceType=0x20;
  
  if ((deviceType>=0x20) && (deviceType<=0x7f))
  {
    uint8_t count=32;
    do
    {
      if (exchange(0x00)==0x53) // SPI (byte 3)
      {
        result=1;
        break;
      }
      exchange(0x00);          // SPI (byte 4)
      exchange(0xac);          // SPI (byte 1)
      exchange(0x53);          // SPI (byte 2)
    } while (count--);
  }
  else
  {
    exchange(0x00); // SPI (byte 3)
  }
  exchange(0x00);          // SPI (byte 4)
  _delay_ms(10);
  return result;
}

void leaveProgrammingMode(uint8_t dev){
  deactivate_output_bit(IO_ISP_SCK);
  deactivate_output_bit(IO_ISP_MOSI);
  deactivateResets();
  _delay_ms(10);
}


void setAddress(uint16_t address){
  addr = address;
}


void writeProgramMemory(uint16_t val){
  exchange(0x40);
  exchange(HIBYTE(addr));
  exchange(LOBYTE(addr));
  exchange(LOBYTE(val));
  
  exchange(0x48);
  exchange(HIBYTE(addr));
  exchange(LOBYTE(addr));
  exchange(HIBYTE(val));
  
  addr++;
}

void issuePageWrite(void){
  exchange(0x4c);
  exchange(HIBYTE(addr-1));
  exchange(LOBYTE(addr-1));
  exchange(0x00);
  _delay_ms(6);
}

uint16_t readProgramMemory(void){
  exchange(0x28);
  exchange(HIBYTE(addr));
  exchange(LOBYTE(addr));
  uint8_t hi = exchange(0x00);
  exchange(0x20);
  exchange(HIBYTE(addr));
  exchange(LOBYTE(addr));
  uint8_t lo = exchange(0x00);
  addr++;
  return lo + (((uint16_t)hi)<<8);
}

void writeDataMemory(uint8_t val){
  exchange(0xc0);
  exchange(HIBYTE(addr));
  exchange(LOBYTE(addr));
  exchange(val);
  _delay_ms(6);
  addr++;
}

uint8_t readDataMemory(void){
  exchange(0xa0);
  exchange(HIBYTE(addr));
  exchange(LOBYTE(addr));
  uint8_t val = exchange(0x00);
  addr++;
  return val;
}

void chipErase(void){
  exchange(0xac);
  exchange(0x80);
  exchange(0x04);
  exchange(0x00);
  _delay_ms(10);
  _delay_ms(10);
  _delay_ms(10);
}

void writeLockBits(uint8_t val){
  exchange(0xac);
  exchange(0xe0);
  exchange(0x00);
  exchange(val);
  _delay_ms(10);
}

void writeFuseBits(void){
  IO_LEDS_RED_PORT = 10 <<2;
}

void readFuseAndLockBits(void){
  IO_LEDS_RED_PORT = 11 <<2;
}


void readSignatureBytes(void){
  exchange(0x30);
  exchange(0x00);
  exchange(0x02);
  uint8_t val = exchange(0x00);

  exchange(0x30);
  exchange(0x00);
  exchange(0x01);
  val = exchange(0x00);
  
  exchange(0x30);
  exchange(0x00);
  exchange(0x00);
  val = exchange(0x00);
}

void writeHiFuse(uint8_t val)
{
  exchange(0xac);
  exchange(0xa8);
  exchange(0x00);
  exchange(val);
  _delay_ms(10);
}

void writeLoFuse(uint8_t val)
{
  exchange(0xac);
  exchange(0xa0);
  exchange(0x00);
  exchange(val);
  _delay_ms(10);
}

void writeExtFuse(uint8_t val)
{
  exchange(0xac);
  exchange(0xa4);
  exchange(0x00);
  exchange(val);
  _delay_ms(10);
}


void universialCommand(void){
  /*
  uint8_t cmd1 = avrisp_getchar();
  uint8_t cmd2 = avrisp_getchar();
  uint8_t cmd3 = avrisp_getchar();
  uint8_t cmd4 = 0x00;
  exchange(cmd1);
  exchange(cmd2);
  exchange(cmd3);
  exchange(cmd4);
  _delay_ms(10);
  */
}

void newUniversalCommand(void){
  /*
  uint8_t cmd1 = avrisp_getchar();
  uint8_t cmd2 = avrisp_getchar();
  uint8_t cmd3 = avrisp_getchar();
  uint8_t cmd4 = avrisp_getchar();
  exchange(cmd1);
  exchange(cmd2);
  exchange(cmd3);
  uint8_t res = exchange(cmd4);
  avrisp_putchar(res);
  _delay_ms(10);
  */
}

