/*  BSD-License:

Copyright (c) 2007 by Nils Springob, nicai-systems, Germany

All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

  * Redistributions of source code must retain the above copyright notice,
    this list of conditions and the following disclaimer.
  * Redistributions in binary form must reproduce the above copyright notice,
    this list of conditions and the following disclaimer in the documentation
    and/or other materials provided with the distribution.
  * Neither the name nicai-systems nor the names of its contributors may be
    used to endorse or promote products derived from this software without
    specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

*/

/*!
 *  @file    ir-rc5.c
 *  @brief   Routinen für die Dekodierung von RC5-Fernbedienungs-Codes
 *  @author  Nils Springob (nils@nicai-systems.de)
 *  @date    2007-05-10
 */

// Infos:
// http://www.sbprojects.com/knowledge/ir/rc5.htm
// http://www.sprut.de/electronic/ir/rc5.htm

// ========================================================================
// RC5 Infrarot-Empfänger
// ========================================================================

#include <avr/interrupt.h>
#include "iodefs.h"
#include "ir_rc5.h"

// -----------------------------------------------------------------------------
// Timer
// -----------------------------------------------------------------------------
#define IR_TIMER_CLOCK      5619
#define IR_SIGNAL_OC        TIM1_COMPA_vect
#define IR_TIMER_CNT        TCNT1
#define IR_TIMER_CCRA       TCCR1A
#define IR_TIMER_CCRA_INIT  0x00
#define IR_TIMER_CCRB       TCCR1B
#define IR_TIMER_CCRB_INIT  (_BV(CS11)|_BV(CS10))
#define IR_TIMER_OCR        OCR1A
#define IR_TIMER_OCR_INIT   (IR_COUNTVAL_SILENCE_MIN-1)
#define IR_TIMER_TIMSK      TIMSK1
#define IR_TIMER_TIMSK_INIT _BV(OCIE1A)


// -----------------------------------------------------------------------------
// Timing
// -----------------------------------------------------------------------------
#define IR_COUNTVAL_PULSE        111
#define IR_COUNTVAL_DELTA        (IR_COUNTVAL_PULSE/4)
#define IR_COUNTVAL_SHORT_MIN    (IR_COUNTVAL_PULSE-IR_COUNTVAL_DELTA)
#define IR_COUNTVAL_SHORT_MAX    (IR_COUNTVAL_PULSE+IR_COUNTVAL_DELTA)
#define IR_COUNTVAL_LONG_MIN     (2*IR_COUNTVAL_PULSE-IR_COUNTVAL_DELTA)
#define IR_COUNTVAL_LONG_MAX     (2*IR_COUNTVAL_PULSE+IR_COUNTVAL_DELTA)
#define IR_COUNTVAL_SILENCE_MIN  (90*IR_COUNTVAL_PULSE)

enum
{
  IR_STATE_SILENCE,
  IR_STATE_IDLE,
  IR_STATE_0,
  IR_STATE_1,
  IR_STATE_H,
  IR_STATE_L
};

enum
{
  IR_DATA_INVALID,
  IR_DATA_SHORT,
  IR_DATA_LONG
};
  

uint8_t ir_state;
uint16_t ir_data_tmp;
uint16_t ir_data;


ISR (IR_SIGNAL_OC)
{
  if (ir_state==IR_STATE_SILENCE)
  {
    IR_TIMER_CCRB = 0x00; // stop timer
    ir_state = IR_STATE_IDLE;
  }
  else
  {
    ir_state = IR_STATE_SILENCE;
    IR_TIMER_CNT = 0;
    IR_TIMER_OCR = IR_COUNTVAL_SILENCE_MIN;
    IR_TIMER_CCRB = IR_TIMER_CCRB_INIT; // start timer
  }
}

ISR (INT0_vect)
{
  uint16_t cnt = IR_TIMER_CNT;
  uint8_t data = IR_DATA_INVALID;
  if (cnt>IR_COUNTVAL_LONG_MAX) data = IR_DATA_INVALID;
  else if (cnt>=IR_COUNTVAL_LONG_MIN) data = IR_DATA_LONG;
  else if (cnt>IR_COUNTVAL_SHORT_MAX) data = IR_DATA_INVALID;
  else if (cnt>=IR_COUNTVAL_SHORT_MIN) data = IR_DATA_SHORT;

  if (ir_state!=IR_STATE_SILENCE)
  {
    if (get_input_bit(IO_IR_RX))
    {
      if (ir_state==IR_STATE_1)
      {
        if (data == IR_DATA_SHORT)
          ir_state=IR_STATE_H;
        else if (data == IR_DATA_LONG)
          ir_state=IR_STATE_0;
        else
          ir_state=IR_STATE_SILENCE;
      }
      else if (ir_state==IR_STATE_L)
      {
        if (data == IR_DATA_SHORT)
          ir_state=IR_STATE_0;
        else
          ir_state=IR_STATE_SILENCE;
      }
      else
      {
        ir_state=IR_STATE_SILENCE;
      }
    }
    else
    {
      if (ir_state==IR_STATE_IDLE)
      {
        ir_state = IR_STATE_1;
        ir_data_tmp = 0x0000;
      }
      else if (ir_state==IR_STATE_0)
      {
        if (data == IR_DATA_SHORT)
          ir_state=IR_STATE_L;
        else if (data == IR_DATA_LONG)
          ir_state=IR_STATE_1;
        else
          ir_state=IR_STATE_SILENCE;
      }
      else if (ir_state==IR_STATE_H)
      {
        if (data == IR_DATA_SHORT)
          ir_state=IR_STATE_1;
        else
          ir_state=IR_STATE_SILENCE;
      }
      else
      {
        ir_state=IR_STATE_SILENCE;
      }
    }
  }


  // finally
  if (ir_state==IR_STATE_SILENCE)
  {
    IR_TIMER_CNT = 0;
    IR_TIMER_OCR = IR_COUNTVAL_SILENCE_MIN;
    IR_TIMER_CCRB = IR_TIMER_CCRB_INIT; // start timer
  }
  else if ( (ir_state==IR_STATE_0) | (ir_state==IR_STATE_1)
          | (ir_state==IR_STATE_H) | (ir_state==IR_STATE_L) )
  {
    IR_TIMER_CNT = 0;
    IR_TIMER_OCR = IR_COUNTVAL_LONG_MAX;
    IR_TIMER_CCRB = IR_TIMER_CCRB_INIT; // start timer
  }

  if (ir_state==IR_STATE_0)
  {
    ir_data_tmp<<=1;
  }
  if (ir_state==IR_STATE_1)
  {
    ir_data_tmp<<=1;
    ir_data_tmp|=1;
  }

  if ((ir_data_tmp & 0x3000)==0x3000)
  {
    ir_data=ir_data_tmp;
    ir_data_tmp=0x0000;
  }
  
}


/*!
 * IR-Daten lesen
 * @return wert von ir_data, löscht anschliessend ir_data
 */
uint16_t ir_read(void) {
  uint16_t retvalue = ir_data;
  ir_data = 0;
  return retvalue;
}


/*!
 * Init IR-System
 */
void ir_init(void) {
  deactivate_output_bit(IO_IR_RX); // Pin auf Input
  set_output_bit(IO_IR_RX);        // Pullup an

  IR_TIMER_CNT = 0;
  IR_TIMER_CCRA = IR_TIMER_CCRA_INIT;
  IR_TIMER_CCRB = IR_TIMER_CCRB_INIT;
  IR_TIMER_OCR = IR_TIMER_OCR_INIT;
  IR_TIMER_TIMSK = IR_TIMER_TIMSK_INIT;

  ir_state = IR_STATE_SILENCE;
  MCUCR = _BV(ISC00);
  GIMSK = _BV(INT0);
}
