/*  BSD-License:

Copyright (c) 2007 by Nils Springob, nicai-systems, Germany

All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

  * Redistributions of source code must retain the above copyright notice,
    this list of conditions and the following disclaimer.
  * Redistributions in binary form must reproduce the above copyright notice,
    this list of conditions and the following disclaimer in the documentation
    and/or other materials provided with the distribution.
  * Neither the name nicai-systems nor the names of its contributors may be
    used to endorse or promote products derived from this software without
    specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

*/

#include "iodefs.h"
#include "i2c.h"
#include <avr/interrupt.h>

uint8_t i2c_state;

uint8_t i2c_rx_size;
uint8_t i2c_tx_size;
uint8_t i2c_tx_pos;

uint8_t i2c_rx_buf[I2C_RX_BUF_SIZE];
uint8_t i2c_tx_buf[I2C_TX_BUF_SIZE];

enum
{
  I2C_Idle,
  I2C_Address,
  I2C_Rx,
  I2C_Rx_Ack,
  I2C_Tx,
  I2C_Tx_Ack,
  I2C_Locked
};




#define I2C_SET_READ_START() \
do { \
  i2c_state = I2C_Idle; \
  deactivate_output_bit(IO_I2C_SDA); \
  USICR = _BV(USISIE) | _BV(USIWM1) | _BV(USICS1); \
  USISR = _BV(USIOIF) | _BV(USIPF); \
} while(0)


static void i2c_set_read_start(void) {
  I2C_SET_READ_START();
}


static void i2c_set_send_ack(void) {
  USIDR = 0x00;
  activate_output_bit(IO_I2C_SDA);
  USISR = _BV(USIOIF) | (14 << USICNT0);
}


static void i2c_set_read_ack(void) {
  deactivate_output_bit(IO_I2C_SDA);
  USIDR = 0x00;
  USISR = _BV(USIOIF) | (14 << USICNT0);
}


static void i2c_set_send_data(void) {
  activate_output_bit(IO_I2C_SDA);
  USISR = _BV(USIOIF) | (0 << USICNT0);
}


static void i2c_set_read_data(void) {
  deactivate_output_bit(IO_I2C_SDA);
  USISR = _BV(USIOIF) | (0 << USICNT0);
}


static void i2c_restart_timer() {
  I2C_TIMER_TCNT = 0; // reset counter
  I2C_TIMER_TIFR = I2C_TIMER_TIMSK_ACTIVE;  // clear interrupt flag
  I2C_TIMER_TIMSK = I2C_TIMER_TIMSK_ACTIVE; // enable interrupt
}

static void i2c_stop_timer() {
  I2C_TIMER_TIMSK = I2C_TIMER_TIMSK_INIT;  // disable interrupt
}

#define I2C_RESET() \
do { \
  i2c_stop_timer(); \
  I2C_SET_READ_START(); \
} while(0)

static void i2c_reset() {
  I2C_RESET();
}

ISR(I2C_TIMER_OCRA_vect) {
  // timeout
  I2C_RESET();
}

void i2c_init(void) {
  //I2C_TIMER_TCCRA = I2C_TIMER_TCCRA_INIT;
  I2C_TIMER_TCCRB = I2C_TIMER_TCCRA_INIT;
  //I2C_TIMER_TCCRC = I2C_TIMER_TCCRA_INIT;
  //I2C_TIMER_TIMSK = I2C_TIMER_TIMSK_INIT;
  I2C_TIMER_OCRA = I2C_TIMER_TIMEOUT;
  
  set_output_bit(IO_I2C_SCL);
  set_output_bit(IO_I2C_SDA);
  activate_output_bit(IO_I2C_SCL);
  i2c_set_read_start();
}

uint8_t i2c_lock(void) {
  uint8_t res = 0;
  cli(); // disable interrupts
  if (i2c_state==I2C_Rx) {
    if (USISR & _BV(USIPF)) {
      // STOP was received on last transmission, go to idle state!
      I2C_RESET();
    }
  }
  
  if (i2c_state==I2C_Idle) {
    USICR = _BV(USIWM1) | _BV(USICS1);
    i2c_state==I2C_Locked;
    res = 1;
  }
  sei(); // enable interrupts
  return res;
}


void i2c_unlock(void) {
  cli(); // disable interrupts
  i2c_state==I2C_Idle;
  USICR = _BV(USISIE) | _BV(USIWM1) | _BV(USICS1);
  sei(); // enable interrupts
}


ISR(USI_START_vect) {
  deactivate_output_bit(IO_I2C_SDA);
  /*
  do {
    if (USISR & _BV(USIPF)) {
      I2C_SET_READ_START();
      return;
    }
  } while (get_input_bit(IO_I2C_SCL));
  */
  i2c_restart_timer();
  i2c_state = I2C_Address;
  // now we should read 8 bits (16 toggles)
  // containing the address and the r/w bit
  USICR = _BV(USISIE) | _BV(USIOIE) | _BV(USIWM1) | _BV(USIWM0) | _BV(USICS1);
  USISR = _BV(USISIF) | _BV(USIOIF) | _BV(USIPF) | _BV(USIDC);
}


ISR(USI_OVF_vect) {
  i2c_restart_timer();
  switch (i2c_state) {
    case I2C_Address:
      if ((USIDR>>1)==I2C_ADDRESS) {
        if (USIDR&0x01) {
          i2c_state = I2C_Tx_Ack;
          i2c_tx_pos = 0;
        } else {
          i2c_state = I2C_Rx_Ack;
          i2c_rx_size = 0;
        }
        i2c_set_send_ack();
      } else {
        I2C_RESET();
      }
      break;
    
    case I2C_Rx:
      if (i2c_rx_size>=I2C_RX_BUF_SIZE) {
        // rx buffer is full -> ignore...
        I2C_RESET();
        break;
      }
      i2c_rx_buf[i2c_rx_size++] = USIDR;
      i2c_state=I2C_Rx_Ack;
      i2c_set_send_ack();
      break;
    
    case I2C_Rx_Ack:
      i2c_state = I2C_Rx;
      i2c_set_read_data();
      break;
    
    case I2C_Tx:
      if (i2c_tx_pos>=i2c_tx_size) {
        // no more tx data -> ignore...
        I2C_RESET();
        break;
      }
    
      i2c_state = I2C_Tx_Ack;
      i2c_set_read_ack();
      break;
    
    case I2C_Tx_Ack:
      if (USIDR) {
        // transmission was completed, master does not want more data
        I2C_RESET();
        break;
      }
      USIDR = i2c_tx_buf[i2c_tx_pos++];
      i2c_state = I2C_Tx;
      i2c_set_send_data();
      break;
  }
  
  if (USISR & _BV(USIPF)) {
    I2C_RESET();
  }
  
}
