/*  BSD-License:

Copyright (c) 2007 by Nils Springob, nicai-systems, Germany

All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

  * Redistributions of source code must retain the above copyright notice,
    this list of conditions and the following disclaimer.
  * Redistributions in binary form must reproduce the above copyright notice,
    this list of conditions and the following disclaimer in the documentation
    and/or other materials provided with the distribution.
  * Neither the name nicai-systems nor the names of its contributors may be
    used to endorse or promote products derived from this software without
    specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

*/

#include "iodefs.h"
#include "i2c.h"
#include "odometry.h"
#include "pid.h"
#include "pwm.h"
#include <util/delay.h>
#include <avr/interrupt.h>

#define LOBYTE(x)        ((uint8_t)(((uint16_t)(x))))
#define HIBYTE(x)        ((uint8_t)(((uint16_t)(x))>>8))
#define MAKE_WORD(hi,lo) (((((uint16_t)hi)<<8)) + ((uint16_t)lo))

// Default values for speed controller
int8_t KP_speed = 3; //5;
int8_t KI_speed = 5; //6;
int8_t KD_speed = 6; //7;


// Default values for position controller
int8_t KP_pos = 2;
int8_t KI_pos = 3;
int8_t KD_pos = 4;

/*
static uint16_t make_word(uint8_t hi, uint8_t lo) {
  uint16_t result;
  ((uint8_t*)(&result))[0]=lo;
  ((uint8_t*)(&result))[1]=hi;
  return result;
}
*/

int main(void) {
  //eeprom_restore_values();

  sei(); // enable interrupts
  odometry_init();
  i2c_init();
  pwm_init();

  while(1) {
    _delay_us(10);
    if (i2c_lock()) {
      if (i2c_rx_size>0) {
        uint8_t command = i2c_rx_buf[0];
        switch (command) {
          case 0x00: // Stop
            pid_change_mode(PID_MODE_OFF);
            pwm_motor_l = 0;
            pwm_motor_r = 0;
            break;
            
          case 0x01: // Set PWM
            if (i2c_rx_size==5) {
              pid_change_mode(PID_MODE_OFF);
              pwm_motor_l = MAKE_WORD(i2c_rx_buf[1], i2c_rx_buf[2]);
              pwm_motor_r = MAKE_WORD(i2c_rx_buf[3], i2c_rx_buf[4]);
            }
            break;
            
          case 0x02: // Set Speed
            if (i2c_rx_size==5) {
              pid_change_mode(PID_MODE_SPEED);
              pid_left_value  = MAKE_WORD(i2c_rx_buf[1], i2c_rx_buf[2]);
              pid_right_value = MAKE_WORD(i2c_rx_buf[3], i2c_rx_buf[4]);
            }
            break;
          
          case 0x03: // Set Speed Parameters
            if (i2c_rx_size==4) {
              KI_speed = i2c_rx_buf[1];
              KP_speed = i2c_rx_buf[2];
              KD_speed = i2c_rx_buf[3];
              // eeprom_store_values();
            }
            break;
          
          case 0x04: // Reset Odometry
            if (i2c_rx_size==5) {
              odometry_left  = MAKE_WORD(i2c_rx_buf[1], i2c_rx_buf[2]);
              odometry_right = MAKE_WORD(i2c_rx_buf[3], i2c_rx_buf[4]);
            }
            break;
          
          case 0x05: // Set Position Parameters
            if (i2c_rx_size==4) {
              KI_pos = i2c_rx_buf[1];
              KP_pos = i2c_rx_buf[2];
              KD_pos = i2c_rx_buf[3];
              // eeprom_store_values();
            }
            break;
          
          case 0x06: // Set Position
            if (i2c_rx_size==7) {
              pid_change_mode(PID_MODE_POS);      
              pid_left_value     = MAKE_WORD(i2c_rx_buf[1], i2c_rx_buf[2]);
              pid_right_value    = MAKE_WORD(i2c_rx_buf[3], i2c_rx_buf[4]);
              pid_position_speed = MAKE_WORD(i2c_rx_buf[5], i2c_rx_buf[6]);
            }
            break;
          
          case 0x07: // Stop immediate
            pid_change_mode(PID_MODE_STOP_IMM);
            pwm_motor_l = (odometry_dir_left>0)?-1023:+1023;
            pwm_motor_r = (odometry_dir_right>0)?-1023:+1023;
            pid_timeout = 100; // 0.5s  (200 is 1 s)
            break;
          
        }
        
        if (pid_mode==PID_MODE_SPEED) {
          pid_ki = KI_speed;
          pid_kp = KP_speed;
          pid_kd = KD_speed;
        } else if (pid_mode==PID_MODE_POS) {
          pid_ki = KI_pos;
          pid_kp = KP_pos;
          pid_kd = KD_pos;
        }

        i2c_rx_size=0;
      }
      i2c_tx_size=8;
      i2c_tx_buf[0]=HIBYTE(odometry_left);
      i2c_tx_buf[1]=LOBYTE(odometry_left);
      i2c_tx_buf[2]=HIBYTE(odometry_right);
      i2c_tx_buf[3]=LOBYTE(odometry_right);
      i2c_tx_buf[4]=HIBYTE(pid_left_speed2);
      i2c_tx_buf[5]=LOBYTE(pid_left_speed2);
      i2c_tx_buf[6]=HIBYTE(pid_right_speed2);
      i2c_tx_buf[7]=LOBYTE(pid_right_speed2);
//      i2c_tx_buf[4]=HIBYTE(pid_left_value);
//      i2c_tx_buf[5]=LOBYTE(pid_left_value);
//      i2c_tx_buf[6]=HIBYTE(pid_right_value);
//      i2c_tx_buf[7]=LOBYTE(pid_right_value);
      i2c_unlock();
    }
  }
}

