/*  BSD-License:

Copyright (c) 2007 by Nils Springob, nicai-systems, Germany

All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

  * Redistributions of source code must retain the above copyright notice,
    this list of conditions and the following disclaimer.
  * Redistributions in binary form must reproduce the above copyright notice,
    this list of conditions and the following disclaimer in the documentation
    and/or other materials provided with the distribution.
  * Neither the name nicai-systems nor the names of its contributors may be
    used to endorse or promote products derived from this software without
    specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

*/

#include "iodefs.h"
#include "odometry.h"
#include "pid.h"
#include <avr/interrupt.h>

#define SUM_DELTA 20
#define DIFF_DELTA 10

uint8_t odometry_cnt;
uint8_t odometry_val;
uint8_t odometry_state;

int8_t odometry_state_right;
int8_t odometry_state_left;
uint8_t odometry_min_sum_left;
uint8_t odometry_min_sum_right;

int16_t odometry_right;
int16_t odometry_left;
int8_t odometry_dir_left;
int8_t odometry_dir_right;



typedef struct
{
  uint8_t val1;
  uint8_t val2;
  int8_t diff_hi;
  int8_t diff_lo;
  uint8_t diff_state;
  uint8_t sum_limit;
  uint8_t sum_state;
  uint8_t sys_state;
} odometry_data;


static odometry_data data_left;
static odometry_data data_right;

enum
{
  ODO_ZERO,
  ODO_ZERO_LOFIX,
  ODO_ZERO_HIFIX,
  ODO_POS,
  ODO_NEG
};

enum
{
  SYS_PASSIVE,
  SYS_NEG,
  SYS_POS,
  SYS_ACT_NEG,
  SYS_ACT_POS
};



void update_sum(odometry_data * data)
{
  uint8_t sum = data->val1/2+data->val2/2;

  uint8_t sum_limit = data->sum_limit;
  uint8_t sum_state = data->sum_state;

  switch (sum_state)
  {
    case ODO_ZERO:
      if (sum<sum_limit)
      {
        sum_limit = sum+SUM_DELTA;
        sum_state = ODO_NEG;
      }
      else
      {
        if (sum_limit<sum-SUM_DELTA)
          sum_limit = sum-SUM_DELTA;
      }
      break;
      
    case ODO_NEG:
      if (sum>sum_limit)
      {
        sum_limit = sum-SUM_DELTA;
        sum_state = ODO_ZERO;
      }
      else
      {
        if (sum_limit>sum+SUM_DELTA)
          sum_limit = sum+SUM_DELTA;
      }
      break;
  }

  data->sum_limit = sum_limit;
  data->sum_state = sum_state;
}

void update_diff(odometry_data * data)
{
  int8_t diff = data->val1/2-data->val2/2;

  int8_t diff_hi = data->diff_hi;
  int8_t diff_lo = data->diff_lo;
  uint8_t diff_state = data->diff_state;

  if ((diff_lo<diff-DIFF_DELTA) && (diff_state!=ODO_ZERO_LOFIX))
    diff_lo = diff-DIFF_DELTA;
  if ((diff_hi>diff+DIFF_DELTA) && (diff_state!=ODO_ZERO_HIFIX))
    diff_hi = diff+DIFF_DELTA;

  if (diff_lo>diff)
    diff_lo = diff;

  if (diff_hi<diff)
    diff_hi = diff;


  switch (diff_state)
  {
    case ODO_ZERO:
    case ODO_ZERO_LOFIX:
    case ODO_ZERO_HIFIX:
      if (diff==diff_lo)
        diff_state = ODO_NEG;
      else if (diff==diff_hi)
        diff_state = ODO_POS;
      break;
      
    case ODO_NEG:
      if (diff==diff_hi)
        diff_hi = diff + 2*DIFF_DELTA, diff_state = ODO_ZERO_HIFIX;
      break;

    case ODO_POS:
      if (diff==diff_lo)
        diff_lo = diff - 2*DIFF_DELTA, diff_state = ODO_ZERO_LOFIX;
      break;
      
  }
  data->diff_hi = diff_hi;
  data->diff_lo = diff_lo;
  data->diff_state = diff_state;
}


int8_t calc_inc(odometry_data * data)
{
  int8_t result = 0;
  uint8_t sys_state = data->sys_state;
  uint8_t diff_state = data->diff_state;
  if (diff_state==ODO_ZERO_LOFIX) diff_state=ODO_ZERO;
  if (diff_state==ODO_ZERO_HIFIX) diff_state=ODO_ZERO;
  
  uint8_t sum_state=data->sum_state;
  switch (sys_state)
  {
    case SYS_PASSIVE:
      if (diff_state==ODO_NEG) sys_state=SYS_NEG;
      else if (diff_state==ODO_POS) sys_state=SYS_POS;
      break;
      
    case SYS_NEG:
      if (diff_state==ODO_ZERO)
      {
        if (sum_state==ODO_ZERO) sys_state=SYS_PASSIVE;
        else sys_state=SYS_ACT_NEG;
      }
      else if (diff_state==ODO_POS)
      {
        sys_state=SYS_POS; result=1;
      }
      break;
      
    case SYS_POS:
      if (diff_state==ODO_ZERO)
      {
        if (sum_state==ODO_ZERO) sys_state=SYS_PASSIVE;
        else sys_state=SYS_ACT_POS;
      }
      else if (diff_state==ODO_NEG)
      {
        sys_state=SYS_NEG; result=-1;
      }
      break;
      
    case SYS_ACT_NEG:
      if (diff_state==ODO_NEG)
        sys_state=SYS_ACT_NEG;
      else if (diff_state==ODO_POS)
        sys_state=SYS_POS, result=1;
      else if (sum_state==ODO_ZERO)
        sys_state=SYS_PASSIVE;
      break;
      
    case SYS_ACT_POS:
      if (diff_state==ODO_POS)
        sys_state=SYS_ACT_POS;
      else if (diff_state==ODO_NEG)
        sys_state=SYS_NEG, result=-1;
      else if (sum_state==ODO_ZERO)
        sys_state=SYS_PASSIVE;
      break;
  }
  data->sys_state=sys_state;
  return result;
}



void odometry_init(void)
{
  activate_output_bit(IO_IRLED);
  set_output_bit(IO_IRLED);

  ADCSRB = _BV(ADLAR); // right adjust result
  //odometry_cnt = 0;
  //ADMUX = 0x00 | odometry_cnt; // ref = Vcc
  ADCSRA = /*_BV(ADPS2) |*/ // prescale faktor = 16 ADC laeuft
           _BV(ADPS1)   // mit 8MHz / 16 = 500 kHz
         | _BV(ADPS0)
         | _BV(ADEN)   // ADC an
         | _BV(ADIF)   // Interrupt flag löschen
         | _BV(ADIE)   // Interrupt an
         | _BV(ADSC);  // Beginne mit der Konvertierung
}


ISR(ADC_vect)
{
  if (odometry_cnt==0){
    data_right.val1=ADCH;
  }
  if (odometry_cnt==1){
    data_right.val2=ADCH;
    update_sum(&data_right);
    update_diff(&data_right);
    int8_t inc = calc_inc(&data_right);
    pid_right_cnt -= inc;
    odometry_right -= inc;
    if (inc != 0) {
      odometry_dir_right = -inc;
    }
  }
  else if (odometry_cnt==2){
    data_left.val1=ADCH;
  }
  else if (odometry_cnt==3){
    data_left.val2=ADCH;
    update_sum(&data_left);
    update_diff(&data_left);
    int8_t inc = calc_inc(&data_left);
    pid_left_cnt += inc;
    odometry_left += inc;
    if (inc != 0) {
      odometry_dir_left = inc;
    }
  }
  
  //odometry_left = data_left.sum_state<<4 | data_left.diff_state;
  //odometry_left = data_left.val1;
  //odometry_right = data_left.val2;
  
  odometry_cnt = (odometry_cnt+1) & 0x03;
  ADMUX = 0x00 | odometry_cnt; // ref = Vcc
  
  ADCSRA = /* _BV(ADPS2)*/  // prescale faktor = 16 ADC laeuft
           _BV(ADPS1)   // mit 8MHz / 16 = 500kHz
         | _BV(ADPS0)
         | _BV(ADEN)   // ADC an
         | _BV(ADIF)   // Interrupt flag löschen
         | _BV(ADIE)   // Interrupt an
         | _BV(ADSC);  // Beginne mit der Konvertierung
}
