/*  BSD-License:

Copyright (c) 2007 by Nils Springob, nicai-systems, Germany

All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

  * Redistributions of source code must retain the above copyright notice,
    this list of conditions and the following disclaimer.
  * Redistributions in binary form must reproduce the above copyright notice,
    this list of conditions and the following disclaimer in the documentation
    and/or other materials provided with the distribution.
  * Neither the name nicai-systems nor the names of its contributors may be
    used to endorse or promote products derived from this software without
    specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

*/

/*! @file       pwm.c
 * @brief       Modul zur Erzeugung der Pwm-Signale
 * @author      Nils Springob (nils@nicai-systems.de)
 * @date        2007-01-05
*/


#include <avr/interrupt.h>
#include <stdlib.h> 

#include "iodefs.h"
#include "pid.h"
#include "pwm.h"

// Clock = 8 kHz
// PS = 1
// freq = 8Mhz/510 = 15.686 kHz
// call PID with 100 Hz:

#define PID_CALL_CNT 156

#define PWM_SIGNAL_OV        TIM0_OVF_vect
#define PWM_TIMER_CNT        TCNT0
#define PWM_TIMER_CCRA       TCCR0A
#define PWM_TIMER_CCRA_INIT  _BV(COM0A1) | _BV(COM0B1) | _BV(COM0B0) | _BV(WGM00)
#define PWM_TIMER_CCRB       TCCR0B
//#define PWM_TIMER_CCRB_INIT  _BV(CS01) | _BV(CS00)
#define PWM_TIMER_CCRB_INIT  _BV(CS00)
#define PWM_TIMER_OCRA       OCR0A
#define PWM_TIMER_OCRA_INIT  0
#define PWM_TIMER_OCRB       OCR0B
#define PWM_TIMER_OCRB_INIT  0

#define PWM_TIMER_IMSK       TIMSK0
#define PWM_TIMER_IMSK_INIT  _BV(TOIE0)

//OC0A is IO_MOTOR_R_EN
//OC0B is IO_MOTOR_L_EN


// diese Variablen enthalten die Soll-Werte: 
int16_t pwm_motor_l;
int16_t pwm_motor_r;
uint8_t pwm_cnt;

void pwm_init(void)
{
  //pwm_motor_l = 0;
  //pwm_motor_r = 0;
  //pwm_cnt = 0;

  //clear_output_bit(IO_MOTOR_L_EN);
  //clear_output_bit(IO_MOTOR_R_EN);

  activate_output_bit(IO_MOTOR_L_EN);
  activate_output_bit(IO_MOTOR_R_EN);
  activate_output_bit(IO_MOTOR_L_DIR);
  activate_output_bit(IO_MOTOR_R_DIR);

  //PWM_TIMER_CNT = 0;
  PWM_TIMER_CCRA = PWM_TIMER_CCRA_INIT;
  PWM_TIMER_CCRB = PWM_TIMER_CCRB_INIT;
  //PWM_TIMER_OCRA = PWM_TIMER_OCRA_INIT;
  //PWM_TIMER_OCRB = PWM_TIMER_OCRB_INIT;
  PWM_TIMER_IMSK |= PWM_TIMER_IMSK_INIT;
}

ISR(PWM_SIGNAL_OV)
{
  pwm_cnt++;
  if (pwm_cnt>=PID_CALL_CNT)
  {
    pid_left_timer_int();
    pid_right_timer_int();
    pwm_cnt=0;
  }

  // LEFT
  if (pwm_motor_l>=0)
  {
    if (get_activation_bit(IO_MOTOR_L_DIR))
    {
      activate_output_bit(IO_MOTOR_L_EN);
    }
    else
    {
      deactivate_output_bit(IO_MOTOR_L_EN);
      activate_output_bit(IO_MOTOR_L_DIR);
    }
    uint16_t val = pwm_motor_l;
    val >>= 2;
    if (val>255) val = 255;
    PWM_TIMER_OCRB = 255-((uint8_t)val);
  }
  else
  {
    if (get_activation_bit(IO_MOTOR_L_DIR))
    {
      deactivate_output_bit(IO_MOTOR_L_EN);
      deactivate_output_bit(IO_MOTOR_L_DIR);
    }
    else
    {
      activate_output_bit(IO_MOTOR_L_EN);
    }
    uint16_t val = -pwm_motor_l;
    val = val>>2;
    if (val>255) val = 255;
    PWM_TIMER_OCRB = 255-((uint8_t)val);
  }

  // RIGHT
  if (pwm_motor_r>=0)
  {
    if (get_activation_bit(IO_MOTOR_R_DIR))
    {
      activate_output_bit(IO_MOTOR_R_EN);
    }
    else
    {
      deactivate_output_bit(IO_MOTOR_R_EN);
      activate_output_bit(IO_MOTOR_R_DIR);
    }
    uint16_t val = pwm_motor_r;
    val >>= 2;
    if (val>255) val = 255;
    PWM_TIMER_OCRA = val;
  }
  else
  {
    if (get_activation_bit(IO_MOTOR_R_DIR))
    {
      deactivate_output_bit(IO_MOTOR_R_EN);
      deactivate_output_bit(IO_MOTOR_R_DIR);
    }
    else
    {
      activate_output_bit(IO_MOTOR_R_EN);
    }
    uint16_t val = -pwm_motor_r;
    val >>= 2;
    if (val>255) val = 255;
    PWM_TIMER_OCRA = val;
  }
}


