/*  BSD-License:

Copyright (c) 2008 by Nils Springob, nicai-systems, Germany

All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

  * Redistributions of source code must retain the above copyright notice,
    this list of conditions and the following disclaimer.
  * Redistributions in binary form must reproduce the above copyright notice,
    this list of conditions and the following disclaimer in the documentation
    and/or other materials provided with the distribution.
  * Neither the name nicai-systems nor the names of its contributors may be
    used to endorse or promote products derived from this software without
    specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

*/

#include "comdata.h"
#include "distance.h"
#include "odometry.h"
#include "pid.h"
#include "pwm.h"
#include "spimaster.h"
#include "persistent.h"
#include "ir_rc5.h"
#include "../../../src/nibo/copro_prot.h"
#include <avr/interrupt.h>
#include <string.h>
//#include <stdlib.h>

uint8_t comdata_state;
uint16_t rc5_code;

static uint8_t check_speed(int16_t speed) {
  return (speed>=-80) && (speed<=80);
}

static void comdata_execute(uint8_t * pos) {
  uint16_t code;
  switch (pos[0]) {
      case COPRO_CMD_STOP_IMMEDIATE:
        cli();
        pid_timeout = 100; // 0.5 s
        pwm_motor_l = (odometry_dir_left==1)?-1023:1023;
        pwm_motor_r = (odometry_dir_right==1)?-1023:1023;
        pid_change_mode(PID_MODE_STOP_IMM);
        sei();
        break;

      case COPRO_CMD_STOP:
        cli();
        pwm_motor_l = 0;
        pwm_motor_r = 0;
        pid_change_mode(PID_MODE_OFF);
        sei();
        break;

      case COPRO_CMD_SETSPEED:
        cli();
        pid_left_value  = - *(int16_t*)(pos+1);
        pid_right_value = - *(int16_t*)(pos+3);
        if (!check_speed(pid_left_value) || !check_speed(pid_right_value)) {
          pid_left_value=0;
          pid_right_value=0;
        }
        pid_change_mode(PID_MODE_SPEED);
        sei();
        break;

      case COPRO_CMD_SETPWM:
        cli();
        pwm_motor_l  = - *(int16_t*)(pos+1);
        pwm_motor_r  = - *(int16_t*)(pos+3);
        pid_change_mode(PID_MODE_OFF);
        sei();
        break;

      case COPRO_CMD_SETTARGET_ABS:
        cli();
        pid_left_value     = - *(int16_t*)(pos+1);
        pid_right_value    = - *(int16_t*)(pos+3);
        pid_position_speed = *(int16_t*)(pos+5);
        pid_change_mode(PID_MODE_POS);
        sei();
        break;
      
      case COPRO_CMD_SETTARGET_REL:
        cli();
        pid_left_value     = odometry_left  - *(int16_t*)(pos+1);
        pid_right_value    = odometry_right - *(int16_t*)(pos+3);
        pid_position_speed = *(int16_t*)(pos+5);
        pid_change_mode(PID_MODE_POS);
        sei();
        break;

      case COPRO_CMD_SETPARAMETERS:
        cli();
        persistent.pid_kp = *(int8_t*)(pos+1);
        persistent.pid_ki = *(int8_t*)(pos+2);
        persistent.pid_kd = *(int8_t*)(pos+3);
        persistent_dirty = 1;
        sei();
        break;
      
      case COPRO_CMD_START_IR:
	ir_set_tx_mode(IR_TX_DISTANCE);
        break;

      case COPRO_CMD_STOP_IR:
	ir_set_tx_mode(IR_TX_NONE);
        break;

      case COPRO_CMD_SEND_RC5:
	code = *(uint16_t*)(pos+1);
	ir_tx_rc5(code);
	ir_set_tx_mode(IR_TX_DATA);
        
        while (ir_tx_mode==IR_TX_DATA) {
        }
        break;
    }
     
}

uint16_t calculate_target_distance() {
  if (pid_mode != PID_MODE_POS) {
    return 0;
  }
  uint16_t l = abs(odometry_left - pid_left_value);
  uint16_t r = abs(odometry_right - pid_right_value);
  return (l>r)?l:r;
}

void comdata_schedule() {
  if (spi_busy()) {
    return;
  }

  uint8_t rx_size = spi_finished();
  if (rx_size==21) {
    comdata_execute(spi_buf+0);
    comdata_execute(spi_buf+7);
    comdata_execute(spi_buf+14);
  }

  switch (comdata_state) {
    case 0:
      return;
      
    case 1:
      // distance message
      if (distance_enable) {
        *(int16_t*)(spi_buf+0)  = distance_values[0];
        *(int16_t*)(spi_buf+2)  = distance_values[1];
        *(int16_t*)(spi_buf+4)  = distance_values[2];
        *(int16_t*)(spi_buf+6)  = distance_values[3];
        *(int16_t*)(spi_buf+8)  = distance_values[4];
      } else {
        memset(spi_buf+0, 0, 10);
      }
      *(int16_t*)(spi_buf+10) = -odometry_left;
      *(int16_t*)(spi_buf+12) = -odometry_right;
      *(int16_t*)(spi_buf+14) = -pid_left_speed2;
      *(int16_t*)(spi_buf+16) = -pid_right_speed2;
      *(int16_t*)(spi_buf+18) = analog_current_left;
      *(int16_t*)(spi_buf+20) = analog_current_right;
      uint16_t rc5_code_tmp = ir_read();
      if (rc5_code_tmp) {
        rc5_code = rc5_code_tmp;
      }
      spi_buf[22] = 0x00; // RC5
      *(uint16_t*)(spi_buf+23) = rc5_code;      
      *(uint16_t*)(spi_buf+25) = calculate_target_distance();
      
      spi_buf[28] = persistent_sysconf[0];
      spi_buf[29] = persistent_sysconf[1];
      spi_buf[30] = persistent_sysconf[2];
      spi_buf[31] = persistent_sysconf[3];
      spi_start(32);
      comdata_state = 0;
      return;
  }
}
