/*  BSD-License:

Copyright (c) 2007 by Nils Springob, nicai-systems, Germany

All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

  * Redistributions of source code must retain the above copyright notice,
    this list of conditions and the following disclaimer.
  * Redistributions in binary form must reproduce the above copyright notice,
    this list of conditions and the following disclaimer in the documentation
    and/or other materials provided with the distribution.
  * Neither the name nicai-systems nor the names of its contributors may be
    used to endorse or promote products derived from this software without
    specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

*/

/*!
 *  @file    ir-rc5.c
 *  @brief   Routinen für die Dekodierung von RC5-Fernbedienungs-Codes
 *  @author  Nils Springob (nils@nicai-systems.de)
 *  @date    2007-05-10
 */

// Infos:
// http://www.sbprojects.com/knowledge/ir/rc5.htm
// http://www.sprut.de/electronic/ir/rc5.htm

#define LOBYTE(x)        (uint8_t)((uint16_t)x)
#define HIBYTE(x)        (uint8_t)(((uint16_t)x)>>8)
#define MAKE_WORD(hi,lo) (((uint8_t)hi*(uint16_t)0x100)+(uint8_t)lo)

// ========================================================================
// RC5 Infrarot-Empfänger
// ========================================================================

#include <avr/interrupt.h>
#include "iodefs.h"
#include "ir_rc5.h"
#include "utils.h"

// -----------------------------------------------------------------------------
// Timer 0
// CTC mode
// PS 1
// TOP 111
// => 144.144 kHz (4 * 36 kHz)
// -----------------------------------------------------------------------------
#define IR_SIGNAL_OC        TIMER0_COMPA_vect
#define IR_TIMER_CNT        TCNT0
#define IR_TIMER_CCRA       TCCR0A
#define IR_TIMER_CCRA_INIT  (_BV(WGM01))
#define IR_TIMER_CCRB       TCCR0B
#define IR_TIMER_CCRB_INIT  (_BV(CS00))
#define IR_TIMER_OCR        OCR0A
#define IR_TIMER_OCR_INIT   (111-1)
#define IR_TIMER_TIMSK      TIMSK0
#define IR_TIMER_TIMSK_INIT _BV(OCIE0A)

/*
#define IR_TIMER_CLOCK      5619
#define IR_SIGNAL_OC        TIMER0_COMPA_vect
#define IR_TIMER_CNT        TCNT0
#define IR_TIMER_CCRA       TCCR0A
#define IR_TIMER_CCRA_INIT  0x00
#define IR_TIMER_CCRB       TCCR0B
#define IR_TIMER_CCRB_INIT  (_BV(CS01)|_BV(CS00))
#define IR_TIMER_OCR        OCR0A
#define IR_TIMER_OCR_INIT   (IR_COUNTVAL_SILENCE_MIN-1)
#define IR_TIMER_TIMSK      TIMSK0
#define IR_TIMER_TIMSK_INIT _BV(OCIE0A)
*/

// -----------------------------------------------------------------------------
// Timing
// -----------------------------------------------------------------------------
#define IR_COUNTVAL_PULSE        128
#define IR_COUNTVAL_DELTA        (IR_COUNTVAL_PULSE/4)
#define IR_COUNTVAL_SHORT_MIN    (IR_COUNTVAL_PULSE-IR_COUNTVAL_DELTA)
#define IR_COUNTVAL_SHORT_MAX    (IR_COUNTVAL_PULSE+IR_COUNTVAL_DELTA)
#define IR_COUNTVAL_LONG_MIN     (2*IR_COUNTVAL_PULSE-IR_COUNTVAL_DELTA)
#define IR_COUNTVAL_LONG_MAX     (2*IR_COUNTVAL_PULSE+IR_COUNTVAL_DELTA)
#define IR_COUNTVAL_SILENCE_MIN  (90*IR_COUNTVAL_PULSE)

enum {
  IR_STATE_SILENCE,
  IR_STATE_IDLE,
  IR_STATE_0,
  IR_STATE_1,
  IR_STATE_H,
  IR_STATE_L
};

enum {
  IR_DATA_INVALID_0,
  IR_DATA_INVALID_1,
  IR_DATA_SHORT_0,
  IR_DATA_SHORT_1,
  IR_DATA_LONG_0,
  IR_DATA_LONG_1
};
  

uint8_t ir_state;
uint16_t ir_data_tmp;
uint16_t ir_data;
uint16_t ir_counter;

uint16_t ir_debug;

volatile uint8_t ir_tx_mode;    // globale Variable um den Modus zum IR-Senden zu definieren

uint8_t ir_tx_pulse;   // 0 ... 128=4*32, Zähler für einzelne Pulse, zählt 144 kHz
uint8_t ir_tx_value;   // Aktuell zu übertragenes codiertes Byte, linksbündig
uint8_t ir_tx_data[4]; // Zu übertragene Bits nach Manchester Codierung (0->10, 1->01)
volatile uint8_t ir_tx_cnt;     // Zu übertragene Anzahl von codierten Bits (RC5: 2*14=28)
uint8_t ir_tx_bitpos;  // Index des aktuell übertragenes Bit
uint8_t ir_tx_bytepos; // Index des aktuell übertragenes Bytes


ISR (IR_SIGNAL_OC) {
  // called with 144 kHz
  
  if (ir_tx_mode==IR_TX_DATA) {
    if (ir_tx_pulse==0) {
      if (ir_tx_cnt==0) {
        clear_output_bit(IO_IR_TX);
        ir_tx_mode = IR_TX_NONE;
        return;
      }
      ir_tx_bitpos &= 0x07;
      if (ir_tx_bitpos==0) {
        ir_tx_value = ir_tx_data[ir_tx_bytepos];
        ir_tx_bytepos++;
      } else {
        ir_tx_value <<= 1;
      }
      ir_tx_bitpos++;
      ir_tx_cnt--;
    }
    if ( ((ir_tx_pulse&0x03) == 0x00) && (ir_tx_value&0x80) ) {
      set_output_bit(IO_IR_TX);
    } else {
      clear_output_bit(IO_IR_TX);
    }
    activate_output_bit(IO_IR_TX);
    ir_tx_pulse++;
    if (ir_tx_pulse==128) {
      ir_tx_pulse=0;
    }
    return;
  }

  if (ir_counter<IR_COUNTVAL_SILENCE_MIN) {
    if (ir_counter == IR_COUNTVAL_LONG_MAX) {
      // timeout
      ir_state = IR_STATE_SILENCE;
    }
    ir_counter++;
  } else {
    if (get_input_bit(IO_IR_RX)) {
      // minimum silence time reached
      IR_TIMER_CCRB = 0x00; // stop timer
      ir_state = IR_STATE_IDLE;
    }
  }
}

ISR (PCINT0_vect) {
  if (ir_tx_mode==IR_TX_DATA) {
    return;
  }
  uint16_t cnt = ir_counter;
  uint8_t data = IR_DATA_INVALID_0;
  if (cnt>IR_COUNTVAL_LONG_MAX) data = IR_DATA_INVALID_0;
  else if (cnt>=IR_COUNTVAL_LONG_MIN) data = IR_DATA_LONG_0;
  else if (cnt>IR_COUNTVAL_SHORT_MAX) data = IR_DATA_INVALID_0;
  else if (cnt>=IR_COUNTVAL_SHORT_MIN) data = IR_DATA_SHORT_0;
  if (!get_input_bit(IO_IR_RX)) {
    data++;
  }
  IR_TIMER_CNT = 0;
  IR_TIMER_CCRB = IR_TIMER_CCRB_INIT; // start timer
  ir_counter = 0;
  
  schedule_irq();

  if (ir_state==IR_STATE_SILENCE) {
    // do nothing, restart silence timer
  } else if (ir_state==IR_STATE_1) {
    if (data == IR_DATA_SHORT_0) {
      ir_state = IR_STATE_H;
    } else if (data == IR_DATA_LONG_0) {
      ir_state = IR_STATE_0;
    } else {
      ir_state = IR_STATE_SILENCE;
    }
  } else if (ir_state==IR_STATE_L) {
    if (data == IR_DATA_SHORT_0) {
      ir_state = IR_STATE_0;
    } else {
      ir_state = IR_STATE_SILENCE;
    }
  } else if (ir_state==IR_STATE_IDLE) {
    if (data&0x01) {
      ir_state = IR_STATE_1;
      ir_data_tmp = 0x0000;
    } else {
      ir_state=IR_STATE_SILENCE;
    }
  } else if (ir_state==IR_STATE_0) {
    if (data == IR_DATA_SHORT_1) {
      ir_state = IR_STATE_L;
    } else if (data == IR_DATA_LONG_1) {
      ir_state = IR_STATE_1;
    } else {
      ir_state = IR_STATE_SILENCE;
    }
  } else if (ir_state==IR_STATE_H) {
    if (data == IR_DATA_SHORT_1) {
      ir_state = IR_STATE_1;
    } else {
      ir_state = IR_STATE_SILENCE;
    }
  } else {
    ir_state = IR_STATE_SILENCE;
  }

  schedule_irq();

  if (ir_state==IR_STATE_0) {
    ir_data_tmp <<= 1;
  }
  
  if (ir_state==IR_STATE_1) {
    ir_data_tmp <<= 1;
    ir_data_tmp |= 1;
  }

  if ((ir_data_tmp & 0x3000)==0x3000) {
    ir_data = ir_data_tmp;
    ir_data_tmp = 0x0000;
  }
  
}

static void ir_tx_reset() {
  uint8_t i;
  ir_tx_cnt = 0;
  for (i=0; i<4; i++) {
    ir_tx_data[i] = 0;
  }
  ir_tx_cnt = 0;
  ir_tx_bitpos = 0x80;
  ir_tx_bytepos = 0;
  ir_tx_pulse = 0;
}

static void ir_tx_putbit(uint8_t val) {
  if (ir_tx_bitpos==0) {
    ir_tx_bitpos = 0x80;
    ir_tx_bytepos++;
  }
  if (val) {
    ir_tx_data[ir_tx_bytepos] += ir_tx_bitpos;
  }
  ir_tx_bitpos >>= 1;
  ir_tx_cnt++;
}

void ir_tx_stop() {
  ir_tx_cnt=0;
}

void ir_tx_rc5(uint16_t code) {
  uint8_t lo = LOBYTE(code);
  uint8_t hi = HIBYTE(code) | 0x20; // Startbit
  uint8_t i;
  ir_tx_reset();
  for (i=0; i<6; i++) {
    if (hi&0x20) {
      ir_tx_putbit(0);
      ir_tx_putbit(1);
    } else {
      ir_tx_putbit(1);
      ir_tx_putbit(0);
    }
    hi <<= 1;
  }
  for (i=0; i<8; i++) {
    if (lo&0x80) {
      ir_tx_putbit(0);
      ir_tx_putbit(1);
    } else {
      ir_tx_putbit(1);
      ir_tx_putbit(0);
    }
    lo <<= 1;    
  }
  
}

void ir_set_tx_mode(uint8_t mode) {
  //clear_output_bit(IO_IR_TX);
  //activate_output_bit(IO_IR_TX);
  ir_tx_mode = mode;
  if (mode==IR_TX_DATA) {
    IR_TIMER_CNT = 0;
    IR_TIMER_CCRB = IR_TIMER_CCRB_INIT; // start timer for transmission
    ir_tx_bitpos = 0;
    ir_tx_bytepos = 0;
  }
}


/*!
 * IR-Daten lesen
 * @return wert von ir_data, löscht anschliessend ir_data
 */
uint16_t ir_read(void) {
  uint16_t retvalue = ir_data;
  ir_data = 0;
  return retvalue;
}


/*!
 * Init IR-System
 */
void ir_init(void) {
 
  deactivate_output_bit(IO_IR_RX); // Pin auf Input
  //set_output_bit(IO_IR_RX);        // Pullup an
  clear_output_bit(IO_IR_RX);        // Pullup aus

  IR_TIMER_CNT = 0;
  IR_TIMER_CCRA = IR_TIMER_CCRA_INIT;
  IR_TIMER_CCRB = IR_TIMER_CCRB_INIT;
  IR_TIMER_OCR = IR_TIMER_OCR_INIT;
  IR_TIMER_TIMSK = IR_TIMER_TIMSK_INIT;

  // wait for silence
  ir_state = IR_STATE_SILENCE;
  PCICR |= _BV(PCIE0);
  PCMSK0 = _BV(0);
}
