/*  BSD-License:

Copyright (c) 2007 by Nils Springob, nicai-systems, Germany

All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

  * Redistributions of source code must retain the above copyright notice,
    this list of conditions and the following disclaimer.
  * Redistributions in binary form must reproduce the above copyright notice,
    this list of conditions and the following disclaimer in the documentation
    and/or other materials provided with the distribution.
  * Neither the name nicai-systems nor the names of its contributors may be
    used to endorse or promote products derived from this software without
    specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

*/

/*! @file       pid.c
 * @brief       Modul zur PID-Regelung der Motoren
 * @author      Nils Springob (nils@nicai-systems.de)
 * @date        2007-01-05
*/

#include "iodefs.h"
#include "pid.h"
#include "pwm.h"
#include "odometry.h"
#include "persistent.h"
        
/*! PID will be called with 100 Hz */
#define PID_FREQ 100


// constant values are 32-times
//#define PID_KI 6
//#define PID_KD 7
//#define PID_KP 5

int16_t pid_left_speed;
int16_t pid_right_speed;
int16_t pid_left_speed2;
int16_t pid_right_speed2;

int16_t pid_position_speed=20;

int16_t pid_left_value;
int16_t pid_right_value;

int8_t pid_left_cnt;
int8_t pid_right_cnt;

uint8_t pid_mode;
uint16_t pid_timeout;

typedef struct {
  int16_t a2;
  int16_t a1;
  int16_t o1;
}
pid_dyndata_t;

static pid_dyndata_t pid_dyndata_left;
static pid_dyndata_t pid_dyndata_right;


void pid_reset(void) {
  pid_dyndata_left.a2 = 0;
  pid_dyndata_left.a1 = 0;
  pid_dyndata_left.o1 = 0;
  pid_dyndata_right.a2 = 0;
  pid_dyndata_right.a1 = 0;
  pid_dyndata_right.o1 = 0;
}

void pid_change_mode(uint8_t new_mode) {
  if (pid_mode==new_mode)
    return;
  pid_reset();
  pid_mode=new_mode;
}


static int16_t pid_calculate_single(int16_t a2, int16_t a1, int16_t o1, int16_t setpoint, int16_t reading) {
  int16_t temp = o1;
  if (temp >= +16384) temp = +16384;
  if (temp <= -16384) temp = -16384;
  temp += persistent.pid_ki * (setpoint - reading);
  temp += persistent.pid_kd * (2*a1 - a2 - reading);
  temp += persistent.pid_kp * (a1 - reading);
  return temp;
}


void pid_left_timer_int(void) {
  if (pid_timeout) {
    pid_timeout--;
  }

  pid_left_speed = (3*pid_left_speed + (PID_FREQ/2)*pid_left_cnt)/4;
  pid_left_speed2 = (3*pid_left_speed2 + pid_left_speed)/4;
  if (pid_mode==PID_MODE_STOP_IMM) {
    if ( ((pwm_motor_l>0)&&(odometry_dir_left>0)) ||
         ((pwm_motor_l<0)&&(odometry_dir_left<0)) ||
         (pid_timeout==0) ) {
      pwm_motor_l=0;
    }
  } else if (pid_mode!=PID_MODE_OFF) {
    int16_t value = pid_left_speed;
    int16_t setpoint = pid_left_value;
    if (pid_mode==PID_MODE_POS) {
      setpoint -= odometry_left;
      if (setpoint>pid_position_speed) setpoint=pid_position_speed;
      if (setpoint<-pid_position_speed) setpoint=-pid_position_speed;
      if (setpoint==0) {
        pid_dyndata_left.a2 = pid_dyndata_left.a1 = pid_dyndata_left.o1 = 0;
      }
    }

    pid_dyndata_left.o1 = pid_calculate_single(pid_dyndata_left.a2,  pid_dyndata_left.a1,
        pid_dyndata_left.o1, setpoint,  value);
    pid_dyndata_left.a2 = pid_dyndata_left.a1;
    pid_dyndata_left.a1 = value;

    pwm_motor_l = pid_dyndata_left.o1 / 16;

  }
  pid_left_cnt = 0;
}

void pid_right_timer_int(void) {
  if (pid_timeout) {
    pid_timeout--;
  }

  pid_right_speed = (3*pid_right_speed + (PID_FREQ/2)*pid_right_cnt)/4;
  pid_right_speed2 = (3*pid_right_speed2 + pid_right_speed)/4;
  if (pid_mode==PID_MODE_STOP_IMM) {
    if ( ((pwm_motor_r>0)&&(odometry_dir_right>0)) ||
         ((pwm_motor_r<0)&&(odometry_dir_right<0)) ||
         (pid_timeout==0) ) {
      pwm_motor_r=0;
    }
  } else if (pid_mode!=PID_MODE_OFF) {
    int16_t value = pid_right_speed;
    int16_t setpoint = pid_right_value;
    if (pid_mode==PID_MODE_POS) {
      setpoint -= odometry_right;
      if (setpoint>pid_position_speed) setpoint=pid_position_speed;
      if (setpoint<-pid_position_speed) setpoint=-pid_position_speed;
      if (setpoint==0) {
        pid_dyndata_right.a2 = pid_dyndata_right.a1 = pid_dyndata_right.o1 = 0;
      }
    }
    
    pid_dyndata_right.o1 = pid_calculate_single(pid_dyndata_right.a2,  pid_dyndata_right.a1,
        pid_dyndata_right.o1, setpoint,  value);
    pid_dyndata_right.a2 = pid_dyndata_right.a1;
    pid_dyndata_right.a1 = value;
    
    pwm_motor_r = pid_dyndata_right.o1 / 16;
  }
  pid_right_cnt = 0;
}

