/*  BSD-License:

Copyright (c) 2007 by Nils Springob, nicai-systems, Germany

All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

  * Redistributions of source code must retain the above copyright notice,
    this list of conditions and the following disclaimer.
  * Redistributions in binary form must reproduce the above copyright notice,
    this list of conditions and the following disclaimer in the documentation
    and/or other materials provided with the distribution.
  * Neither the name nicai-systems nor the names of its contributors may be
    used to endorse or promote products derived from this software without
    specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

*/

/*! @file       pwm.c
 * @brief       Modul zur Erzeugung der Pwm-Signale
 * @author      Nils Springob (nils@nicai-systems.de)
 * @date        2007-01-05
*/

//#define NIBO_2_5

#include <avr/interrupt.h>
#include <stdlib.h> 

#include "iodefs.h"
#include "pid.h"
#include "pwm.h"
#include "utils.h"
#include "comdata.h"

// CLK = 16 MHz
// PS = 1
// TOP = 511
// freq = CLK/(PS*2*TOP)  16Mhz/1022 = 15.656 kHz
// call PID with 100 Hz:
// PWM, Phase Correct, 9-bit
// Set OC1A/OC1B on Compare Match when upcounting. Clear OC1A/OC1B on Compare Match when downcounting.

#define PID_CALL_CNT 156

#define PWM_SIGNAL_OV        TIMER1_OVF_vect
#define PWM_TIMER_CNT        TCNT1
#define PWM_TIMER_CCRA       TCCR1A
#define PWM_TIMER_CCRA_INIT  _BV(COM1A1) | _BV(COM1A0) | _BV(COM1B1) | _BV(COM1B0) | _BV(WGM11)
#define PWM_TIMER_CCRB       TCCR1B
//#define PWM_TIMER_CCRB_INIT  _BV(CS11) | _BV(CS10)
#define PWM_TIMER_CCRB_INIT  _BV(CS10)
#define PWM_TIMER_OCRA       OCR1A
#define PWM_TIMER_OCRA_INIT  0
#define PWM_TIMER_OCRB       OCR1B
#define PWM_TIMER_OCRB_INIT  0

#define PWM_TIMER_IMSK       TIMSK1
#define PWM_TIMER_IMSK_INIT  _BV(TOIE1)

//OC0A is IO_MOTOR_R_EN
//OC0B is IO_MOTOR_L_EN

// diese Variablen enthalten die Soll-Werte:
int16_t pwm_motor_l;
int16_t pwm_motor_r;
uint8_t pwm_cnt;

extern uint8_t is_nibo_2_5;


void pwm_init(void) {
  pwm_motor_l = 0;
  pwm_motor_r = 0;
  pwm_cnt = 0;


  clear_output_bit(IO_MOTOR_L_PWM);
  clear_output_bit(IO_MOTOR_R_PWM);

  activate_output_bit(IO_MOTOR_L_PWM);
  activate_output_bit(IO_MOTOR_R_PWM);
  activate_output_bit(IO_MOTOR_L_DIR);
  activate_output_bit(IO_MOTOR_R_DIR);

  PWM_TIMER_CCRA = PWM_TIMER_CCRA_INIT;
  PWM_TIMER_CCRB = PWM_TIMER_CCRB_INIT;
  PWM_TIMER_IMSK |= PWM_TIMER_IMSK_INIT;
}

static void do_right_pwm() {
  int16_t sval = pwm_motor_r;
  if (sval<0) {
    uint16_t val = -sval;
    if (get_output_bit(IO_MOTOR_R_DIR)) {
      clear_output_bit(IO_MOTOR_R_DIR);
    }
    val = val>>1;
    if (val>511) val = 511;
    PWM_TIMER_OCRB = (is_nibo_2_5)?(511-val):val;
  } else {
    uint16_t val = sval;
    if (!get_output_bit(IO_MOTOR_R_DIR)) {
      set_output_bit(IO_MOTOR_R_DIR);
    }
    val >>= 1;
    if (val>511) val = 511;
    PWM_TIMER_OCRB = val;
  } 
}

static void do_left_pwm() {
  int16_t sval = pwm_motor_l;
  if (sval<0) {
    uint16_t val = -sval;
    if (get_output_bit(IO_MOTOR_L_DIR)) {
      clear_output_bit(IO_MOTOR_L_DIR);
    }
    // L_DIR is LOW
    val >>= 1;
    if (val>511) val = 511;  
    PWM_TIMER_OCRA = (is_nibo_2_5)?(511-val):val;
  } else {
    uint16_t val = sval;
    if (!get_output_bit(IO_MOTOR_L_DIR)) {
      set_output_bit(IO_MOTOR_L_DIR);
    }
    // L_DIR is HIGH
    val >>= 1;
    if (val>511) val = 511;  
    PWM_TIMER_OCRA = val;
  }
}

ISR(PWM_SIGNAL_OV) {
  static uint8_t running=0;
  if (running) {
    return;
  }
  running=1;
  
  pwm_cnt++;
  if (pwm_cnt>=PID_CALL_CNT) {
    // 100 Hz
    pid_left_timer_int();
    schedule_irq();
    pid_right_timer_int();
    schedule_irq();
    comdata_signal();
    pwm_cnt=0;
  }
  
  do_left_pwm();
  
  schedule_irq();

  do_right_pwm();
  
  running=0;
}


