/*  BSD-License:

Copyright (c) 2008 by Nils Springob, nicai-systems, Germany

All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

  * Redistributions of source code must retain the above copyright notice,
    this list of conditions and the following disclaimer.
  * Redistributions in binary form must reproduce the above copyright notice,
    this list of conditions and the following disclaimer in the documentation
    and/or other materials provided with the distribution.
  * Neither the name nicai-systems nor the names of its contributors may be
    used to endorse or promote products derived from this software without
    specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

*/

#include "iodefs.h"
#include "spimaster.h"
#include "utils.h"
#include "comdata.h"
#include <avr/interrupt.h>
#include <string.h>

uint8_t spi_buf[32];

enum {
  SPI_IDLE,
  SPI_LOCKED,
  SPI_COMMAND,
  SPI_ADDRESS,
  SPI_READ,
  SPI_WRITE,
  SPI_CHK,
};

//uint8_t spi_rxbuf[sizeof(comdata)];

volatile uint8_t spi_status;
uint8_t spi_size;
uint8_t spi_pos;
uint8_t spi_rx_size;
uint8_t spi_tx_size;
uint8_t spi_rx_chk;
uint8_t spi_tx_chk;

inline static void spi_tx(uint8_t val) {
  SPDR = val;
  spi_tx_chk ^= val;
}

inline static uint8_t spi_rx() {
  uint8_t val = SPDR;
  spi_rx_chk ^= val;
  return val;
}

void spi_init() {
  set_output_bit(IO_SPI_SS);
  activate_output_bit(IO_SPI_SCK);
  activate_output_bit(IO_SPI_MOSI);
  activate_output_bit(IO_SPI_SS);

  /* Enable SPI, Master, set clock rate fck/16  -> 1us/Bit = 8us/Byte plus 32us Pause -> 40us/Byte == 250Byte/10ms */
  SPCR = _BV(SPIE)|_BV(SPE)|_BV(MSTR)|_BV(SPR0);
  spi_status = SPI_IDLE;
}

uint8_t spi_finished() {
  uint8_t res = spi_rx_size;
  spi_rx_size = 0;
  return res;
}

uint8_t spi_busy() {
  if (spi_size) {
    return 1;
  }
  return 0;
}


void spi_start(uint8_t size) {
  clear_output_bit(IO_SPI_SS);
  spi_tx_size = size;
  spi_size = size+1;
  spi_pos = 0;
  spi_rx_chk = 0x96;
  spi_tx_chk = 0x96;
  spi_tx(size);
  SPCR = _BV(SPIE)|_BV(SPE)|_BV(MSTR)|_BV(SPR1);
}


ISR(SPI_STC_vect) {
  // receive
  uint8_t rx_data = spi_rx();
  
  if (spi_pos==0) {
    spi_rx_size = rx_data;
    if (spi_rx_size>sizeof(spi_buf)) {
      spi_rx_size = 0;
      spi_size = 0;
    } else if (spi_rx_size+1>spi_size) {
      spi_size = spi_rx_size+1;
    }
  } else {
    if (spi_pos-1<spi_rx_size) {
      spi_buf[spi_pos-1]=rx_data;
    } else if (spi_pos-1==spi_rx_size) {
      if (spi_rx_chk!=0) {
        // chk invalid
        spi_rx_size=0;
      }
    }
  }

  // start timer 2 to call back in 32 us
  TCCR2B = _BV(CS21); // PS: 8
  OCR2A = 64; // 32us later
  GTCCR = _BV(PSRASY); // Reset Prescaler
  TIFR2 = _BV(OCF2A); // Clear OC Bit
  TIMSK2 = _BV(OCIE2A); // Enable Interrupt
}


ISR(TIMER2_COMPA_vect) {
  //stop timer 2
  TCCR2B = 0;
  TIMSK2 = 0;
  
  if (spi_pos<spi_tx_size) {
    // transmit
    spi_tx(spi_buf[spi_pos]);
  } else if (spi_pos==spi_tx_size) {
    spi_tx(spi_tx_chk);
  } else if (spi_pos<spi_size) {
    spi_tx(0xff);
  } else {
    SPCR = _BV(SPE)|_BV(MSTR)|_BV(SPR1);
    set_output_bit(IO_SPI_SS);
    spi_size=0;
  }
  spi_pos++;
}

