/*  BSD-License:

Copyright (c) 2007 by Nils Springob, nicai-systems, Germany

All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

  * Redistributions of source code must retain the above copyright notice,
    this list of conditions and the following disclaimer.
  * Redistributions in binary form must reproduce the above copyright notice,
    this list of conditions and the following disclaimer in the documentation
    and/or other materials provided with the distribution.
  * Neither the name nicai-systems nor the names of its contributors may be
    used to endorse or promote products derived from this software without
    specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

*/

#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <avr/eeprom.h>
#include <inttypes.h>
#include <stdio.h>

const char progkey[] EEMEM = "Test EEPROM";

const uint8_t copro_hex[] PROGMEM =
#include "copro_hex.h"

const uint8_t copro_check_hex[] PROGMEM =
#include "copro_check_hex.h"

#include "nibo/niboconfig.h"

#include "nibo/iodefs.h"
#include "nibo/delay.h"
#include "nibo/display.h"
#include "nibo/gfx.h"
#include "avrisp.h"
#include "nibo/copro.h"
#include "nibo/floor.h"
#include "nibo/bot.h"
#include "nibo/analog.h"
#include "nibo/pwm.h"
#include "nibo/leds.h"
#include "nibo/spi.h"

uint16_t value_off_ll;
uint16_t value_off_rl;
uint16_t value_off_lf;
uint16_t value_off_rf;
uint16_t value_on_ll;
uint16_t value_on_rl;
uint16_t value_on_lf;
uint16_t value_on_rf;

uint16_t old_rc5_cmd;
uint8_t sequence_state;

#define IC_FLASH_SIZE 8192

#define LOBYTE(x)        (uint8_t)((uint16_t)x)
#define HIBYTE(x)        (uint8_t)(((uint16_t)x)>>8)
#define MAKE_WORD(hi,lo) ((hi*0x100)+lo)

void start_rc_mode();

void measure_ground();
uint8_t diff_ground(int16_t val);

void test_sensors();
void test_leds();
void circuit_test();

void print_hexdigit(uint8_t val) {
  if (val<10)
    val+='0';
  else
    val+='a'-10;
  gfx_print_char(val);
}


void print_hex (uint8_t val) {
  char c1=val/16;
  char c2=val%16;
  if (c1<10) c1+='0'; else c1+='a'-10;
  if (c2<10) c2+='0'; else c2+='a'-10;
  gfx_print_char(c1);
  gfx_print_char(c2);
}


void led_red_status(uint8_t val) {
  val=6-val;
  uint8_t port_val=0xff;
  port_val >>= val;
  IO_LEDS_RED_PORT = port_val;
}


void led_green_status(uint8_t val) {
  val=6-val;
  uint8_t port_val=0xff;
  port_val >>= val;
  IO_LEDS_GREEN_PORT = port_val;
}


uint8_t detect_mode() {
  
  // activate pull-up inputs
  deactivate_output_bit(IO_INPUT_2);
  deactivate_output_bit(IO_INPUT_3);
  set_output_bit(IO_INPUT_2);
  set_output_bit(IO_INPUT_3);
  
  delay(1);
  //if ((get_input_bit(IO_INPUT_2)==0)&&(get_input_bit(IO_INPUT_3)==0)) {
    // both pins connected to ground -> circuit test
  //  return 0;
  //}

  char buf[11];
  char key[]   = "Test EEPROM";
  char nokey[] = "***********";
  eeprom_read_block (buf, (const void *) 0, sizeof(buf));
  
  if (strncmp(buf, key, sizeof(buf))==0) {
    eeprom_write_block (nokey, (const void *) 0, sizeof(buf));
    return 0;
  }

  activate_output_bit(IO_INPUT_2);
  clear_output_bit(IO_INPUT_2);
  delay(1);
  if (get_input_bit(IO_INPUT_3)) {
    // jumper open
    return 1;
  }
  // jumper closed
  return 2;
}

void test_leds(void);
void test_bot(void);
void show_error(void);
void ic_init(void);
void copro_spi_init(void);
void show_boot_message_gfx(void);

void copro_check_flash(void);
void copro_program_flash(void);


uint8_t copro_status = 0;



void draw_head() {
  gfx_fill(0x00);
  gfx_move(0, 0);
  gfx_set_proportional(1);
  gfx_draw_mode(GFX_DM_JAM2);
  gfx_print_text("NIBO 2  Lib ");
  gfx_print_text_P(bot_lib_version);
  gfx_print_text("  Rev. ");
  gfx_print_text_P(bot_lib_revision);
  gfx_set_proportional(0);
  gfx_move(0, 9);
  gfx_hline(128);
}

void do_validate() {
  draw_head();
  gfx_move(0, 13);
  gfx_set_proportional(1);
  gfx_draw_mode(GFX_DM_JAM2);
  gfx_print_text("Validating COPRO firmware");
  gfx_set_proportional(0);
  delay(200);
  copro_check_flash();
  if (copro_status!=0x02) {
    show_error();
  }
}


void do_program() {
  draw_head();
  gfx_move(0, 13);
  gfx_set_proportional(1);
  gfx_draw_mode(GFX_DM_JAM2);
  gfx_print_text("Updating COPRO firmware");
  gfx_set_proportional(0);
  IO_LEDS_GREEN_PORT = 0;
  IO_LEDS_RED_PORT = 0x1e;
  delay(200);
  copro_program_flash();


  if (copro_status==0xf0) {
    gfx_move(0, 44);
    gfx_set_proportional(1);
    gfx_draw_mode(GFX_DM_JAM2);
    gfx_print_text("ERROR: Unable to enter");
    gfx_move(0, 55);
    gfx_print_text("COPRO programming mode");
    show_error();
  }
  
  while(1) {
    IO_LEDS_GREEN_PORT = IO_LEDS_GREEN_MASK;
    IO_LEDS_RED_PORT = 0;
    delay(950);
    IO_LEDS_GREEN_PORT = 0;
    delay(50);
  }
}









int main (void) {
  deactivate_output_bit(IO_INPUT_2);
  activate_output_bit(IO_INPUT_3);
  set_output_bit(IO_INPUT_2);
  clear_output_bit(IO_INPUT_3);

  switch (detect_mode()) {
    case 0: circuit_test(); break;
    case 1: test_leds(); break;
    case 2: break;
  }

  sei(); // enable interrupts

  activate_output_bit(IO_LED_RG);
  set_output_bit(IO_LED_RG);

  activate_output_bit(IO_LED_WHITE);
  activate_output_bit(IO_DISP_LIGHT);
  IO_LEDS_RED_DDR = IO_LEDS_RED_MASK;
  IO_LEDS_GREEN_DDR = IO_LEDS_GREEN_MASK;
  IO_LEDS_RED_PORT = 0xff;
  IO_LEDS_GREEN_PORT = 0xff;
  delay(50);

  led_red_status(0);
  led_green_status(0);
  delay(200);

  led_green_status(1);
  delay(200);

  
  if (display_init(NIBO_DISPLAY_TYPE)!=0) {
    set_output_bit(IO_DISP_LIGHT);
    if (display_type==NIBO_DISPLAY_TYPE)
      show_boot_message_gfx();
    delay(500);
    delay(500);
  }
  
  
  led_green_status(2);
  delay(200);

  copro_status = 0;

  ic_init();
  copro_spi_init();

  if (copro_status==1) {
    led_green_status(3);
  } else {
    led_red_status(3);
  }
  delay(200);


  gfx_move(15, 24);
  gfx_print_text("                ");
  gfx_move(15, 32);
  gfx_print_text("                ");
  gfx_move(15, 40);
  gfx_print_text("                ");
  gfx_move(15, 48);
  gfx_print_text("                ");
  gfx_move(15, 56);
  gfx_print_text("                ");


  ic_init();
  delay(100);
  led_green_status(4);

  delay(200);

  gfx_move(15, 24);
  gfx_print_char('0'+copro_status);

  led_green_status(5);
  delay(200);


  if (!get_input_bit(IO_INPUT_1)) {
    do_program();
  } else {
    do_validate();
    delay(200);
    delay(200);
    delay(200);
    delay(200);
    test_sensors();
  }
}


void plot_bar(uint8_t x, uint8_t y, uint8_t h, uint8_t l, uint8_t val) {
  gfx_draw_mode(GFX_DM_JAM2);
  gfx_move(x, y);  gfx_box(val, h);
  gfx_draw_mode(GFX_DM_JAM2_INV);
  gfx_move(x+val, y);  gfx_box(l-val, h);
  gfx_draw_mode(GFX_DM_JAM2);
}

void plot_value(uint8_t x, uint8_t y, uint8_t val) {
  gfx_move(x, y);
  print_hex(val);
  plot_bar(x, y+8, 1, 16, val/16);
}

void plot_logo(uint8_t x, uint8_t y) {
  gfx_move(x+6, y+0);  gfx_box(14, 3);
  gfx_move(x+6, y+5);  gfx_box(14, 6);
  gfx_move(x+0, y+7);  gfx_box(4, 6);
  gfx_move(x+22, y+7);  gfx_box(4, 6);
}

void print_hex16(uint8_t x, uint8_t y, uint16_t val) {
  gfx_move(x, y);
  print_hex(HIBYTE(val));
  print_hex(LOBYTE(val));
}

void test_sensors() {
  IO_LEDS_RED_PORT=0x00;
  IO_LEDS_GREEN_PORT=0x00;

  activate_output_bit(IO_LED_RG);
  set_output_bit(IO_LED_RG);
  activate_output_bit(IO_LED_WHITE);
  activate_output_bit(IO_DISP_LIGHT);
  IO_LEDS_RED_DDR = IO_LEDS_RED_MASK;
  IO_LEDS_GREEN_DDR = IO_LEDS_GREEN_MASK;
  
  spi_init();
  delay(100);

  gfx_fill(0x00);
  gfx_move(0, 0);
  gfx_set_proportional(1);
  gfx_draw_mode(GFX_DM_JAM2);
  gfx_print_text("NIBO 2  Lib ");
  gfx_print_text_P(bot_lib_version);
  gfx_print_text("  Rev. ");
  gfx_print_text_P(bot_lib_revision);

  
  gfx_set_proportional(0);

  gfx_move(0, 9);
  gfx_hline(128);

  plot_logo(52, 26);

  analog_init();

  uint32_t cnt=0;

  copro_setSpeedParameters(5, 6, 7);

  while(1) {
    copro_ir_startMeasure();
    copro_update();

    // Motorwerte
    print_hex16(  0, 13, copro_ticks_r);
    print_hex16(102, 13, copro_ticks_l);
    print_hex16(  0, 23, copro_speed_r);
    print_hex16(102, 23, copro_speed_l);
    print_hex16(  0, 33, copro_current_r);
    print_hex16(102, 33, copro_current_l);
    
    // DISTANCES
    plot_value(  0, 55, HIBYTE(copro_distance[0]));
    plot_value( 29, 55, HIBYTE(copro_distance[1]));
    plot_value( 58, 55, HIBYTE(copro_distance[2]));
    plot_value( 87, 55, HIBYTE(copro_distance[3]));
    plot_value(116, 55, HIBYTE(copro_distance[4]));

    // Bodensensoren
    measure_ground();

    plot_value(  0, 44, diff_ground(value_off_rf-value_on_rf));
    plot_value( 38, 44, diff_ground(value_off_rl-value_on_rl));
    plot_value( 78, 44, diff_ground(value_off_ll-value_on_ll));
    plot_value(116, 44, diff_ground(value_off_lf-value_on_lf));

    gfx_move(40, 13);
    char c='-';
    if ((copro_sysconf[0]==2) && (copro_sysconf[1]==50)) c='+';
    gfx_print_char(c);

    
    // Batterie
    float vbatt;
    if (copro_sysconf[1]==50) {
      vbatt = 0.4 + (analog_getValue(ANALOG_SUPPLY)*(10./1023));
    } else {
      vbatt = 0.8 + (analog_getValue(ANALOG_SUPPLY)*(15.0/1023));
      //vbatt = (analog_getValue(ANALOG_SUPPLY)-24)*0.01734;
    }
    char buffer[20];
    sprintf(buffer, "%4.1fV", vbatt);
    gfx_move(50, 13);
    gfx_print_text(buffer);
    delay(200);

    cnt++;

    switch (cnt) {
//       case 20: copro_setSpeed(20, 20); break;
//       case 30: copro_setSpeed(0, 0); break;
//       case 40: copro_setSpeed(-20, -20); break;
//       case 50: copro_setSpeed(0, 0); break;
       case 20: copro_setTargetAbs(145, 145, 30); break; // 50cm
       case 40: copro_setTargetAbs(  0,   0, 30); break;
       case 59: cnt=19; break;
//       case 59: copro_setTargetRel(27, -27, 10); break;
//       case 99: cnt=19; break;
    }
    
  }
}


void ic_init() {
  activate_output_bit(IO_RESET_CO);
  clear_output_bit(IO_RESET_CO);
  delay(10);
  set_output_bit(IO_RESET_CO);
  delay(20);
}


uint8_t ic_check_flash(uint32_t ptr) {
  uint8_t result=0;

  gfx_draw_mode(GFX_DM_JAM2);
  gfx_move(0, 24);
  gfx_set_proportional(1);
  gfx_print_text("check");
  gfx_set_proportional(0);
  
  IO_LEDS_GREEN_PORT = 0x1e;
  IO_LEDS_RED_PORT = 0x00;
  
  ic_init();
  enterProgrammingMode(_BV(IO_RESET_CO_BIT));
  setAddress(0);
  for (uint16_t addr=0; addr<IC_FLASH_SIZE; addr+=2) {
    uint16_t data1 = readProgramMemory();
    uint16_t data2 = pgm_read_word_far(ptr+addr);
    if (data1!=data2) {
      if (result!=1) {
        result=1;
        gfx_draw_mode(GFX_DM_JAM2);
        gfx_move(0, 44);
        gfx_set_proportional(1);
        gfx_print_text("first diff addr=");
        gfx_set_proportional(0);
        print_hexdigit((addr>>9)&0x0f);
        print_hexdigit((addr>>5)&0x0f);
        print_hexdigit((addr>>1)&0x0f);
        gfx_move(0, 54);
        gfx_set_proportional(1);
        gfx_print_text("r=");
        gfx_set_proportional(0);
        print_hexdigit((data1>>12)&0x0f);
        print_hexdigit((data1>>8)&0x0f);
        print_hexdigit((data1>>4)&0x0f);
        print_hexdigit(data1&0x0f);
        gfx_set_proportional(1);
        gfx_print_text(" w=");
        gfx_set_proportional(0);
        print_hexdigit((data2>>12)&0x0f);
        print_hexdigit((data2>>8)&0x0f);
        print_hexdigit((data2>>4)&0x0f);
        print_hexdigit(data2&0x0f);
        //delay(1000);
      }
    }
    if ((addr&0x003f)==0x00) {
      gfx_move(30, 24);
      gfx_set_proportional(1);
      gfx_print_text("addr=");
      gfx_set_proportional(0);
      print_hexdigit((addr>>9)&0x0f);
      print_hexdigit((addr>>5)&0x0f);
      print_hexdigit((addr>>1)&0x0f);

      plot_bar(0, 35, 4, 128, addr/(IC_FLASH_SIZE/128));
    }
  
    if ((addr&0x00ff)==0x00) {
      IO_LEDS_GREEN_PORT = 0x1e;
      IO_LEDS_RED_PORT = 0x00;
    }
    if ((addr&0x00ff)==0x80) {
      IO_LEDS_GREEN_PORT = 0x00;
      IO_LEDS_RED_PORT = 0x00;
    }

  }

  if (result!=1) {
    gfx_move(0, 44);
    gfx_set_proportional(1);
    gfx_print_text("Validation successful!");
    gfx_move(0, 54);
    gfx_set_proportional(1);
    gfx_print_text("Please wait...");
  }
  
  leaveProgrammingMode(_BV(IO_RESET_CO_BIT));
  delay(100);
  return result;
}


//
// IC3 - Motco
//

void copro_spi_init(void) {
  if (enterProgrammingMode(_BV(IO_RESET_CO_BIT))) {
    copro_status=1;
  }
  leaveProgrammingMode(_BV(IO_RESET_CO_BIT));
  delay(100);
}


void copro_check_flash(void) {
  if (ic_check_flash((uint32_t) copro_hex)) {
    copro_status=1;
  } else {
    copro_status=2;
  }
}


void copro_program_flash(void) {
  ic_init();
  if (enterProgrammingMode(_BV(IO_RESET_CO_BIT))==0) {
    // error COPRO not responding!
    leaveProgrammingMode(_BV(IO_RESET_CO_BIT));
    delay(100);
    copro_status=0xf0;
    return;
  }
  chipErase();
  writeLoFuse(0xf7);
  setAddress(0);
  uint32_t ptr = (uint32_t) copro_hex;
  for (uint16_t addr=0; addr<sizeof(copro_hex); addr+=2) {
    uint16_t data = pgm_read_word_far(ptr+addr);
    writeProgramMemory(data);
    if (((addr>>1)&0x1f)==0x1f) {
      issuePageWrite();
    }
    if ((addr&0x003f)==0x00) {
      gfx_move(30, 24);
      gfx_set_proportional(1);
      gfx_print_text("addr=");
      gfx_set_proportional(0);
      print_hexdigit((addr>>9)&0x0f);
      print_hexdigit((addr>>5)&0x0f);
      print_hexdigit((addr>>1)&0x0f);
      plot_bar(0, 35, 4, 128, addr/(IC_FLASH_SIZE/128));
    }
  }
  gfx_move(0, 44);
  gfx_set_proportional(1);
  gfx_print_text("Update finished.");
  gfx_move(0, 54);
  gfx_set_proportional(1);
  gfx_print_text("Please reboot!");

  copro_status=2;
  leaveProgrammingMode(_BV(IO_RESET_CO_BIT));
  delay(100);
}

void copro_program_flash2(void) {
  ic_init();
  if (enterProgrammingMode(_BV(IO_RESET_CO_BIT))==0) {
    // error COPRO not responding!
    leaveProgrammingMode(_BV(IO_RESET_CO_BIT));
    delay(100);
    copro_status=0xf0;
    return;
  }
  chipErase();
  writeLoFuse(0xf7);
  
  writeHiFuse(0xd7); // new boards >V2.50 will store Version in EEPROM
  setAddress(128);
  writeDataMemory(2);  // NIBO 2
  writeDataMemory(50); // Board Rev. 50
  writeDataMemory(0);
  writeDataMemory(0);
  
  setAddress(0);
  uint32_t ptr = (uint32_t) copro_check_hex;
  for (uint16_t addr=0; addr<sizeof(copro_check_hex); addr+=2) {
    uint16_t data = pgm_read_word_far(ptr+addr);
    writeProgramMemory(data);
    if (((addr>>1)&0x1f)==0x1f) {
      issuePageWrite();
    }
    if ((addr&0x003f)==0x00) {
      gfx_move(30, 24);
      gfx_set_proportional(1);
      gfx_print_text("addr=");
      gfx_set_proportional(0);
      print_hexdigit((addr>>9)&0x0f);
      print_hexdigit((addr>>5)&0x0f);
      print_hexdigit((addr>>1)&0x0f);
      plot_bar(0, 35, 4, 128, addr/(IC_FLASH_SIZE/128));
    }
  }
  gfx_move(0, 44);
  gfx_set_proportional(1);
  gfx_print_text("Update finished.");
  gfx_move(0, 54);
  gfx_set_proportional(1);
  gfx_print_text("Please reboot!");

  copro_status=2;
  leaveProgrammingMode(_BV(IO_RESET_CO_BIT));
  delay(100);
}



void show_boot_message_gfx(void) {
  display_init(NIBO_DISPLAY_TYPE);
  display_wait(0x00);
  display_write(0x00, 0x3f);
  display_wait(0x00);
  display_write(0x00, 0xc0);
  display_wait(0x00);
  display_write(0x00, 0x40);
  display_wait(0x00);
  display_write(0x00, 0xb8);

  display_wait(0x01);
  display_write(0x01, 0x3f);
  display_wait(0x01);
  display_write(0x01, 0xc0);
  display_wait(0x01);
  display_write(0x01, 0x40);
  display_wait(0x01);
  display_write(0x01, 0xb8);

  gfx_fill(0x00);

  gfx_move(0, 0);
  gfx_set_proportional(1);
  gfx_print_text("NIBO 2  Lib ");
  gfx_print_text_P(bot_lib_version);
  gfx_print_text("  Rev. ");
  gfx_print_text_P(bot_lib_revision);
  gfx_move(5, 12);
  gfx_print_text("(c) 2011 by nicai-systems");

  gfx_set_proportional(0);

  gfx_move(15, 24);
  gfx_print_text("0123456789:;<=>?");
  gfx_move(15, 32);
  gfx_print_text("@ABCDEFGHIJKLMNO");
  gfx_move(15, 40);
  gfx_print_text("PQRSTUVWXYZ[\\]^_");
  gfx_move(15, 48);
  gfx_print_text("`abcdefghijklmno");
  gfx_move(15, 56);
  gfx_print_text("pqrstuvwxyz{|}~");
  gfx_print_char(127);
  
}


void test_leds(void) {
  activate_output_bit(IO_LED_RG);
  set_output_bit(IO_LED_RG);

  activate_output_bit(IO_LED_WHITE);
  activate_output_bit(IO_DISP_LIGHT);
  IO_LEDS_RED_DDR = IO_LEDS_RED_MASK;
  IO_LEDS_GREEN_DDR = IO_LEDS_GREEN_MASK;

  IO_LEDS_GREEN_PORT=0;
  IO_LEDS_RED_PORT=0;
  delay(1000);
    
  while(1) {
    for (uint8_t i=0; i<8; ++i) {
      IO_LEDS_RED_PORT = _BV(i);
      delay(1000);
      IO_LEDS_RED_PORT = 0;
      IO_LEDS_GREEN_PORT = _BV(i);
      delay(1000);
      IO_LEDS_GREEN_PORT = 0;
    }
    set_output_bit(IO_LED_WHITE);
    delay(1000);
    clear_output_bit(IO_LED_WHITE);
    delay(1000);
    set_output_bit(IO_DISP_LIGHT);
    delay(1000);
    clear_output_bit(IO_DISP_LIGHT);
    delay(1000);
    
    IO_LEDS_RED_PORT = IO_LEDS_RED_MASK;
    IO_LEDS_GREEN_PORT = IO_LEDS_GREEN_MASK;
    set_output_bit(IO_LED_WHITE);
    set_output_bit(IO_DISP_LIGHT);
    delay(2000);

    IO_LEDS_RED_PORT = 0;
    IO_LEDS_GREEN_PORT = 0;
    clear_output_bit(IO_LED_WHITE);
    clear_output_bit(IO_DISP_LIGHT);
    delay(1000);
  }
}

void show_error(void) {
  activate_output_bit(IO_LED_WHITE);
  activate_output_bit(IO_DISP_LIGHT);
  IO_LEDS_RED_DDR = IO_LEDS_RED_MASK;
  IO_LEDS_GREEN_DDR = IO_LEDS_GREEN_MASK;

  IO_LEDS_GREEN_PORT=0;

  while (1) {
    delay(50);
    IO_LEDS_RED_PORT = IO_LEDS_RED_MASK;
    delay(50);
    IO_LEDS_RED_PORT = 0;
    delay(50);
    IO_LEDS_RED_PORT = IO_LEDS_RED_MASK;
    delay(50);
    IO_LEDS_RED_PORT = 0;
    delay(50);
    IO_LEDS_RED_PORT = IO_LEDS_RED_MASK;
    delay(50);
    IO_LEDS_RED_PORT = 0;
    delay(50);
    IO_LEDS_RED_PORT = IO_LEDS_RED_MASK;
    delay(50);
    IO_LEDS_RED_PORT = 0;
    delay(600);
  }

}

void measure_ground() {
    delay_us(25);
    value_off_ll = 2*analog_getValue(ANALOG_LINE_L0);
    value_off_rl = 2*analog_getValue(ANALOG_LINE_R0);
    value_off_lf = 2*analog_getValue(ANALOG_FLOOR_L0);
    value_off_rf = 2*analog_getValue(ANALOG_FLOOR_R0);
    value_on_ll = 2*analog_getValue(ANALOG_LINE_L1);
    value_on_rl = 2*analog_getValue(ANALOG_LINE_R1);
    value_on_lf = 2*analog_getValue(ANALOG_FLOOR_L1);
    value_on_rf = 2*analog_getValue(ANALOG_FLOOR_R1);
}

uint8_t check_ground(int16_t val) {
  return val<200;
}

uint8_t diff_ground(int16_t val) {
  if(val>255)
    val=255;
  if (val<0)
    val=0;
  return val;
}

void circuit_test() {
  activate_output_bit(IO_LED_RG);
  set_output_bit(IO_LED_RG);
  IO_LEDS_RED_DDR = IO_LEDS_RED_MASK;
  IO_LEDS_GREEN_DDR = IO_LEDS_GREEN_MASK;
  copro_program_flash2();
  
  // signal COPRO check start (all ISP lines low):
  clear_output_bit(IO_ISP_SCK);
  clear_output_bit(IO_ISP_MOSI);
  clear_output_bit(IO_ISP_MISO);
  clear_output_bit(IO_ISP_SS);
  
  activate_output_bit(IO_ISP_SCK);
  activate_output_bit(IO_ISP_MOSI);
  activate_output_bit(IO_ISP_MISO);
  activate_output_bit(IO_ISP_SS);

  while(1) {
    IO_LEDS_RED_PORT = IO_LEDS_RED_MASK;
    IO_LEDS_GREEN_PORT = IO_LEDS_GREEN_MASK;
    delay_ms(100);
    IO_LEDS_RED_PORT = 0;
    IO_LEDS_GREEN_PORT = 0;
    delay_ms(100);
  }
}


