/*  BSD-License:

Copyright (c) 2007 by Nils Springob, nicai-systems, Germany

All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

  * Redistributions of source code must retain the above copyright notice,
    this list of conditions and the following disclaimer.
  * Redistributions in binary form must reproduce the above copyright notice,
    this list of conditions and the following disclaimer in the documentation
    and/or other materials provided with the distribution.
  * Neither the name nicai-systems nor the names of its contributors may be
    used to endorse or promote products derived from this software without
    specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

*/

#include "iodefs.h"
#include "analog.h"
#include "persistent.h"
#include <util/delay.h>
#include <util/atomic.h>
#include <avr/interrupt.h>

#define ANALOG_MASK 0x8f

// usage of the 8 bits for idx:  --SSCCCD
// S: Sample number: 0,1,2,3
// C: Channel: 0,1,2,3,4,5
// D: Drop: 0,1
uint8_t idx_act;
uint8_t idx_new;


uint16_t analog_raw[8];

uint16_t analog_data[8];


inline static void select_mux(uint8_t idx) {
  switch (idx) {
    case  0: ADMUX=0x00; break; // single, ADC0 @5V
    case  2: ADMUX=0x01; break; // single, ADC1 @5V
    case  4: ADMUX=0x02; break; // single, ADC2 @5V
    case  6: ADMUX=0x03; break; // single, ADC3 @5V
    case  8: ADMUX=0x07; break; // single, ADC7 @5V
    case 10: ADMUX=0xa2; break; // single, Temperature @1.1V
    //case 12: ADMUX=0a2; break; // single, Temperature @1.1V
    //case 14: ADMUX=0x2; break; // single, Temperature @1.1V
  }
}

void analog_init() {
  PORTA &= ~ANALOG_MASK;
  idx_act=0;
  idx_new=1;
  select_mux(idx_act);
  // PS=128 -> ADCCLK=62.5kHz@8Mhz -> 4464 samples/sec -> 744 samples/sec/channel
  // Enable, start free-running, irq mode
  ADCSRA = _BV(ADEN) | _BV(ADSC) | _BV(ADATE) | _BV(ADIF) | _BV(ADIE) | _BV(ADPS2) | _BV(ADPS1) | _BV(ADPS0);
}


ISR(ADC_vect) {
  uint16_t val = ADC;
  // store Result, drop first result...
  // combine 4 raw samples to one processed sample
  if (idx_act&0x01) {
    uint8_t ch = (idx_act>>1)&0x07;
    if ((idx_act & 0xf0)==0x30) {
      val += analog_raw[ch];
      analog_raw[ch] = 0;
      analog_data[ch] = val/4;
    } else {
      analog_raw[ch] += val;
    }
  }

  idx_act = idx_new;
  idx_new++;
  if ((idx_new&0x0f)==12) {
    // skip 12,13,14,15
    idx_new += 4;
  }
  if (idx_new>=64) {
    idx_new=0;
  }

  select_mux(idx_new);
}

int16_t analog_get_distance() {
  int16_t v0 = persistent.dist_v0;
  int16_t d0 = persistent.dist_d0;
  int16_t alpha = persistent.dist_alpha;

  int16_t val = analog_get_value(4);
  if (val>v0) {
    val -= v0;
  } else {
    val = 1;
  }
  int16_t d = alpha / val;
  d -= d0;
  if (d<0) {
    return 0;
  } else if (d>255) {
    return 255;
  } else {
    return d;
  }
}

int16_t analog_get_value(uint8_t idx) {
  int16_t val;
  ATOMIC_BLOCK(ATOMIC_RESTORESTATE) {
    val = analog_data[idx];
  }
  return val;
}


