/*  BSD-License:

Copyright (c) 2007 by Nils Springob, nicai-systems, Germany

All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

  * Redistributions of source code must retain the above copyright notice,
    this list of conditions and the following disclaimer.
  * Redistributions in binary form must reproduce the above copyright notice,
    this list of conditions and the following disclaimer in the documentation
    and/or other materials provided with the distribution.
  * Neither the name nicai-systems nor the names of its contributors may be
    used to endorse or promote products derived from this software without
    specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

*/

#include "iodefs.h"
#include "i2c.h"
#include "comm.h"
#include "transaction.h"
#include "servo.h"
#include "analog.h"
#include "persistent.h"
#include <avr/interrupt.h>

#define LOBYTE(x)        (uint8_t)((uint16_t)x)
#define HIBYTE(x)        (uint8_t)(((uint16_t)x)>>8)
#define MAKE_WORD(hi,lo) (((uint8_t)hi*(uint16_t)0x100)+(uint8_t)lo)


#define CHECK_SIZE(s) if (size!=(s+1)) return 0

#define SET_MASKED(reg, val, mask) reg = (reg & ~((uint8_t)(mask))) | (((uint8_t)(val)) & ((uint8_t)(mask)))

//uint8_t global_mode  = MODE_MOVE;
uint8_t global_mode   = MODE_MEASURE;
uint8_t global_pos    = 180;
uint8_t global_target = 0;
uint8_t global_step   = 5;

uint8_t global_distances[181];

int16_t global_actdist;
int16_t global_adcval[7];
int16_t global_temperature;


enum {
  CMD_RESET = 0x00,
  CMD_MOVE = 0x01,
  CMD_MEASURE = 0x02,
  CMD_READ = 0x03,
  CMD_GETPOS = 0x04,
  CMD_GETDIST = 0x05,
  CMD_GETBUSY = 0x06,
  CMD_SETDDR = 0x10,
  CMD_SETPORT = 0x11,
  CMD_GETPIN = 0x12,
  CMD_GETANALOG = 0x13,
  CMD_MOVESERVO = 0x20,
  CMD_GETSERVOCALIB = 0x21,
  CMD_SETSERVOCALIB = 0x22,
  CMD_GETDISTCALIB = 0x23,
  CMD_SETDISTCALIB = 0x24
};


static inline uint8_t do_reset(uint8_t size) {
  CHECK_SIZE(0);
  global_mode = MODE_IDLE;
  return 0;
}

static inline uint8_t do_move(uint8_t size) {
  CHECK_SIZE(1);
  global_mode = MODE_MOVE;
  global_target = comm_buf[1];
  return 0;
}

static inline uint8_t do_measure(uint8_t size) {
  CHECK_SIZE(2);
  global_mode = MODE_MEASURE;
  global_target = comm_buf[1];
  global_step = comm_buf[2];
  return 0;
}

static inline uint8_t do_read(uint8_t size) {
  CHECK_SIZE(2);
  uint8_t pos = comm_buf[1];
  uint8_t step = comm_buf[2];
  uint8_t idx = 1;
  
  while ((pos<=180) && (idx<17)) {
    comm_buf[idx] = global_distances[pos];
    pos += step;
    idx++;
  }
  
  return 16;
}

static inline uint8_t do_getpos(uint8_t size) {
  CHECK_SIZE(0);
  comm_buf[1] = global_pos;
  return 1;
}

static inline uint8_t do_getdist(uint8_t size) {
  CHECK_SIZE(0);
  comm_buf[1] = global_actdist;
  return 1;
}

static inline uint8_t do_getbusy(uint8_t size) {
  CHECK_SIZE(0);
  comm_buf[1] = global_mode;
  return 1;
}

static inline uint8_t do_setddr(uint8_t size) {
  CHECK_SIZE(1);
  uint8_t val = comm_buf[1];
  SET_MASKED(DDRA, val, 0x0f);
  val >>= 4;
  SET_MASKED(DDRB, val, 0x05);
  return 0;
}

static inline uint8_t do_setport(uint8_t size) {
  CHECK_SIZE(1);
  uint8_t val = comm_buf[1];
  SET_MASKED(PORTA, val, 0x0f);
  val >>= 4;
  SET_MASKED(PORTB, val, 0x05);
  return 0;
}

static inline uint8_t do_getpin(uint8_t size) {
  CHECK_SIZE(0);
  uint8_t val = PINB & 0x05;
  val <<= 4;
  val |= PINA & 0x0f;
  comm_buf[1] = val;
  return 1;
}

static inline uint8_t do_getanalog(uint8_t size) {
  CHECK_SIZE(1);
  uint8_t idx = comm_buf[1];
  uint16_t val = 0;
  if (idx<6) {
    val = analog_get_value(idx);
  }
  comm_buf[1] = HIBYTE(val);
  comm_buf[2] = LOBYTE(val);
  return 2;
}

static inline uint8_t do_moveservo(uint8_t size) {
  CHECK_SIZE(2);
  global_mode = MODE_MOVESERVO;
  servo_set_irq(MAKE_WORD(comm_buf[1], comm_buf[2]));
  return 0;
}

static inline uint8_t do_getservocalib(uint8_t size) {
  CHECK_SIZE(0);
  comm_buf[1] = HIBYTE(persistent.servo_lo);
  comm_buf[2] = LOBYTE(persistent.servo_lo);
  comm_buf[3] = HIBYTE(persistent.servo_mid);
  comm_buf[4] = LOBYTE(persistent.servo_mid);
  comm_buf[5] = HIBYTE(persistent.servo_hi);
  comm_buf[6] = LOBYTE(persistent.servo_hi);
  return 6;
}

static inline uint8_t do_setservocalib(uint8_t size) {
  CHECK_SIZE(6);
  persistent.servo_lo = MAKE_WORD(comm_buf[1], comm_buf[2]);
  persistent.servo_mid = MAKE_WORD(comm_buf[3], comm_buf[4]);
  persistent.servo_hi = MAKE_WORD(comm_buf[5], comm_buf[6]);
  persistent_dirty = 1;
  return 0;
}

static inline uint8_t do_getdistcalib(uint8_t size) {
  CHECK_SIZE(0);
  comm_buf[1] = HIBYTE(persistent.dist_v0);
  comm_buf[2] = LOBYTE(persistent.dist_v0);
  comm_buf[3] = HIBYTE(persistent.dist_d0);
  comm_buf[4] = LOBYTE(persistent.dist_d0);
  comm_buf[5] = HIBYTE(persistent.dist_alpha);
  comm_buf[6] = LOBYTE(persistent.dist_alpha);
  return 6;
}

static inline uint8_t do_setdistcalib(uint8_t size) {
  CHECK_SIZE(6);
  persistent.dist_v0 = MAKE_WORD(comm_buf[1], comm_buf[2]);
  persistent.dist_d0 = MAKE_WORD(comm_buf[3], comm_buf[4]);
  persistent.dist_alpha = MAKE_WORD(comm_buf[5], comm_buf[6]);
  persistent_dirty = 1;
  return 0;
}


uint8_t comm_onTransaction(uint8_t size) {
  switch (comm_buf[0]) {
    case CMD_RESET:         return do_reset(size);
    case CMD_MOVE:          return do_move(size);
    case CMD_MEASURE:       return do_measure(size);
    case CMD_READ:          return do_read(size);
    case CMD_GETPOS:        return do_getpos(size);
    case CMD_GETDIST:       return do_getdist(size);
    case CMD_GETBUSY:       return do_getbusy(size);
    case CMD_SETDDR:        return do_setddr(size);
    case CMD_SETPORT:       return do_setport(size);
    case CMD_GETPIN:        return do_getpin(size);
    case CMD_GETANALOG:     return do_getanalog(size);
    case CMD_MOVESERVO:     return do_moveservo(size);
    case CMD_GETSERVOCALIB: return do_getservocalib(size);
    case CMD_SETSERVOCALIB: return do_setservocalib(size);
    case CMD_GETDISTCALIB:  return do_getdistcalib(size);
    case CMD_SETDISTCALIB:  return do_setdistcalib(size);
  }
  return 0;
}
