#include <avr/interrupt.h>
#include <avr/pgmspace.h>

#include <nibobee/iodefs.h>
#include <nibobee/motpwm.h>
#include <nibobee/motpid.h>
#include <nibobee/delay.h>
#include <nibobee/sens.h>
#include <nibobee/usart.h>
#include <nibobee/led.h>
#include <nibobee/line.h>
#include <nibobee/odometry.h>
#include <nibobee/base.h>
#include <nibobee/clock.h>
#include "niboprot.h"
#include "sensoric.h"

static char buffer[256];
static uint16_t pos;
static uint16_t len;
static uint8_t mode;
static uint8_t motor_mode;
static uint8_t colli_mode;
static uint8_t echo;

#define READ_REG(x) nsp_register_flags[x]&(NSP_FLAG_REQUESTED|NSP_FLAG_REPORTED)
#define WRITE_REG(x) nsp_register_flags[x]&NSP_FLAG_MODIFIED_REMOTE

#define BLACK_MAX 10


static void handle_colli() {
  
  if (clock_get_timer()==0) { 
    switch (colli_mode) {
      case 0: {
        if (sensoric_floor_alert) {
          // start colli
          colli_mode = 1;
          motpid_stop(1);
          odometry_reset();
	  clock_set_timer(100);
        }
      } break;
      
      case 1: {
	if (sensoric_floor_alert<0) {
          motpid_setSpeed(10, -10);
	} else {
	  motpid_setSpeed(-10, 10);
	}
        clock_set_timer(2);
        colli_mode = 2;
      } break;

      case 2: {
        int16_t odoD = odometry_getRight(0)-odometry_getLeft(0);
	if ((odoD>40)||(odoD<-40)) {
          motpid_stop(1);
	  clock_set_timer(200);
          sensoric_floor_alert = 0;
          colli_mode = 3;
	} else {
          clock_set_timer(2);
	}
      } break;

      case 3: {
        // get back to remote control...
	nsp_registers[NSPREG_MOTL] = 0;
	nsp_registers[NSPREG_MOTR] = 0;
        nsp_register_flags[NSPREG_MOTMODE] |= NSP_FLAG_MODIFIED_REMOTE;
        colli_mode = 0;
      } break;
	
    }
  }
}


static void update() {
  
  if (READ_REG(NSPREG_BOTID)) {
    //nsp_registers[NSPREG_BOTID] = 0x4e32;   // N2 NIBO 2
    nsp_registers[NSPREG_BOTID] = 0x4e62; // Nb NIBObee
  }

  if (READ_REG(NSPREG_VERSION)) {
    nsp_registers[NSPREG_VERSION] = 2;
  }
  
  if (READ_REG(NSPREG_VSUPPLY)) {
    nsp_registers[NSPREG_VSUPPLY] = 4800;//nibobee_getMillivolt ();
  }
  
  if (WRITE_REG(NSPREG_LEDS)) {
    uint8_t val = nsp_registers[NSPREG_LEDS];
    led_set(LED_L_YE, val&0x01);
    led_set(LED_L_RD, val&0x02);
    led_set(LED_R_RD, val&0x04);
    led_set(LED_R_YE, val&0x08);
  }
  
  if (READ_REG(NSPREG_SENS)) {
    uint8_t val=0;
    int8_t l = sens_getLeft ();
    int8_t r = sens_getRight ();
    if (sensoric_floor_alert < 0) {
      l = -1;
    }
    if (sensoric_floor_alert > 0) {
      r = -1;
    }
    
    if (l<0) val |= 0x01;
    if (l>0) val |= 0x02;
    if (r<0) val |= 0x10;
    if (r>0) val |= 0x20;
    nsp_registers[NSPREG_SENS] = val;
  }
  
  if (colli_mode==0) {
    if (WRITE_REG(NSPREG_MOTL) || WRITE_REG(NSPREG_MOTR) || WRITE_REG(NSPREG_MOTMODE) ) {
      uint8_t mode = nsp_registers[NSPREG_MOTMODE];
      if (mode != motor_mode) {
        if (mode==2) {
          motpid_release();
        } else {
          motpid_init();
        }
      }
      int16_t l = nsp_registers[NSPREG_MOTL];
      int16_t r = nsp_registers[NSPREG_MOTR];
      int16_t s = (l+r)/2;
      int16_t d = (l-r)/2;
      
      if (d<-256) {
	d += 96;
      } else if (d>256) {
	d -= 96;
      } else if (d<-128) {
	d /= 2;
	d += 32;
      } else if (d>128) {
	d /= 2;
	d -= 32;
      } else {
	d /= 4;
      }
      
      if (s<0) {
        s=0;
      }
      
      l = s + d;
      r = s - d;
      
    
      switch (nsp_registers[NSPREG_MOTMODE]) {
        case 0: motpid_stop(1); break;
        case 1: motpid_stop(0); break;
        case 2: motpwm_setLeft (l); 
                motpwm_setRight(r); break;
        case 3: motpid_setSpeed(l, r); break;
      } 
    }
  }

  if (READ_REG(NSPREG_ODOL)) {
    nsp_registers[NSPREG_ODOL] = odometry_getLeft(0);
  }

  if (READ_REG(NSPREG_ODOR)) {
    nsp_registers[NSPREG_ODOR] = odometry_getRight(0);
  }

  
  if (READ_REG(NSPREG_LINEC)) {
    nsp_registers[NSPREG_LINEC] = line_get(LINE_C);
  }
  
  if (READ_REG(NSPREG_LINEL)) {
    nsp_registers[NSPREG_LINEL] = line_get(LINE_L);
  }

  if (READ_REG(NSPREG_LINER)) {
    nsp_registers[NSPREG_LINER] = line_get(LINE_R);
  }
  
  
}


static void send_help() {
  PGM_P text = PSTR(NSP_HELP_TEXT NSP_REGISTER_TEXT);
  char c=1;
  while (c) {
    if (!usart_txfull()) {
      c=pgm_read_byte(text++);
      usart_putchar(c);
    }
  }
  
}

int main_remote() {
  sei(); // enable interrupts
  led_init();
  line_init();
  sens_init();
  odometry_init();
  usart_setbaudrate(9600);
  usart_enable();

#if 0 
  usart_putchar('A');
  usart_putchar('T');
  usart_putchar('+');
  usart_putchar('N');
  usart_putchar('A');
  usart_putchar('M');
  usart_putchar('E');

  usart_putchar('B');
  usart_putchar('L');
  usart_putchar('U');
  usart_putchar('E');
  usart_putchar('2');
  delay(1000);
  delay(1000);
#endif 
  
  mode = 1;
  pos=0;

  
  motpwm_init();
  sensoric_init(SENSORIC_MODE_FLOOR);
  motor_mode=99;
  delay(100);
  
  while (1) {
    handle_colli();
    if (mode==1) {
      // receive request
      if (!usart_rxempty()) {
        char c = usart_getchar();
        if (pos==0) {
          echo = (c!='$');
        }
        if (echo) {
          usart_putchar(c);
        }
        if (c!='\r') {
          buffer[pos++] = c;
        }
        if (c=='\n') {
          nsp_request(buffer, pos-1);
          update();
          if (nsp_show_help()) {
            send_help();
          }
          // generate reply
          len = nsp_reply(buffer, sizeof(buffer));
          pos = 0;
          mode = 2;
        }
      }
    }
    if (mode==2) {
      // send reply
      if (pos<len) {
        if (!usart_txfull()) {
          usart_putchar(buffer[pos++]);
        }
      } else {
        pos = 0;
        len = 0;
        mode = 1;
      }
    }
  }
  
  return 0;
}
