#include <nibobee/iodefs.h>
#include <nibobee/motpwm.h>
#include <nibobee/motpid.h>
#include <nibobee/analog.h>
#include <nibobee/led.h>
#include <nibobee/line.h>
#include <nibobee/delay.h>
#include <nibobee/clock.h>
#include <nibobee/odometry.h>
#include <nibobee/sens.h>
#include <nibobee/usart.h>


#include <stdio.h>
#include <stdint.h>
#include <stdbool.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include <util/delay.h>

#include "sensoric.h"
#include "util.h"

#define MS(x) ((x)/2)

#define BLACK_MAX 10

static void clock_delay_n(uint16_t n) {
  while (n--) {
    clock_wait_timer();
  }
}

// special function
void modpid_load_pwm(int16_t l, int16_t r);

enum {
  SENS_COLL_L,
  SENS_COLL_R,
  SENS_LINE_L,
  SENS_LINE_C,
  SENS_LINE_R,
  SENS_BKITXS_L,
  SENS_BKITXS_R,
  SENS_CNT,
  // symbolic values:
  SENS_COLL,
  SENS_LINE,
  SENS_BKITXS,
  SENS_ALL
};


typedef struct {
  int16_t val;
  int16_t min;
  int16_t max;
} sensor_t;

sensor_t sensors[SENS_CNT];


uint8_t mode;
uint16_t counter;
int16_t lval, rval;
int8_t last_coll;

uint16_t logcnt;

enum {
  MODE_HALT,
  MODE_GO,
  MODE_GO_L,
  MODE_GO_R,
  MODE_SLOW,
  MODE_SLOW_L,
  MODE_SLOW_R,
  MODE_BACK,
  MODE_BACK_L,
  MODE_BACK_R,
  MODE_BACK_HALT
};


// Targets
int16_t target_l;
int16_t target_r;
int16_t target_d;

// Sequencer
// Halt: 0x0, forwards: 0xF, backwards: 0xB
uint8_t pico_seq;


void sensor_update() {
  // update BKit-XS sensors   
  sensors[SENS_BKITXS_L].val = sensoric_getBKitXS_L();
  sensors[SENS_BKITXS_R].val = sensoric_getBKitXS_R();
  
  // update Collision sensor
  int16_t bkxs_l = sensors[SENS_BKITXS_L].val;
  int8_t sl = sens_getLeft();
  if (sl) {
    sensors[SENS_COLL_L].val = -4*sl;
  } else if (bkxs_l>8*6) {
    sensors[SENS_COLL_L].val = 3;
  } else if (bkxs_l>8*4) {
    sensors[SENS_COLL_L].val = 2;
  } else if (bkxs_l>8*2) {
    sensors[SENS_COLL_L].val = 1;
  } else {
    sensors[SENS_COLL_L].val = 0;
  }
    
  int16_t bkxs_r = sensors[SENS_BKITXS_R].val;
  int8_t sr = sens_getRight();
  if (sr) {
    sensors[SENS_COLL_R].val = -4*sr;
  } else if (bkxs_r>8*6) {
    sensors[SENS_COLL_R].val = 3;
  } else if (bkxs_r>8*4) {
    sensors[SENS_COLL_R].val = 2;
  } else if (bkxs_r>8*2) {
    sensors[SENS_COLL_R].val = 1;
  } else {
    sensors[SENS_COLL_R].val = 0;
  }

  // update Floor sensor
  sensors[SENS_LINE_L].val = (3 * sensors[SENS_LINE_L].val + line_get(LINE_L) + 3)/4;
  sensors[SENS_LINE_C].val = (3 * sensors[SENS_LINE_C].val + line_get(LINE_C) + 3)/4;
  sensors[SENS_LINE_R].val = (3 * sensors[SENS_LINE_R].val + line_get(LINE_R) + 3)/4;
  
  static uint8_t floor_cnt = 0;
  
  if (sensoric_floor_cnt!=floor_cnt) {
    floor_cnt = sensoric_floor_cnt;
    log_hex8(sensoric_floor_cnt);   
    log_str(":");
    log_shex16(sensoric_floor_delta_t);   
    log_str("\n");
  }
#if 1
  if (logcnt) {
    logcnt--;
  } else {        
    /*
    //log_hex8(sensors[SENS_LINE_L].val/16);
    //log_str(":");
    log_hex4(sensoric_floor_l);
    //log_str(" ");
    //log_hex8(sensors[SENS_LINE_C].val/16);
    //log_str(":");
    log_hex4(sensoric_floor_c);
    //log_str(" ");
    //log_hex8(sensors[SENS_LINE_R].val/16);
    //log_str(":");
    log_hex4(sensoric_floor_r);
*/
    
    
    //log_str(" ");
    log_hex16(sensors[SENS_BKITXS_L].val);
    log_str(":");
    log_hex16(sensors[SENS_BKITXS_R].val);
    
    /*
    
    log_str(" ");
    log_hex4(sensoric_floor_class&0x01);
    log_str(":");
    log_shex16(sensoric_floor_delta_t);
    */
    
    log_str("\n");
    logcnt=200;
    //logcnt=50;
  }
#endif  
  led_set(LED_L_RD, (sensoric_floor_c==3)||(sensoric_floor_c==0));
  led_set(LED_R_RD, (sensoric_floor_c==3)||(sensoric_floor_c==0));
  led_set(LED_L_YE, (sensoric_floor_c==2)||(sensoric_floor_c==0));
  led_set(LED_R_YE, (sensoric_floor_c==2)||(sensoric_floor_c==0));
  
}

void allow(uint8_t sensor, int16_t minval, int16_t maxval) {
  if (sensor<SENS_CNT) {
    sensors[sensor].max=maxval;
    sensors[sensor].min=minval;
    return;
  }
  if ((sensor==SENS_COLL)||(sensor==SENS_ALL)) {
    sensors[SENS_COLL_L].max=sensors[SENS_COLL_R].max=maxval;
    sensors[SENS_COLL_L].min=sensors[SENS_COLL_R].min=minval;
  }
  if ((sensor==SENS_LINE)||(sensor==SENS_ALL)) {
    sensors[SENS_LINE_L].max=sensors[SENS_LINE_C].max=sensors[SENS_LINE_R].max=maxval;
    sensors[SENS_LINE_L].min=sensors[SENS_LINE_C].min=sensors[SENS_LINE_R].min=minval;
  }
  if ((sensor==SENS_BKITXS)||(sensor==SENS_ALL)) {
    sensors[SENS_BKITXS_L].max=sensors[SENS_BKITXS_R].max=maxval;
    sensors[SENS_BKITXS_L].min=sensors[SENS_BKITXS_R].min=minval;
  }  
}

static void allow_max(uint8_t sensor, int16_t maxval) {
  allow(sensor, INT16_MIN, maxval);
}

static void allow_min(uint8_t sensor, int16_t minval) {
  allow(sensor, minval, INT16_MAX);
}

static void ignore(uint8_t sensor) {
  allow(sensor, INT16_MIN, INT16_MAX);
}

// return true if all sensors are in range
uint8_t check(uint8_t sensor) {
  if (sensor<SENS_CNT) {
    return !((sensors[sensor].val>sensors[sensor].max) || (sensors[sensor].val<sensors[sensor].min));
  }
  uint8_t res = true;
  if ((sensor==SENS_COLL)||(sensor==SENS_ALL)) {
    res &= check(SENS_COLL_L) && check(SENS_COLL_R);
  }
  if ((sensor==SENS_LINE)||(sensor==SENS_ALL)) {
    res &= check(SENS_LINE_L) && check(SENS_LINE_C) && check(SENS_LINE_R);
  }
  if ((sensor==SENS_BKITXS)||(sensor==SENS_ALL)) {
    res &= check(SENS_BKITXS_L) && check(SENS_BKITXS_R);
  }
  return res;
}

uint8_t check_target() {
  return !(target_d || target_l || target_r);
}

uint8_t bot_halt(uint16_t time) {
  motpid_stop(1);
  pico_seq = 0x00;
  while (time--) {
    clock_wait_timer();
  }
}

uint8_t bot_wait_sens(uint16_t time) {
  uint8_t res = 1;
  while (time--) {
    clock_wait_timer();
    
    // check sensors...
    res = 0;
    sensor_update();
    if (check(SENS_ALL)) break;
    res = 1;
  }
  return res;
}

uint8_t bot_go(int8_t speedL, int8_t speedR, uint16_t time) {
  uint8_t new_pico_seq = 0x00;
  if (speedL>0) new_pico_seq |= 0xF0;
  if (speedL<0) new_pico_seq |= 0xB0;
  if (speedR>0) new_pico_seq |= 0x0F;
  if (speedR<0) new_pico_seq |= 0x0B;
  
  if ((new_pico_seq!=pico_seq) && (pico_seq!=0x00)) {
    motpid_stop(1);
    clock_delay_n(MS(50));
  }

  odometry_reset();
  
  motpid_setSpeed(speedL, speedR);
  
  if ((new_pico_seq!=pico_seq) && (new_pico_seq!=0x00)) {    
    switch (new_pico_seq) {
      case 0xFF: modpid_load_pwm(+200, +200); break;
      case 0xBB: modpid_load_pwm(-200, -200); break;
      case 0xFB: modpid_load_pwm(+300, -300); break;
      case 0xBF: modpid_load_pwm(-300, +300); break;
      case 0x0B: modpid_load_pwm(0, -300); break;
      case 0x0F: modpid_load_pwm(0, +300); break;
      case 0xF0: modpid_load_pwm(+300, 0); break;
      case 0xB0: modpid_load_pwm(-300, 0); break;
    }    
  }

  pico_seq = new_pico_seq;

  uint8_t res = 1;
  while (time--) {
     clock_wait_timer();

    // check sensors...
    sensor_update();
    res = 0;
    if (!check(SENS_ALL)) break;

    // check for success
    res = 1;
    int16_t odoL = odometry_getLeft(0);
    int16_t odoR = odometry_getRight(0);
    if ((target_l>0) && (odoL>target_l)) break;
    if ((target_l<0) && (odoL<target_l)) break;
    if ((target_r>0) && (odoR>target_r)) break;
    if ((target_r<0) && (odoR<target_r)) break;
    if ((target_d>0) && (odoR-odoL>target_d)) break;
    if ((target_d<0) && (odoR-odoL<target_d)) break;    
  }
  int16_t odoL = odometry_getLeft(0);
  int16_t odoR = odometry_getRight(0);
  int16_t odoD = odoR-odoL;
  
  if (target_l>0) target_l = (target_l>odoL)?(target_l-odoL):0;
  if (target_l<0) target_l = (target_l<odoL)?(target_l-odoL):0;
  if (target_r>0) target_r = (target_r>odoR)?(target_r-odoR):0;
  if (target_r<0) target_r = (target_r<odoR)?(target_r-odoR):0;
  if (target_d>0) target_d = (target_d>odoD)?(target_d-odoD):0;
  if (target_d<0) target_d = (target_d<odoD)?(target_d-odoD):0;

  /*
  if (target_l) target_l -= odometry_getLeft(0);
  if (target_r) target_r -= odometry_getRight(0);
  if (target_d) target_d -= odometry_getRight(0)-odometry_getLeft(0);
  */
  
  return res;
}

void set_target(int16_t ticksL, int16_t ticksR) {
  target_l = ticksL;
  target_r = ticksR;
  target_d = 0;
}

void set_target_dir(int16_t ticksD) {
  target_l = 0;
  target_r = 0;
  target_d = ticksD;
}


uint8_t cmd_go(int16_t ticks, uint8_t speed, uint8_t coll) {
  set_target(ticks, ticks);
  if (ticks>0) {
    allow_max(SENS_COLL, coll);
    return bot_go(speed, speed, MS(5000));
  } else if (ticks<0) {
    allow_min(SENS_COLL, -coll);
    return bot_go(-speed, -speed, MS(5000));
  }
}

uint8_t cmd_turn(int8_t ticks, uint8_t speed, uint8_t coll) {
  set_target_dir(ticks);
  if (ticks>0) {
    // counterclockwise
    allow_min(SENS_COLL_L, -coll);
    allow_max(SENS_COLL_R, coll);
    return bot_go(-speed, speed, MS(5000));
  } else if (ticks<0) {
    // clockwise
    allow_min(SENS_COLL_R, -coll);
    allow_max(SENS_COLL_L, coll);
    return bot_go(speed, -speed, MS(5000));
  } else {
    bot_halt(MS(100));
  }
}

uint8_t cmd_turn_space(int8_t ticks, uint8_t speed, uint8_t coll, uint8_t space) {
  set_target_dir(ticks);
  if (ticks>0) {
    // counterclockwise
    allow_min(SENS_COLL_L, -coll);
    allow(SENS_COLL_R, space, coll);
    return bot_go(-speed, speed, MS(5000));
  } else {
    // clockwise
    allow_min(SENS_COLL_R, -coll);
    allow(SENS_COLL_L, space, coll);
    return bot_go(speed, -speed, MS(5000));
  }
}


/////////////////////////////////////

uint8_t escape2() {
  led_set(LED_L_RD, 1);
  led_set(LED_R_RD, 1);
  led_set(LED_L_YE, 0);
  led_set(LED_R_YE, 0);
  
  ignore(SENS_ALL);
  bot_halt(MS(100));
  uint8_t res = true;
  
  if (sensors[SENS_LINE_L].val<sensors[SENS_LINE_R].val) {
    log_str("escape R\n");
    allow_max(SENS_COLL_L, 3);
    allow_min(SENS_COLL_R, -3);
    set_target_dir(-40);
    while (bot_go(+10, -10, MS(5000))==0) {
      log_str("coll\n");
      led_set(LED_L_YE, 1);
      led_set(LED_R_YE, 1);
      bot_halt(MS(100));  
      while (bot_wait_sens(MS(100))==0) {
	// collision
      }
      led_set(LED_L_YE, 0);
      led_set(LED_R_YE, 0);
      log_str("clear\n");
    }
  } else {
    log_str("escape L\n");
    allow_max(SENS_COLL_R, 3);
    allow_min(SENS_COLL_L, -3);
    set_target_dir(+40);
    while (bot_go(-10, +10, MS(5000))==0) {
      log_str("coll\n");
      led_set(LED_L_YE, 1);
      led_set(LED_R_YE, 1);
      bot_halt(MS(100));  
      while (bot_wait_sens(MS(100))==0) {
	// collision
      }
      led_set(LED_L_YE, 0);
      led_set(LED_R_YE, 0);
      log_str("clear\n");
    }
  }
  led_set(LED_L_RD, 0);
  led_set(LED_R_RD, 0);
  
  log_str(res?"SUCCESS\n":"FAIL\n");
  return res;
}

/////////////////////////////////////
uint8_t runaround() {
  int8_t turn = 0;
  int16_t coll_l = sensors[SENS_LINE_L].val;
  int16_t coll_r = sensors[SENS_LINE_R].val;
  
  if (coll_l<coll_r) {
    turn = +1;
  } else if (coll_l>coll_r) {
    turn = -1;
  } else if (sensors[SENS_BKITXS_L].val<sensors[SENS_BKITXS_R].val) {
    turn = +1;
  } else {
    turn = -1;
  }
  bot_halt(MS(100));
  log_str("runaround\n");
  allow_min(SENS_LINE, BLACK_MAX);
  
  uint8_t res = cmd_go(-20, 10, 3)
             && cmd_turn(turn*18, 10, 3)
             && cmd_go(20, 10, 3);

  log_str(res?"SUCCESS\n":"FAIL\n");
  return res;
}

/////////////////////////////////////


uint8_t from_line(int8_t turnticks) {
  bot_halt(MS(10));
  
  ignore(SENS_ALL);
  cmd_turn((turnticks<0)?-10:10, 10, 3);
  allow_max((turnticks<0)?SENS_BKITXS_R:SENS_BKITXS_L, 0x30);  
  cmd_turn((turnticks<0)?-50:50, 10, 3);
  uint8_t res = check(SENS_BKITXS);
  
  allow_min((turnticks<0)?SENS_BKITXS_R:SENS_BKITXS_L, 0x20);
  allow_max((turnticks<0)?SENS_BKITXS_L:SENS_BKITXS_R, 0x20);
  cmd_turn((turnticks<0)?-2:2, 10, 3);
  

  if (!res) {
    // Dose gefunden
    bot_halt(MS(100));
    sensoric_reset_floor();
    ignore(SENS_BKITXS);
    allow_min(SENS_LINE, BLACK_MAX);   
    cmd_go(40, 10, 3);
    cmd_go(50, 17, 3);  

    int16_t delta = sensoric_angle_hist[1]+sensoric_angle_hist[2];
    cmd_turn(delta/2, 10, 3);
    cmd_go(70, 17, 3);
  
    bot_halt(MS(10));
    clock_delay_n(MS(1000));

    cmd_go(-20, 17, 3);
    cmd_turn((turnticks<0)?-62:62, 10, 3);
    
    cmd_go(120, 17, 3);
    
    // langsam vorwärts bis zur Linie
    cmd_go(50, 10, 3);    
    bot_halt(MS(50));

    if (check(SENS_LINE_L) != check(SENS_LINE_R)) {
      // falls schräg ein Tick weiter zum Testen!
      ignore(SENS_LINE);
      cmd_go(1, 10, 3);
      bot_halt(MS(50));
    }

    
    if (check(SENS_LINE_L) == check(SENS_LINE_R)) {    
      // Roboter steht 90 Grad zur Linie drehen
      ignore(SENS_LINE);
      
      if (!check(SENS_LINE_L)) {
	// steht auf der Linie!
	// 3 Ticks vorwärts
        cmd_go(3, 10, 3);
        bot_halt(MS(10));
      }
      
      if (check(SENS_LINE_L) == check(SENS_LINE_R)) {        
        if (check(SENS_LINE_L)) {
          // steht hinter der Linie!
          allow_min((turnticks<0)?SENS_LINE_R:SENS_LINE_L, BLACK_MAX);
        } else {
          // steht auf der Linie!
          allow_max((turnticks<0)?SENS_LINE_R:SENS_LINE_L, BLACK_MAX);
        }
      
        // 360 Grad drehen bis schräg auf Linie...
        cmd_turn((turnticks<0)?-122:122, 10, 3);
        bot_halt(MS(10));
      }
    }
    allow_min(SENS_LINE, BLACK_MAX);   
    
    
    // zurück bis zur Linie falls zu weit gefahren
    //cmd_go(-5, 10, 3);
    //bot_halt(MS(100));
    
    
  } else {    
    ignore(SENS_BKITXS);
    allow_min(SENS_LINE, BLACK_MAX);
    cmd_turn((turnticks<0)?-15:15, 10, 3);
  }
  bot_halt(MS(10));
  clock_delay_n(MS(100));
  return 0;
}


//////////////////////////

int main_obst() {
  activate_output_group(IO_LEDS);  // LED bits als Output

  led_set(LED_L_RD, 1);    
  led_set(LED_R_RD, 1);
  _delay_ms(200);
  led_set(LED_L_RD, 0);
  led_set(LED_R_RD, 0);
  _delay_ms(800);

  sensoric_init(SENSORIC_MODE_FLOOR);
  
  mode = MODE_HALT;
  counter = 200;
  int16_t fl=2*BLACK_MAX, fc=2*BLACK_MAX, fr=2*BLACK_MAX;
  
  clock_set_timer(2);
  ignore(SENS_ALL);
   
  // Hauptschleife:
  while(1) {
    sei();
    clock_wait_timer();
    
    sensor_update();
    
    uint8_t sensL = 0;
    uint8_t sensR = 0;    
    uint8_t coll = 0;
    
    int16_t lv, rv;
    
    fl *= 3;  fc *= 3; fr *= 3;
    fl += line_get(LINE_L);
    fc += line_get(LINE_C);
    fr += line_get(LINE_R);    
    fl /= 4;  fc /= 4; fr /= 4;
    
    
    if ((fl<BLACK_MAX) || (fc<BLACK_MAX) || (fr<BLACK_MAX)) {
      escape2();
      mode = MODE_HALT;
      counter = 100;
    }    
    
    
    lv = analog_getValue(ANALOG_EXT2H);
    rv = analog_getValue(ANALOG_EXT3H);
    lv -= analog_getValue(ANALOG_EXT2);
    rv -= analog_getValue(ANALOG_EXT3);
    
    lv = 8 * INTERVAL(0, lv, 20);
    rv = 8 * INTERVAL(0, rv, 20);
    
    rval = (31 * rval + rv + 15)/32;
    lval = (31 * lval + lv + 15)/32;
       
    if (sens_getLeft()) {
      coll = 1;
      sensL = 3;      
    } else if (lval>8*6) {
      sensL = 3;
    } else if (lval>8*4) {
      sensL = 2;
    } else if (lval>8*2) {
      sensL = 1;
    }
    
    if (sens_getRight()) {
      coll = 1;
      sensR = 3;
    } else if (rval>8*6) {
      sensR = 3;
    } else if (rval>8*4) {
      sensR = 2;
    } else if (rval>8*2) {
      sensR = 1;
    }
    
    /*
    led_set(LED_L_RD, (sensL > 1));
    led_set(LED_R_RD, (sensR > 1));

    led_set(LED_L_YE, ((sensL == 1) || (sensL == 2)));
    led_set(LED_R_YE, ((sensR == 1) || (sensR == 2)));
    */
    
    if (mode == MODE_BACK_R) {
      int16_t ol = odometry_getLeft(0)-odometry_getRight(0);
      if (ol < -35) {
	counter = 0;
      } 
      if ((ol < -15) && (!sensL && !sensR)) {
	counter = 0;
      }
    }
    if (mode == MODE_BACK_L) {
      int16_t or = odometry_getRight(0)-odometry_getLeft(0);
      if (or < -35) {
	counter = 0;
      }
      if ((or < -15) && (!sensL && !sensR)) {
	counter = 0;
      }
    }
    if (mode == MODE_BACK) {
      if ((odometry_getRight(0) < -40) || (odometry_getLeft(0) < -40)) {
	counter = 0;
      }
    }
        
    if (counter) {
        counter--;
    } else {
      switch (mode) {
        case MODE_HALT:
	  if (sensR>sensL) {
	    mode = MODE_BACK_R;
	    odometry_reset();
            counter=1500;
	  } else if (sensL>sensR) {
	    mode = MODE_BACK_L;
	    odometry_reset();
            counter=1500;
	  } else if (sensL>0) {
	    mode = MODE_BACK;
	    odometry_reset();
	    counter=1500;
	  } else {
	    mode = MODE_SLOW;
	  }
          break;
	
        case MODE_GO:
        case MODE_GO_L:
        case MODE_GO_R:
        case MODE_SLOW:
        case MODE_SLOW_L:
        case MODE_SLOW_R:
	  if (coll) {
	    if (sensR>sensL) {
	      last_coll = +1;
	    
	    } else if (sensL>sensR) {
	      last_coll = -1;
	    
	    } else /* sensL==sensR */ {
	    
	    }
	    runaround();
	    mode = MODE_SLOW;
	    
	  } else if (sensR>sensL) {
	    if (sensR==3) {
	      last_coll = +1;
	      runaround();
	      mode = MODE_SLOW;
	    } else if (sensR==2) {
	      last_coll = +1;
	      runaround();
              mode = MODE_SLOW;
	    } else {
	      mode = MODE_GO;    
	    }
	  } else if (sensL>sensR) {
	    if (sensL==3) {
	      last_coll = -1;
	      runaround();
	      mode = MODE_SLOW;
	    } else if (sensL==2) {
	      last_coll = -1;
	      runaround();
	      mode = MODE_SLOW;
	    } else {
	      mode = MODE_GO;
	    }
	  } else /* sensL==sensR */ {
	    if (sensL==3) {
	      mode = MODE_BACK;
	      counter = 200;
	    } else if (sensL==2) {
	      mode = MODE_HALT;
	      counter = 200;
	    } else if (sensL==1) {
	      mode = MODE_SLOW;
	    } else {
	      mode = MODE_GO;
	    }    
	  }
	  break;
	  
        case MODE_BACK:
	  mode = MODE_BACK_HALT;
          counter = 100;
	  break;
	  
        case MODE_BACK_L:
	  mode = MODE_BACK_HALT;
          counter = 100;
	  break;
	  
        case MODE_BACK_R:
	  mode = MODE_BACK_HALT;
          counter = 100;
	  break;
	  
	case MODE_BACK_HALT:
	  if (last_coll==+1) {
	    mode = MODE_BACK_R;
	    last_coll=0;
	  } else if (last_coll==-1) {
	    mode = MODE_BACK_L;
	    last_coll=0;
	  } else {
            mode = MODE_SLOW;
	  }
	  break;
	  
      }
    }
    
    switch (mode) {
        case MODE_HALT:      motpid_stop(1); break;
        case MODE_GO:        motpid_setSpeed ( 17,  17); break;
        case MODE_GO_L:      motpid_setSpeed ( 10,  22); break;
        case MODE_GO_R:      motpid_setSpeed ( 22,  10); break;
        case MODE_SLOW:      motpid_setSpeed ( 10,  10); break;
        case MODE_SLOW_L:    motpid_setSpeed (  0,  10); break;
        case MODE_SLOW_R:    motpid_setSpeed ( 10,   0); break;
        case MODE_BACK:      motpid_setSpeed (-10, -10); break;
        case MODE_BACK_L:    motpid_setSpeed ( -5,  -10); break;
        case MODE_BACK_R:    motpid_setSpeed (-10,   -5); break;
        case MODE_BACK_HALT: motpid_stop(1); break;
    }
    
  }
  return 0;
}
