#include <nibobee/iodefs.h>
#include <nibobee/motpwm.h>
#include <nibobee/motpid.h>
#include <nibobee/analog.h>
#include <nibobee/line.h>
#include <nibobee/led.h>
#include <nibobee/sens.h>
#include <nibobee/delay.h>
#include <nibobee/clock.h>
#include <nibobee/odometry.h>

#include <stdio.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include <util/delay.h>

#include "util.h"
#include "sensoric.h"

uint8_t sensoric_mode;
int8_t sensoric_line_pos;
int8_t sensoric_line_posf;
uint8_t sensoric_line_valid;
uint8_t sensoric_floor_color;
uint16_t sensoric_floor_dt;

uint16_t sensoric_angle_hist[SENSORIC_ANGLE_HIST_SIZE];


/* POSITION */
int16_t sensoric_ori;
int16_t sensoric_x;
int16_t sensoric_y;


static void update_line_pos() {
  int16_t l = line_get(LINE_L);
  int16_t c = line_get(LINE_C);
  int16_t r = line_get(LINE_R);
  int8_t actpos = sensoric_line_pos;
  int16_t pos = 0;
 
  if ((l>=0x30) && (c>=0x30) && (r>=0x30)) {
    // NO LINE!!
    if (actpos<0) {
      pos = -31;
    } if (actpos>0) {
      pos = +31;
    } else {
      //sensoric_line_valid = 0;
    }
    
  } else if (c<0x30) {
    // CENTER
    int16_t d = l-r;
    if (d < -0x100) {
      pos = -4 + (d+0x100)/0x80;
      pos = INTERVAL(-8, pos, -4);
    } else if (d > 0x100) {
      pos = +4 + (d+0x100)/0x80;
      pos = INTERVAL(4, pos, 8);
    } else {
      pos = d/0x40;
      pos = INTERVAL(-4, pos, 4);
    }
    sensoric_line_valid = 1;
  } else if (l<r) {
    // LEFT
    int16_t s = l+c;
    pos = -s / 0x45 - 8;
    pos = INTERVAL(-24, pos, -8);
    sensoric_line_valid = 1;
  } else {
    // RIGHT
    int16_t s = l+c;
    pos = s / 0x45 + 8;
    pos = INTERVAL(8, pos, 24);
    sensoric_line_valid = 1;
  }
  if (actpos<pos) actpos++;
  if (actpos>pos) actpos--;
  sensoric_line_pos = actpos;
  if (actpos<pos) actpos++;
  if (actpos>pos) actpos--;
  sensoric_line_posf = actpos;
}

/* FLOOR */

uint8_t hist_l[8];
uint8_t hist_c[8];
uint8_t hist_r[8];

uint16_t sum_l;
uint16_t sum_c;
uint16_t sum_r;

uint8_t hpos;

uint8_t sensoric_floor_l;
uint8_t sensoric_floor_c;
uint8_t sensoric_floor_r;

uint8_t sensoric_floor_cnt;
int16_t sensoric_floor_delta_t;
uint16_t delta_grey;
uint16_t delta_t;
int8_t delta_side;

int8_t sensoric_floor_alert;



// BKit-XS
int16_t sensoric_bkxs_l = 0;
int16_t sensoric_bkxs_r = 0;


static uint8_t calcLClass(int16_t val) {
  if (val>8*0x38) return 3;
  if (val>8*0x28) return 2;
  if (val>8*0x08) return 1;
  return 0;
}

static uint8_t clip_line(int16_t val) {
  if (val<0) return 0;
  if (val>16*255) return 255;
  return val/16;
}

static void update_bkitxs() {
  // update BKit-XS sensors
  int16_t value;
  
  value = analog_getValue(ANALOG_EXT2H);
  value -= analog_getValue(ANALOG_EXT2);
  value = 8 * INTERVAL(0, value, 64);
  sensoric_bkxs_l = (15 * sensoric_bkxs_l + value + 15)/16;

  value = analog_getValue(ANALOG_EXT3H);
  value -= analog_getValue(ANALOG_EXT3);    
  value = 8 * INTERVAL(0, value, 64);
  sensoric_bkxs_r = (15 * sensoric_bkxs_r + value + 15)/16;
}       



static void update_floor() {
  uint8_t l = clip_line(line_get(LINE_L));
  sum_l = sum_l - hist_l[hpos] + l;
  hist_l[hpos] = l;
  
  uint8_t c = clip_line(line_get(LINE_C));
  sum_c = sum_c - hist_c[hpos] + c; 
  hist_c[hpos] = c;

  uint8_t r = clip_line(line_get(LINE_R)); 
  sum_r = sum_r - hist_r[hpos] + r; 
  hist_r[hpos] = r;
  
  if (++hpos>7) {
    hpos=0;
  }
  
  sensoric_floor_l = calcLClass(sum_l);
  sensoric_floor_c = calcLClass(sum_c);
  sensoric_floor_r = calcLClass(sum_r);
  
  if (sensoric_floor_alert==0) {
    if ( (sum_l<8*3) || (sum_c<8*5) || (sum_r<8*3) ) {
      sensoric_floor_alert = (sum_l<sum_r)? -1 : 1;
    }
  }
  
  uint8_t grey = ((sensoric_floor_l==1)?0x10:0x00) | ((sensoric_floor_r==1)?0x01:0x00);
  
  if ((grey==0x11)||(grey==0x00)) {
    if ((grey ^ delta_grey)) {
      delta_grey = grey;
      sensoric_floor_delta_t = delta_t;
      sensoric_floor_delta_t *= delta_side;
      if (sensoric_floor_cnt<SENSORIC_ANGLE_HIST_SIZE) {
        sensoric_angle_hist[sensoric_floor_cnt] = sensoric_floor_delta_t;
      }
      sensoric_floor_cnt++;
      delta_t = 0;      
    }
  } else {
    if (delta_t==0) {
      delta_side = ((grey^delta_grey) ==0x01)?-1:1;
    }
    delta_t++;
  }
}

void sensoric_reset_floor() {
  ATOMIC_BLOCK(ATOMIC_RESTORESTATE) {
    update_floor();
    sensoric_floor_delta_t = 0;
    sensoric_floor_cnt = 0;
    delta_t = 0;    
  }
}



void sensoric_init(uint8_t mode) {
  if (sensoric_mode != mode) {
    sensoric_mode = mode;
    if (mode==SENSORIC_MODE_LINE) {
      sensoric_line_pos = 0;
      sensoric_line_valid = 0;
    } else if (mode==SENSORIC_MODE_FLOOR) {
      delta_t = 0;
      delta_grey = 0xff;
      sensoric_floor_delta_t = 0;
      sensoric_floor_cnt = 0;
    }
  }
  
}


void analog_irq_hook() {
  if (sensoric_mode==SENSORIC_MODE_LINE) {
    update_line_pos();
  }
  if (sensoric_mode==SENSORIC_MODE_FLOOR) {
    update_floor();
    update_bkitxs();
  }
}


