/*
 * charger.c - NIBObee charger
 *
 * Autor..........: Nils Springob nils@nicai-systems.de
 * Description....: Provides functions for NIBObee charging
 * Licence........: GNU GPL v2
 * Creation Date..: 2009-10-08
 */

#include <avr/io.h>

#include "charger.h"


// 8 hours * 150 mA = 1200 mAh = 120%
#define CHARGE_MINUTES (60*8)

#define VOLTAGE(v) (int16_t)(1024/1.1/11*(v))
#define CURRENT(c) VOLTAGE(3.3*(c))


#define charge_on() PORTB &= (uint8_t)~_BV(2)
#define charge_off() PORTB |= _BV(2)


#define MUX_VOLTAGE_SUPP 0x21
#define MUX_VOLTAGE_BAT  0x81
#define MUX_CURRENT      0x88


uint16_t charge_cnt = CHARGE_MINUTES*60/2;
int16_t ms;
uint8_t charge_mode;

uint16_t voltage_supp;
uint16_t voltage_bat;
uint16_t current;
uint8_t curr_err_cnt;

enum {
  CHARGE_POLL,
  CHARGE_ACTIVE,
  CHARGE_FIN,
  CHARGE_ERR
};

static void start_measure(uint8_t mux) {
  DDRA &= (uint8_t)~_BV(0);
  DDRA &= (uint8_t)~_BV(1);
  PORTA &= (uint8_t)~_BV(0);
  PORTA &= (uint8_t)~_BV(1);
  ADCSRA = 0x00;
  ADMUX = mux;
  ADCSRA = _BV(ADEN) | _BV(ADSC) | _BV(ADPS2) | _BV(ADPS1) | _BV(ADPS0);
}


static void do_charge_poll() {
  if (ms==490) {
    start_measure(MUX_VOLTAGE_SUPP);
  } else if (ms==491) {
    voltage_supp = ADC;
  } else if (ms==492) {
    start_measure(MUX_VOLTAGE_BAT);
  } else if (ms==493) {
    voltage_bat = ADC;
    if (voltage_bat>VOLTAGE(0.2)) {
      charge_mode = CHARGE_ACTIVE;
    }
  } else if (ms==997) {
    start_measure(MUX_CURRENT);
  } else if (ms==998) {
    current = ADC;
  }


  if (ms>499) {
    charge_on();
  } else {
    charge_off();
  }
  if (ms>999) {
    ms=0;
  }
}


static void do_charge_active() {
  if (ms==16) {
    start_measure(MUX_VOLTAGE_SUPP);
  } else if (ms==17) {
    voltage_supp = ADC;
  
  } else if (ms==18) {
    start_measure(MUX_VOLTAGE_BAT);
    if ((voltage_bat<VOLTAGE(0.2)) && (current<CURRENT(0.05))) {
      charge_mode = CHARGE_POLL;
    }
  } else if (ms==19) {
    voltage_bat = ADC;
  
  } else if (ms==997) {
    start_measure(MUX_CURRENT);
  } else if (ms==998) {
    current = ADC;
    
    if (current>CURRENT(0.6)) {
      curr_err_cnt++;
      if (curr_err_cnt>5) {
        charge_mode = CHARGE_ERR;
      }
    } else {
      curr_err_cnt = 0;
    }
    
  }


  if (ms>20) {
    charge_on();
  } else {
    charge_off();
  }

  if (charge_cnt==0) {
    charge_mode = CHARGE_FIN;
  }

  if (ms>999) {
    ms=0;
    charge_cnt--;
  }
}


static void do_charge_fin() {
  if (ms<10) {
    charge_on();
  } else {
    charge_off();
  }
  if (ms>1999) {
    ms=0;
  }
}

static void do_charge_err() {
  if (ms<10) {
    charge_on();
  } else if (ms<200) {
    charge_off();
  } else  if (ms<210) {
    charge_on();
  } else {
    charge_off();
  }
  if (ms>999) {
    ms=0;
  }
}


// called each ms
void do_charger() {
  ms++;
  switch (charge_mode) {
    case CHARGE_POLL:   do_charge_poll(); break;
    case CHARGE_ACTIVE: do_charge_active(); break;
    case CHARGE_FIN:    do_charge_fin(); break;
    case CHARGE_ERR:    do_charge_err(); break;
    default: ms=0;
  }
}

void do_charger_stop() {
  charge_off();
}

void init_charger() {
  DDRB =_BV(2);
  DIDR0 = _BV(0) | _BV(1); // disable digital buffers for A0 and A1
}
