/*
 * isp.c - part of USBasp - modified version for NIBObee charger/programmer
 *
 * Autor..........: Thomas Fischl <tfischl@gmx.de>
 * Description....: Provides functions for communication/programming
 *                  over ISP interface
 * Licence........: GNU GPL v2 (see Readme.txt)
 * Creation Date..: 2005-02-23
 */

#include <avr/io.h>
#include <util/delay.h>
#include "isp.h"
#include "clock.h"
#include "usbasp.h"

#define spiHWdisable() SPCR = 0

uchar sck_sw_type;
uchar sck_sw_delay;
uchar sck_spcr;
uchar sck_spsr;

uchar swap_msio;

void spiHWenable() {

}

void ispSetSCKOption(uchar option) {

	if (option == USBASP_ISP_SCK_AUTO)
		option = USBASP_ISP_SCK_375;

	if (option >= USBASP_ISP_SCK_93_75) {
                sck_sw_type = 1;

		switch (option) {

		case USBASP_ISP_SCK_1500:
			sck_sw_delay = 1;
                        break;
		case USBASP_ISP_SCK_750:
			sck_sw_delay = 2;
			break;
		case USBASP_ISP_SCK_375:
		default:
			sck_sw_delay = 4;
			break;
		case USBASP_ISP_SCK_187_5:
			sck_sw_delay = 8;
			break;
		case USBASP_ISP_SCK_93_75:
			sck_sw_delay = 16;
			break;
		}

	} else {
                sck_sw_type = 0;
		switch (option) {

		case USBASP_ISP_SCK_32:
			sck_sw_delay = 4; //3

			break;
		case USBASP_ISP_SCK_16:
			sck_sw_delay = 8; //6

			break;
		case USBASP_ISP_SCK_8:
			sck_sw_delay = 15; //12

			break;
		case USBASP_ISP_SCK_4:
			sck_sw_delay = 30; //24

			break;
		case USBASP_ISP_SCK_2:
			sck_sw_delay = 60; //48

			break;
		case USBASP_ISP_SCK_1:
			sck_sw_delay = 120; //96

			break;
		case USBASP_ISP_SCK_0_5:
			sck_sw_delay = 240; //192
			break;
		}
	}
}

void ispDelay() {
  if (sck_sw_type) {
    uint8_t delay = sck_sw_delay;
    while(delay--) {
      _delay_us(0.66666666);
    }
  } else {
    uint8_t starttime = TIMERVALUE;
    while ((uint8_t) (TIMERVALUE - starttime) < sck_sw_delay) {
    }
  }
}

void ispConnect() {
	/* all ISP pins are inputs before */
	/* now set output pins */
        if (swap_msio) {
	  ISP_DDR |= (1 << ISP_RST) | (1 << ISP_SCK) | (1 << ISP_MISO);
	} else {
	  ISP_DDR |= (1 << ISP_RST) | (1 << ISP_SCK) | (1 << ISP_MOSI);
	}
	/* reset device */
	ISP_OUT &= ~(1 << ISP_RST); /* RST low */
	ISP_OUT &= ~(1 << ISP_SCK); /* SCK low */

	/* positive reset pulse > 2 SCK (target) */
	ispDelay();
	ISP_OUT |= (1 << ISP_RST); /* RST high */
	ispDelay();
	ISP_OUT &= ~(1 << ISP_RST); /* RST low */
	//if (ispTransmit == ispTransmit_sw0) {
	//	spiHWenable();
	//}
}

void ispDisconnect() {
	/* set all ISP pins inputs */
	ISP_DDR &= ~((1 << ISP_RST) | (1 << ISP_SCK) | (1 << ISP_MOSI) | (1 << ISP_MISO));
	/* switch pullups off */
	ISP_OUT &= ~((1 << ISP_RST) | (1 << ISP_SCK) | (1 << ISP_MOSI) | (1 << ISP_MISO));
	/* disable hardware SPI */
	//spiHWdisable();
}

uchar ispTransmit(uchar byte) {
  uchar i;
  for (i = 0; i < 8; i++) {
    /* set MSB to MOSI-pin */
    if ((byte & 0x80) != 0) {
      if (swap_msio) {
        ISP_OUT |= (1 << ISP_MISO); /* MOSI high */
      } else {
        ISP_OUT |= (1 << ISP_MOSI); /* MOSI high */
      }
    } else {
      if (swap_msio) {
        ISP_OUT &= ~(1 << ISP_MISO); /* MOSI low */
      } else {
        ISP_OUT &= ~(1 << ISP_MOSI); /* MOSI low */
      }
    }
    /* shift to next bit */
    byte = byte << 1;

    // rising edge
    ispDelay();
    ISP_OUT |= (1 << ISP_SCK); /* SCK high */
    /* receive data */
    if (swap_msio) {
      if ((ISP_IN & (1 << ISP_MOSI)) != 0) {
        byte++;
      }
    } else {
      if ((ISP_IN & (1 << ISP_MISO)) != 0) {
        byte++;
      }
    }

    /* falling edge */
    ispDelay();
    ISP_OUT &= ~(1 << ISP_SCK); /* SCK low */
  }
  return byte;
}

#ifdef _DONT_COMPILE_

uint8_t exchange(uint8_t val) {
  uint8_t res=0x00;
  for (uint8_t i=0; i<8; ++i) {
    // SETUP, falling edge
    clear_output_bit(IO_ISP_SCK);
    set_output_bitval(IO_ISP_MOSI, (val&0x80));
    val <<= 1;
    uint8_t delay=param_sck_duration;
    while(delay--) __asm volatile("nop");
    // SAMPLE, rising edge
    set_output_bit(IO_ISP_SCK);
    res <<= 1;
    res |= get_input_bit(IO_ISP_MISO)?0x01:0x00;
    delay=param_sck_duration;
    while(delay--) __asm volatile("nop");
    clear_output_bit(IO_ISP_SCK);
  }
  return res;
}

void cmd_enter_progmode_isp() {
  set_output_bit(IO_ISP_MISO);
  set_output_bit(IO_ISP_MOSI);
  set_output_bit(IO_ISP_SCK);
  clear_output_bit(IO_ISP_RESET);

  activate_output_bit(IO_ISP_RESET);
  activate_output_bit(IO_ISP_SCK);
  if (swap_msio) {
    activate_output_bit(IO_ISP_MISO);
    deactivate_output_bit(IO_ISP_MOSI);
  } else {
    activate_output_bit(IO_ISP_MOSI);
    deactivate_output_bit(IO_ISP_MISO);
  }
  delay_ms(com_message[2]); // stabDelay

  clear_output_bit(IO_ISP_SCK);
  if (swap_msio) {
    clear_output_bit(IO_ISP_MISO);
  } else {
    clear_output_bit(IO_ISP_MOSI);
  }
  delay_us(100);

  // positive reset pulse
  set_output_bit(IO_ISP_RESET);
  delay_us(100);

  clear_output_bit(IO_ISP_RESET);
  delay_ms(com_message[3]); // exeDelay

  uint8_t byteDelay = com_message[5];
  uint8_t count = com_message[4]; // synchLoops
  if (count<1) {
    count=1;
  }
  while(count--) {
    uint8_t res[4];
    res[0] = exchange(com_message[8]);
    delay_ms(byteDelay);
    res[1] = exchange(com_message[9]);
    delay_ms(byteDelay);
    res[2] = exchange(com_message[10]);
    delay_ms(byteDelay);
    res[3] = exchange(com_message[11]);
    delay_ms(byteDelay);

    uint8_t poll_index = com_message[7];
    if ((poll_index==0) || (res[poll_index-1]==com_message[6])) {
      led_flash();
      avrisp_mode = AVRISP_PROG;
      _delay_ms(10);
      com_message[1] = STATUS_CMD_OK;
      com_message_size = 2;
      return;
    }
  }
  com_message[1] = STATUS_CMD_FAILED;
  com_message_size = 2;
}

uchar ispTransmit_sw(uchar value) {
  USICR = _BV(USIWM0) | _BV(USICS1) | _BV(USICLK);
  USIDR = value;
  USISR = _BV(USIOIF);
  do {
    ispDelay();
    USICR = _BV(USIWM0) | _BV(USICS1) | _BV(USICLK) | _BV(USITC);
  } while (!(USISR & _BV(USIOIF)));
  value = USIDR;
  USICR = 0;
  return value;
}

#endif


uchar ispEnterProgrammingMode() {
	uchar check;
	uchar count = 32;

	while (count--) {
		ispTransmit(0xAC);
		ispTransmit(0x53);
		check = ispTransmit(0);
		ispTransmit(0);

		if (check == 0x53) {
			return 0;
		}

		//spiHWdisable();

		/* pulse SCK */
		ISP_OUT |= (1 << ISP_SCK); /* SCK high */
		ispDelay();
		ISP_OUT &= ~(1 << ISP_SCK); /* SCK low */
		ispDelay();

	}

	return 1; /* error: device dosn't answer */
}

uchar ispReadFlash(unsigned long address) {
	ispTransmit(0x20 | ((address & 1) << 3));
	ispTransmit(address >> 9);
	ispTransmit(address >> 1);
	return ispTransmit(0);
}

uchar ispWriteFlash(unsigned long address, uchar data, uchar pollmode) {

	/* 0xFF is value after chip erase, so skip programming
	 if (data == 0xFF) {
	 return 0;
	 }
	 */

	ispTransmit(0x40 | ((address & 1) << 3));
	ispTransmit(address >> 9);
	ispTransmit(address >> 1);
	ispTransmit(data);

	if (pollmode == 0)
		return 0;

	if (data == 0x7F) {
		clockWait(15); /* wait 4,8 ms */
		return 0;
	} else {

		/* polling flash */
		uchar retries = 30;
		uint8_t starttime = TIMERVALUE;
		while (retries != 0) {
			if (ispReadFlash(address) != 0x7F) {
				return 0;
			};

			if ((uint8_t) (TIMERVALUE - starttime) > CLOCK_T_320us) {
				starttime = TIMERVALUE;
				retries--;
			}

		}
		return 1; /* error */
	}

}

uchar ispFlushPage(unsigned long address, uchar pollvalue) {
	ispTransmit(0x4C);
	ispTransmit(address >> 9);
	ispTransmit(address >> 1);
	ispTransmit(0);

	if (pollvalue == 0xFF) {
		clockWait(15);
		return 0;
	} else {

		/* polling flash */
		uchar retries = 30;
		uint8_t starttime = TIMERVALUE;

		while (retries != 0) {
			if (ispReadFlash(address) != 0xFF) {
				return 0;
			};

			if ((uint8_t) (TIMERVALUE - starttime) > CLOCK_T_320us) {
				starttime = TIMERVALUE;
				retries--;
			}

		}

		return 1; /* error */
	}

}

uchar ispReadEEPROM(unsigned int address) {
	ispTransmit(0xA0);
	ispTransmit(address >> 8);
	ispTransmit(address);
	return ispTransmit(0);
}

uchar ispWriteEEPROM(unsigned int address, uchar data) {

	ispTransmit(0xC0);
	ispTransmit(address >> 8);
	ispTransmit(address);
	ispTransmit(data);

	clockWait(30); // wait 9,6 ms

	return 0;
}
