/*  BSD-License:

Copyright (c) 2007 by Nils Springob, nicai-systems, Germany

All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

  * Redistributions of source code must retain the above copyright notice,
    this list of conditions and the following disclaimer.
  * Redistributions in binary form must reproduce the above copyright notice,
    this list of conditions and the following disclaimer in the documentation
    and/or other materials provided with the distribution.
  * Neither the name nicai-systems nor the names of its contributors may be
    used to endorse or promote products derived from this software without
    specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

*/

#include "nibo/niboconfig.h"
#ifndef NIBO_USE_FLOOR
#error NIBO_USE_FLOOR must be defined in niboconfig.h when this file is compiled
#endif
#ifndef NIBO_USE_ADC
#error NIBO_USE_ADC must be defined in niboconfig.h when this file is compiled
#endif

#include "nibo/iodefs.h"
#include "nibo/floor.h"
#include "nibo/analog.h"
#include <avr/eeprom.h>
#include <util/delay.h> 
#include <avr/interrupt.h>

#define LINE_EEPROM ((void *)0)

#define RANGE 0x400
#define FACTOR 0x100
#define CALSAMPLES 0x20

#define GAIN_FACT ((uint32_t)RANGE*(uint32_t)FACTOR*(uint32_t)CALSAMPLES)

#ifdef __cplusplus
extern "C" {
#endif

/*! @obsolete */
uint16_t floor_absolute [4]; 

/*! @obsolete */
uint16_t floor_relative [4];
  
  
static uint16_t cal_gain[4];
static uint16_t cal_offset[4];

typedef struct {
  uint16_t black[4]; // sum of 2*16 samples
  uint16_t white[4]; // sum of 2*16 samples
} line_cal_t;

line_cal_t line_cal;


void floor_init() {
  floor_readPersistent();
  analog_init();
} 


static uint16_t get_differential(uint8_t idx) {
  uint16_t val0 = analog_getValue(ANALOG_LINE_L0+idx);
  uint16_t val1 = analog_getValue(ANALOG_LINE_L1+idx);
  return (val0>val1)?(val0-val1):0;
}

static uint16_t get_absolute1(uint8_t idx) {
  uint16_t val = analog_getValue(ANALOG_LINE_L1+idx);
  return (val<0x3ff)?(0x3ff-val):0;
}

static uint16_t get_absolute0(uint8_t idx) {
  uint16_t val = analog_getValue(ANALOG_LINE_L0+idx);
  return (val<0x3ff)?(0x3ff-val):0;
}


// new version...
void floor_calculateCalibration() {
  // 0x2000000L = 0x1000 * 0x100 * 0x20
  //               range   factor calib
  cal_gain[0] = GAIN_FACT / (line_cal.white[0]-line_cal.black[0]);
  cal_gain[1] = GAIN_FACT / (line_cal.white[1]-line_cal.black[1]);
  cal_gain[2] = GAIN_FACT / (line_cal.white[2]-line_cal.black[2]);
  cal_gain[3] = GAIN_FACT / (line_cal.white[3]-line_cal.black[3]);
  cal_offset[0] = line_cal.black[0] / CALSAMPLES;
  cal_offset[1] = line_cal.black[1] / CALSAMPLES;
  cal_offset[2] = line_cal.black[2] / CALSAMPLES;
  cal_offset[3] = line_cal.black[3] / CALSAMPLES;
}


/*
 * Differenz von LED an und aus mit Schwarzoffset
 */
uint16_t floor_get(uint8_t idx) {
  if (idx<4) {
    uint16_t val16 = get_differential(idx);
    val16 -= (val16>cal_offset[idx])?cal_offset[idx]:val16;
    uint32_t val = val16;
    val *= cal_gain[idx];
    val /= FACTOR;
    if (val>INT16_MAX) {
      return INT16_MAX;
    }
    return val;
  }
  return 0;
} 


/*
 * Wert bei LED aus ohne Schwarzoffset
 */
uint16_t floor_getAbs(uint8_t idx) {
  if (idx<4) {
    uint16_t val16 = get_absolute0(idx);
    uint32_t val = val16;
    val *= cal_gain[idx];
    val /= FACTOR;
    if (val>INT16_MAX) {
      return INT16_MAX;
    }
    return val;
  }
  return 0;
}


uint16_t floor_getRaw0(uint8_t idx) {
  if (idx<4) {
    return get_absolute0(idx);
  }
  return 0;
}


uint16_t floor_getRaw1(uint8_t idx) {
  if (idx<4) {
    return get_absolute1(idx);
  }
  return 0;
}


uint16_t floor_getRawDiff(uint8_t idx) {
  if (idx<4) {
    return get_differential(idx);
  }
  return 0;
}


void floor_calibrateBlackCh(uint8_t idx) {
  uint8_t i;
  line_cal.black[idx]=0;
  for (i=0; i<CALSAMPLES; ++i) {    
    analog_wait_update();
    line_cal.black[idx] += get_differential(idx);
  }
  floor_calculateCalibration();
}


void floor_calibrateWhiteCh(uint8_t idx) {
  uint8_t i;
  line_cal.white[idx]=0;
  for (i=0; i<CALSAMPLES; ++i) {
    analog_wait_update();
    line_cal.white[idx] += get_differential(idx);
  }
  floor_calculateCalibration();
}


void floor_calibrateBlack() {
  uint8_t i;
  line_cal.black[0]=0;
  line_cal.black[1]=0;
  line_cal.black[2]=0;
  line_cal.black[3]=0;
  for (i=0; i<CALSAMPLES; ++i) {
    analog_wait_update();
    line_cal.black[0] += get_differential(0);
    line_cal.black[1] += get_differential(1);
    line_cal.black[2] += get_differential(2);
    line_cal.black[3] += get_differential(3);
  }
  floor_calculateCalibration();
}


void floor_calibrateWhite() {
  uint8_t i;
  line_cal.white[0]=0;
  line_cal.white[1]=0;
  line_cal.white[2]=0;
  line_cal.white[3]=0;
  for (i=0; i<CALSAMPLES; ++i) {
    analog_wait_update();
    line_cal.white[0] += get_differential(0);
    line_cal.white[1] += get_differential(1);
    line_cal.white[2] += get_differential(2);
    line_cal.white[3] += get_differential(3);
  }
  floor_calculateCalibration();
}


void floor_readPersistent() {
  eeprom_read_block (&line_cal, LINE_EEPROM, sizeof(line_cal_t));
  if (line_cal.black[0]==0xffff) {
    line_cal.black[0] = 0;
    line_cal.black[1] = 0;
    line_cal.black[2] = 0;
    line_cal.black[3] = 0;
    line_cal.white[0] = CALSAMPLES * 0x80;
    line_cal.white[1] = CALSAMPLES * 0x80;
    line_cal.white[2] = CALSAMPLES * 0x80;
    line_cal.white[3] = CALSAMPLES * 0x80;
  }
  floor_calculateCalibration();
}


void floor_writePersistent() {
  eeprom_write_block (&line_cal, LINE_EEPROM, sizeof(line_cal_t));
}


/* OBSOLETE functionality - will be removed... */

uint16_t floor_absolute [4]; 
uint16_t floor_relative [4];

void floor_update() {
  sei();
  floor_absolute[0]=floor_getAbs(0);
  floor_absolute[1]=floor_getAbs(1);
  floor_absolute[2]=floor_getAbs(2);
  floor_absolute[3]=floor_getAbs(3);
  floor_relative[0]=floor_get(0);
  floor_relative[1]=floor_get(1);
  floor_relative[2]=floor_get(2);
  floor_relative[3]=floor_get(3);
}


#ifdef __cplusplus
} // extern "C"
#endif
