/*  BSD-License:

Copyright (c) 2007 by Nils Springob, nicai-systems, Germany

All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

  * Redistributions of source code must retain the above copyright notice,
    this list of conditions and the following disclaimer.
  * Redistributions in binary form must reproduce the above copyright notice,
    this list of conditions and the following disclaimer in the documentation
    and/or other materials provided with the distribution.
  * Neither the name nicai-systems nor the names of its contributors may be
    used to endorse or promote products derived from this software without
    specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

*/

/*! @file    line.c
 *  @brief   Zuordnung der physikalischen Pins zu symbolischen Namen
 *  @author  Nils Springob (nils@nicai-systems.de)
 *  @date    2009-08-19
 */

#include "nibobee/iodefs.h"
#include "nibobee/line.h"
#include "nibobee/base.h"
#include "nibobee/delay.h"
#include <avr/eeprom.h>

#define LINE_EEPROM ((void *)0)

#define RANGE 0x400
#define FACTOR 0x100
#define CALSAMPLES 0x20

#define GAIN_FACT ((uint32_t)RANGE * (uint32_t)FACTOR * (uint32_t)CALSAMPLES)
//      0x800000L =          0x400 *           0x100  *           0x20

#ifdef __cplusplus
extern "C" {
#endif

static uint16_t cal_gain[3];
static uint16_t cal_offset[3];

typedef struct { 
  uint16_t black[3]; // sum of CALSAMPLES samples
  uint16_t white[3]; // sum of CALSAMPLES samples
} line_cal_t;

line_cal_t line_cal;


void line_calculateCalibration();

void line_init() {
  if (!(nibobee_initialization & NIBOBEE_ANALOG_INITIALIZED)) {
    analog_init();
  }
  line_readPersistent();
  activate_output_bit(IO_LINE_EN);
}


void line_writePersistent() {
  eeprom_write_block (&line_cal, LINE_EEPROM, sizeof(line_cal_t));
}


void line_readPersistent() {
  eeprom_read_block (&line_cal, LINE_EEPROM, sizeof(line_cal_t));
  if (line_cal.white[0]==0xffff) {
    line_cal.white[0] = CALSAMPLES*0x200;
    line_cal.white[1] = CALSAMPLES*0x300;
    line_cal.white[2] = CALSAMPLES*0x200;
    line_cal.black[0] = CALSAMPLES*4;
    line_cal.black[1] = CALSAMPLES*4;
    line_cal.black[2] = CALSAMPLES*4;
  }
  line_calculateCalibration();
}

/////// NEW

static uint16_t get_differential(uint8_t idx) {
  // attention: black is 5V!!
  int16_t res = (int16_t)analog_getValue(ANALOG_L1+idx)-(int16_t)analog_getValue(ANALOG_L0+idx);
  return (res>0)?res:0;
}

/*
static uint16_t get_absolute1(uint8_t idx) {
  int16_t res = (int16_t)analog_getValue(ANALOG_L1+idx);
  return (res>0)?res:0;
}
*/

static uint16_t get_absolute0(uint8_t idx) {
  int16_t res = (int16_t)analog_getValue(ANALOG_L0+idx);
  return (res>0)?res:0;
}


// new version...
void line_calculateCalibration() {
  // 0x800000L = 0x400 * 0x100 * 0x20
  // GAIN_FACT = range   factor calib
  cal_gain[0] = GAIN_FACT / (line_cal.white[0]-line_cal.black[0]);
  cal_gain[1] = GAIN_FACT / (line_cal.white[1]-line_cal.black[1]);
  cal_gain[2] = GAIN_FACT / (line_cal.white[2]-line_cal.black[2]);
  cal_offset[0] = line_cal.black[0] / CALSAMPLES;
  cal_offset[1] = line_cal.black[1] / CALSAMPLES;
  cal_offset[2] = line_cal.black[2] / CALSAMPLES;
}

/*
 * Differenz von LED an und aus mit Schwarzoffset
 */
uint16_t line_get(uint8_t idx) {
  if (idx<3) {
    uint16_t val16 = get_differential(idx);
    val16 -= (val16>cal_offset[idx])?cal_offset[idx]:val16;
    uint32_t val = val16;
    val *= cal_gain[idx];
    val /= FACTOR;
    if (val>INT16_MAX) {
      return INT16_MAX;
    }
    return val;
  }
  return 0;
} 


/*
 * Wert bei LED aus ohne Schwarzoffset
 */
uint16_t line_getAbs(uint8_t idx) {
  if (idx<3) {
    uint16_t val16 = get_absolute0(idx);
    uint32_t val = val16;
    val *= cal_gain[idx];
    val /= FACTOR;
    if (val>INT16_MAX) {
      return INT16_MAX;
    }
    return val;
  }
  return 0;
}

void line_calibrateBlack() {
  uint8_t i;
  line_cal.black[0]=0;
  line_cal.black[1]=0;
  line_cal.black[2]=0;
  for (i=0; i<CALSAMPLES; ++i) {
    analog_wait_update();
    line_cal.black[0] += get_differential(0);
    line_cal.black[1] += get_differential(1);
    line_cal.black[2] += get_differential(2);
  }
  line_calculateCalibration();
}


void line_calibrateWhite() {
  uint8_t i;
  line_cal.white[0]=0;
  line_cal.white[1]=0;
  line_cal.white[2]=0;
  for (i=0; i<CALSAMPLES; ++i) {
    analog_wait_update();
    line_cal.white[0] += get_differential(0);
    line_cal.white[1] += get_differential(1);
    line_cal.white[2] += get_differential(2);
  }
  line_calculateCalibration();
}

#ifdef __cplusplus
} // extern "C"
#endif

