/*  BSD-License:

Copyright (c) 2007 by Nils Springob, nicai-systems, Germany

All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

  * Redistributions of source code must retain the above copyright notice,
    this list of conditions and the following disclaimer.
  * Redistributions in binary form must reproduce the above copyright notice,
    this list of conditions and the following disclaimer in the documentation
    and/or other materials provided with the distribution.
  * Neither the name nicai-systems nor the names of its contributors may be
    used to endorse or promote products derived from this software without
    specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

*/

/**
 * This program converts an intel hex file to a C char array.
 * The sourcefilename, outputfilename and the size of the array must be given.
 */

#include <boost/program_options.hpp>
#include <boost/array.hpp>
#include <iostream>
#include <iomanip>
#include <fstream>
#include <sstream>
#if defined _WIN32 || defined __CYGWIN__
typedef unsigned char uint8_t;
#else
#include <stdint.h>
#endif

namespace po = boost::program_options;

using namespace std;

class parse_exception : std::runtime_error {
  public:
    parse_exception(const std::string &str) : std::runtime_error(str)
    {
    }
};


int parse_hex(string str) {
  unsigned int val = 0;
  istringstream i(str);
  i >> hex;
  i >> val;
  return val;
}

void read_inputline(string line, std::vector<uint8_t> & mem)
{
  //  0123456790123456790123456790123456790123456
  //  :ssaaaattddddddddddddddddddddddddddddddddcc
  // ":1000000010C029C028C027C026C025C0B7C023C043"
  // ":1007C00010F0A61BB70B881F991F5A95A9F78095A3"
  // ":0E07D0009095682F792F8A2F9B2F0895FFCFC9"
  // ":00000001FF"
  
  unsigned int length = line.size();
  if (length==0) {
    return;
  }

  if (line[length-1]==13) {
    line.erase(--length);
  }
  if (length<11) {
    throw runtime_error("error while parsing \""+line+"\"");
  }
  if (line[0]!=':') {
    throw runtime_error("error while parsing \""+line+"\"");
  }
  unsigned int size = 0xff & parse_hex(line.substr(1, 2));

  if (length!=2*size+11) {
    throw runtime_error("line length error \""+line+"\"");
  }
  
  unsigned int address = 0xffff & parse_hex(line.substr(3, 4));
  unsigned int type = 0xff & parse_hex(line.substr(7, 2));

  if (type==0) {
    for (unsigned int i=0; i<size; ++i) {
      unsigned int val = 0xff & parse_hex(line.substr(2*i+9, 2));
      mem[address++] = val;
      if (address>=mem.size()) {
        throw runtime_error("memory overflow while parsing \""+line+"\"");
      }
    }
  }
}

int read_input(istream & in, std::vector<uint8_t> & mem) {
  string line;
  while (getline(in, line))
    read_inputline(line, mem);

  return 0;
}

int write_output(ostream & out, std::vector<uint8_t> & mem) {
  int length = mem.size();
  out << "{" << endl;
  for (unsigned int i=0; i<length; ++i) {
    if (i%16==0) {
      out << "  /* 0x" << setfill('0') << setw(4) << hex << i << " */ ";
    }
    out << "0x" << setfill('0') << setw(2) << hex << (unsigned int)mem[i] << ", ";
    if (i%16==15) {
      out << endl;
    }
  }
  
  out << "};" << endl;
  cout << "write successful.\n";
  return 0;
}

int hex2carray(string inputfilename, string outputfilename, int memsize) {
  std::vector<uint8_t> mem(memsize,0xff);
  
  ifstream in(inputfilename.c_str());
  if (!in) {
    cout << "Error opening input file.\n";
    return -1;
  }
  
  ofstream out(outputfilename.c_str());
  if (!out) {
    cout << "Error opening output file.\n";
    return -1;
  }
  
  int res = read_input(in, mem);
  if (res<0) return res;
  cout << "read successful.\n";
  return write_output(out, mem);
}


int main (int argc, char * argv[]) {
  // Declare the supported options.
  po::options_description desc("Allowed options");
  desc.add_options()
    ("help", "produce help message")
    ("size,s", po::value<int>(), "set output size")
    ("input-file", po::value< string >(), "input file")
    ("output-file,o", po::value< string >(), "output file")
  ;

  po::positional_options_description p;
  p.add("input-file", 1);
  p.add("output-file", 1);

  po::variables_map vm;
  po::store(po::command_line_parser(argc, argv).
          options(desc).positional(p).run(), vm);
  po::notify(vm);

  if (vm.count("help")) {
    cout << desc << "\n";
    return 1;
  }

  if (!vm.count("size")) {
    cout << "Size was not set.\n";
    return -1;
  }
  
  if (!vm.count("input-file")) {
    cout << "Please specify an input file.\n";
    return -1;
  }
  
  if (!vm.count("output-file")) {
    cout << "Please specify an output file.\n";
    return -1;
  }

  int memsize = vm["size"].as<int>();
  string inputfilename = vm["input-file"].as<string>();
  string outputfilename = vm["output-file"].as<string>();
   
  cout << "Size is " << memsize << ".\n";
  cout << "Input file is " << inputfilename << "\n";
  cout << "Output file is " << outputfilename << "\n";
  
  try {
    return hex2carray(inputfilename, outputfilename, memsize);
  } catch (const exception & e) {
    cout << flush;
    cerr << "Error: " << e.what() << endl;
    return -1;
  }
}

